-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("achievementTracker", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LATr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-AchievementTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local CreateFont, CreateFrame, SetFont, ShowUIPanel, GetTime = CreateFont, CreateFrame, SetFont, ShowUIPanel, GetTime
local unpack, format, pairs, ipairs, sub, band, wipe = unpack, string.format, pairs, ipairs, string.sub, bit.band, wipe
local tonumber, time, SecondsToTime, ItemRefTooltip = tonumber, time, SecondsToTime, ItemRefTooltip
local GetAchievementInfo, GetAchievementLink, ChatEdit_InsertLink = GetAchievementInfo, GetAchievementLink, ChatEdit_InsertLink
local GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements = GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements
local GetAchievementNumCriteria, GetAchievementCriteriaInfo = GetAchievementNumCriteria, GetAchievementCriteriaInfo
local UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown

-----------------
--- Variables ---
-----------------
local _, prdb, chdb, options

module.numTrackedAchievement = 0
module.CELL_MARGIN_H = 6
module.modName = LATr["Achievement Tracker"]
module.trackerAnchor = "ANCHOR_RIGHT"
module.cATrFontTQ2 = CreateFont("cATrFontTQ2")
module.scATrFontTQ2 = CreateFont("scATrFontTQ2")
module.aATrFontTQ2 = CreateFont("aATrFontTQ2")
module.acATrFontTQ2 = CreateFont("acATrFontTQ2")
module.sbATrFontTQ2 = CreateFont("sbATrFontTQ2")
module.fixATrFontTQ2 = CreateFont("fixATrFontTQ2")
module.plus = "Interface\\Buttons\\UI-PlusButton-Up"
module.minus = "Interface\\Buttons\\UI-MinusButton-Up"

module.category, module.achievements, module.visibleAchievements, module.visibleCriteria = {}, {}, {}, {}
module.anchorPosition = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"}
module.outlines = {[""] = LATr["None"], ["OUTLINE"] = LATr["Outline"], ["THICKOUTLINE"] = LATr["Thick Outline"]}
module.achievementPosition = {[1] = LATr["Above"], [2] = LATr["Below"]}
module.anchorSetPoint = {
	[1] = {point = "TOPLEFT", relPoint = "BOTTOMLEFT", x = 0, y = 24},
	[2] = {point = "TOP", relPoint = "BOTTOM", x = 0, y = 24},
	[3] = {point = "TOPRIGHT", relPoint = "BOTTOMRIGHT", x = 0, y = 24},
	[4] = {point = "BOTTOMLEFT", relPoint = "TOPLEFT", x = 0, y = -24},
	[5] = {point = "BOTTOM", relPoint = "TOP", x = 0, y = -24},
	[6] = {point = "BOTTOMRIGHT", relPoint = "TOPRIGHT", x = 0, y = -24}}
module.interactivityList = {
	["chatLink"] = LATr["Insert achievement link in chat"],
	["achievementTooltip"] = LATr["Display of achievement tooltip"],
	["removeTrackedAchievement"] = LATr["Remove achievement from tracker"],
	["displayAchievement"] = LATr["Diplay achievement frame"],
	["nothing"] = LATr["Do nothing"],
}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		timedAchievements = true,
		point = "CENTER",
		x = 0,
		y = 0,
		showInQuestTracker = false,
		positionInQuestTracker = 1,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedCriteria = true,
		showByCategory = true,
		categoryFont = LSM.DefaultMedia.font,
		categoryFontSize = 12,
		categoryFontOutline = "",
		categoryCollapseButtonSize =13,
		subCategoryFont = LSM.DefaultMedia.font,
		subCategoryFontSize = 12,
		subCategoryFontOutline = "",
		subCategoryCollapseButtonSize = 13,
		achievementFont = LSM.DefaultMedia.font,
		achievementFontSize = 10,
		achievementFontOutline = "",
		displayAchievementImage = true,
		achievementImageSize = 13,
		achievementCriteriaFont = LSM.DefaultMedia.font,
		achievementCriteriaFontSize = 8,
		achievementCriteriaFontOutline = "",
		statusBarFont = LSM.DefaultMedia.font,
		statusBarFontSize = 8,
		statusBarFontOutline = "",
		numBooleanCriteriaColumns = 2,
		maxWidth = 250,
		minWidth = 250,
		statusMaxWidth = 70,
		statusMinWidth = 70,
		hasMaxHeight = true,
		maxHeight = 400,
		subCategoryCellPadding = 5,
		achievementCellPadding = 5,
		criteriaCellPadding = 10,
		rightCellPadding = 0,
		scale = 1,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		colorizeACName = true,
		defaultACNameColor = "ffffff",
		colorizeACCounter = true,
		defaultACCounterColor = "ffffff",
		colorizeASCName = true,
		defaultASCNameColor = "ffffff",
		colorizeASCCounter = true,
		defaultASCCounterColor = "ffffff",
		colorizeAname = true,
		defaultANameColor = "ffffff",
		colorizeATimer = true,
		defaultATimerColor = "ffffff",
		colorizeACrName = true,
		defaultACrNameColor = "ffffff",
		colorizeACrStatus = true,
		defaultACrStatusColor = "ffffff",
		displayAsStatusBar = true,
		statusBarTexture = LSM.DefaultMedia.statusbar,
		colorizeStatusBar = true,
		defaultStatusBarColor = {0, 0.6, 0},
		colorizeStatusBarText = false,
		defaultStatusBarTextColor = "ffffff",
		defaultDescriptionColor = "ffffff",
		showSBDescription = true,
		statusBarHeight = 12,
		leftStatusBarPadding = 40,
		rightStatusBarPadding = 40,
		disableInteraction = false,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "nothing",
		mouseRightAlt = "nothing",
		mouseRight = "achievementTooltip",
		mouseLeftShift = "removeTrackedAchievement",
		mouseLeftControl = "nothing",
		mouseLeftAlt = "nothing",
		mouseLeft = "displayAchievement",
		mouseMiddleShift = "nothing",
		mouseMiddleControl = "nothing",
		mouseMiddleAlt = "nothing",
		mouseMiddle = "nothing",
	},
}

---------------
--- Options ---
---------------
local function normalizeWidth(min, max)
	if max < min then return max else return min end
end
local optGetter, optSetter, displayAchievementImageSetter, achievementImageSizeSetter, achievementCellPaddingSetter, criteriaCellPaddingSetter, showByCategorySetter, categoryCollapseButtonSizeSetter, subCategoryCellPaddingSetter, subCategoryCollapseButtonSizeSetter, maxWidthSetter, minWidthSetter, statusMaxWidthSetter, statusMinWidthSetter, interactionSetter, numBooleanCriteriaColumnsSetter, taSetter, questTrackerSetter, colorGetter, colorSetter, hexColorGetter, hexColorSetter, catFontSetter, subCatFontSetter, achFontSetter, achCritFontSetter, sbFontSetter, anchorSetter, combatSetter, interactionDisabled
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function displayAchievementImageSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if value then
			module.achievementImageSize = prdb.achievementImageSize
			module.achievementCellPadding = prdb.achievementCellPadding + module.achievementImageSize
		else
			module.achievementImageSize = 0
			module.achievementCellPadding = prdb.achievementCellPadding
		end
		if prdb.showByCategory then
			module.subAchievementCellPadding = module.achievementCellPadding + module.subCategoryCellPadding
		else
			module.subAchievementCellPadding = module.achievementCellPadding
		end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function achievementImageSizeSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if prdb.displayAchievementImage then
			module.achievementImageSize = value
			module.achievementCellPadding = prdb.achievementCellPadding + value
			if prdb.showByCategory then
				module.subAchievementCellPadding = module.achievementCellPadding + module.subCategoryCellPadding
			else
				module.subAchievementCellPadding = module.achievementCellPadding
			end
		end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function achievementCellPaddingSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if prdb.displayAchievementImage then
			module.achievementCellPadding = value + module.achievementImageSize
		else
			module.achievementCellPadding = value
		end
		module.achievementIconPadding = value
		if prdb.showByCategory then
			module.subAchievementCellPadding = module.achievementCellPadding + module.subCategoryCellPadding
			module.subAchievementIconPadding = value + module.subCategoryCellPadding
		else
			module.subAchievementCellPadding = module.achievementCellPadding
			module.subAchievementIconPadding = value
		end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function criteriaCellPaddingSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.criteriaCellPadding = value
		if prdb.showByCategory then
			module.subCriteriaCellPadding = value + module.subCategoryCellPadding
		else
			module.subCriteriaCellPadding = value
		end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function showByCategorySetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if value then
			module.subAchievementCellPadding = module.achievementCellPadding + module.subCategoryCellPadding
			module.subAchievementIconPadding = prdb.achievementCellPadding + module.subCategoryCellPadding
			module.subCriteriaCellPadding = module.criteriaCellPadding + module.subCategoryCellPadding
		else
			module.subAchievementCellPadding = module.achievementCellPadding
			module.subAchievementIconPadding = module.achievementIconPadding
			module.subCriteriaCellPadding = module.criteriaCellPadding
		end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function categoryCollapseButtonSizeSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.categoryCellPadding = value
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function subCategoryCellPaddingSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.subCategoryCellPadding = value + prdb.subCategoryCollapseButtonSize
		if prdb.showByCategory then
			module.subAchievementCellPadding = module.achievementCellPadding + module.subCategoryCellPadding
			module.subAchievementIconPadding = prdb.achievementCellPadding + module.subCategoryCellPadding
			module.subCriteriaCellPadding = module.criteriaCellPadding + module.subCategoryCellPadding
		end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function subCategoryCollapseButtonSizeSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.subCategoryCellPadding = prdb.subCategoryCellPadding + value
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateAchievementTracker()
	end
	function maxWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
			QTIP:Release(module.tracker)
			module.tracker = nil
			module.maxWidth = value
			module.minWidth = normalizeWidth(prdb.minWidth, value)
			module:updateAchievementTracker()
		end
	end
	function minWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
			QTIP:Release(module.tracker)
			module.tracker = nil
			module.minWidth = normalizeWidth(value, prdb.maxWidth)
			module:updateAchievementTracker()
		end
	end
	function statusMaxWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
			QTIP:Release(module.tracker)
			module.statusMaxWidth = value
			module.statusMinWidth = normalizeWidth(prdb.statusMinWidth, value)
			module:updateAchievementTracker()
		end
	end
	function statusMinWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
			QTIP:Release(module.tracker)
			module.statusMinWidth = normalizeWidth(value, prdb.statusMaxWidth)
			module:updateAchievementTracker()
		end
	end
	function interactionSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
			QTIP:Release(module.tracker)
			module.tracker = nil
			module.disableInteraction = value
			module:updateAchievementTracker()
		end
	end
	function numBooleanCriteriaColumnsSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
			QTIP:Release(module.tracker)
			module.tracker = nil
			module.columnMaxWidth = module.maxWidth / value
			module.totalMaxWidth = module.maxWidth + module.statusMaxWidth + module.CELL_MARGIN_H * value
			module.columnMinWidth = normalizeWidth(module.minWidth / value, module.columnMaxWidth)
			module.totalMinWidth = normalizeWidth(module.minWidth + module.statusMinWidth + module.CELL_MARGIN_H * value, module.totalMaxWidth)
			module:updateAchievementTracker()
		end
	end
	function questTrackerSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		if not value or not core.moduleState["questsTracker"] then
			QTIP:Release(module.tracker)
			module.tracker = nil
			module.maxWidth = prdb.maxWidth
			module.minWidth = normalizeWidth(prdb.minWidth, module.maxWidth)
			module.statusMaxWidth = prdb.statusMaxWidth
			module.statusMinWidth = normalizeWidth(prdb.statusMinWidth, module.statusMaxWidth)
			module.columnMaxWidth = module.maxWidth / prdb.numBooleanCriteriaColumns
			module.totalMaxWidth = module.maxWidth + module.statusMaxWidth + module.CELL_MARGIN_H * prdb.numBooleanCriteriaColumns
			module.columnMinWidth = normalizeWidth(module.minWidth / prdb.numBooleanCriteriaColumns, module.columnMaxWidth)
			module.totalMinWidth = normalizeWidth(module.minWidth + module.statusMinWidth + module.CELL_MARGIN_H * prdb.numBooleanCriteriaColumns, module.totalMaxWidth)
			core.cbh:Fire("updateQuestsTracker")
			module:updateAchievementTracker()
		elseif core.moduleState["questsTracker"] then
			core.cbh:Fire("updateAchievementsTrackerParameters")
		end
	end
	function taSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if value then
			module:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
		else
			module:UnregisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
		end
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, r, g, b, a)
		prdb[info[#info]][1] = r
		prdb[info[#info]][2] = g
		prdb[info[#info]][3] = b
		prdb[info[#info]][4] = a
		if not core.moduleState["achievementTracker"] then return end
		module:updateAchievementTracker()
	end
	function hexColorGetter(info)
		return tonumber(sub(prdb[info[#info]], 1, 2), 16) / 255, tonumber(sub(prdb[info[#info]], 3, 4), 16) / 255, tonumber(sub(prdb[info[#info]], 5, 6), 16) / 255
	end
	function hexColorSetter(info,r, g, b)
		if not core.moduleState["achievementTracker"] then return end
		prdb[info[#info]] = format("%02x%02x%02x", r*255, g*255, b*255)
		module:updateAchievementTracker()
	end
	function catFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
		module:updateAchievementTracker()
	end
	function subCatFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
		module:updateAchievementTracker()
	end
	function achFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
		module:updateAchievementTracker()
	end
	function achCritFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
		module:updateAchievementTracker()
	end
	function sbFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.sbATrFontTQ2:SetFont(LSM:Fetch("font", prdb.statusBarFont), prdb.statusBarFontSize, prdb.statusBarFontOutline)
		module:updateAchievementTracker()
	end
	function anchorSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		if value then
			module.anchor:Show()
		else
			module.anchor:Hide()
		end
		module:updateAchievementTracker()
	end
	function combatSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["achievementTracker"] then return end
		module.hideInCombat = nil
		if value then
			core.RegisterCallback(module, "outOfCombat")
			core.RegisterCallback(module, "inCombat")
		else
			core.UnregisterCallback(module, "outOfCombat")
			core.UnregisterCallback(module, "inCombat")
		end
		module:updateAchievementTracker()
	end
	function interactionDisabled()
		if prdb.disableInteraction then
			return true
		else
			return false
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				autotrack = {
					type = "group",
					name = LATr["Auto track"],
					args = {
						timedAchievements = {
							name = LATr["Timed achievements"],
							type = "toggle",
							set = taSetter,
							width = "full",
							order = 1,
						},
					},
				},
				display = {
					type = "group",
					name = LATr["Display"],
					args = {
						showInQuestTracker = {
							name = LATr["Show achievements in quests tracker"],
							type = "toggle",
							set = questTrackerSetter,
							width = "full",
							order = 1,
						},
						positionInQuestTracker = {
							name = LATr["Position of achievements relative to quests in the tracker"],
							type = "select",
							values = module.achievementPosition,
							set = questTrackerSetter,
							width = "full",
							order = 2,
						},
						showAnchor = {
							name = LATr["Show achievement tracker anchor"],
							type = "toggle",
							set = anchorSetter,
							width = "full",
							order = 3,
						},
						anchorPosition = {
							name = LATr["Anchor position"],
							type = "select",
							values = module.anchorPosition,
							width = "full",
							order = 4,
						},
						toggleTracker = {
							name = LATr["Display tracker"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						hideTrackerInCombat = {
							name = LATr["Hide achievement tracker in combat"],
							type = "toggle",
							set = combatSetter,
							width = "full",
							order = 6,
						},
						hideCompletedCriteria = {
							name = LATr["Hide completed achievement criteria"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						showByCategory = {
							name = LATr["Display achievements by category and subcategory"],
							type = "toggle",
							set = showByCategorySetter,
							width = "full",
							order = 8,
						},
					},
				},
				fonts = {
					type = "group",
					name = LATr["Tracker fonts"],
					args = {
						category = {
							type = "group",
							name = LATr["Achievement category"],
							set = catFontSetter,
							inline = true,
							order = 1,
							args = {
								categoryFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								categoryFontSize = {
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								categoryFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
						subCategory = {
							type = "group",
							name = LATr["Achievement sub category"],
							set = subCatFontSetter,
							inline = true,
							order = 2,
							args = {
								subCategoryFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								subCategoryFontSize = {
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								subCategoryFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
						achievement = {
							type = "group",
							name = LATr["Achievement"],
							set = achFontSetter,
							inline = true,
							order = 3,
							args = {
								achievementFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								achievementFontSize = {
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								achievementFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
						achievementCriteria = {
							type = "group",
							name = LATr["Achievement criteria"],
							set = achCritFontSetter,
							inline = true,
							order = 4,
							args = {
								achievementCriteriaFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								achievementCriteriaFontSize = { 
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								achievementCriteriaFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
						statusBar = {
							type = "group",
							name = LATr["Status Bar"],
							set = sbFontSetter,
							inline = true,
							order = 5,
							args = {
								statusBarFont = {
									name = LATr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								statusBarFontSize = {
									name = LATr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								statusBarFontOutline = {
									name = LATr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
					},
				},
				appearance = {
					type = "group",
					name = LATr["Tracker appearance"],
					args = {
						scale = {
							name = LATr["Achievement tracker scale"],
							type = "range",
							min = 0,
							max = 1,
							step = 0.1,
							width = "full",
							order = 1,
						},
						hasMaxHeight = {
							name = LATr["Set achievements tracker maximum height after which it will be scrollable"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						maxHeight = {
							name = LATr["Achievement tracker max height"],
							type = "range",
							min = 100,
							max = 2000,
							step = 5,
							width = "full",
							order = 3,
						},
						categoryCollapseButtonSize = {
							name = LATr["Category collapse Button size"],
							type = "range",
							set = categoryCollapseButtonSizeSetter,
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 4,
						},
						subCategoryCollapseButtonSize = {
							name = LATr["Sub category collapse Button size"],
							type = "range",
							set = subCategoryCollapseButtonSizeSetter,
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 5,
						},
						displayAchievementImage = {
							name = LATr["Display achievements icon"],
							type = "toggle",
							set = displayAchievementImageSetter,
							width = "full",
							order = 6,
						},
						achievementImageSize = {
							name = LATr["Achievement image size"],
							type = "range",
							set = achievementImageSizeSetter,
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 7,
						},
						subCategoryCellPadding = {
							name = LATr["Sub category cell padding"],
							type = "range",
							set = subCategoryCellPaddingSetter,
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 8,
						},
						achievementCellPadding = {
							name = LATr["Achievement cell padding"],
							type = "range",
							set = achievementCellPaddingSetter,
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 9,
						},
						criteriaCellPadding = {
							name = LATr["Achievement criteria cell padding"],
							type = "range",
							set = criteriaCellPaddingSetter,
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 10,
						},
						rightCellPadding = {
							name = LATr["Right-most cell right padding to avoid cell clipping"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							width = "full",
							order = 11,
						},
						maxWidth = {
							name = LATr["Achievement tracker column max width"],
							type = "range",
							set = maxWidthSetter,
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 12,
						},
						minWidth = {
							name = LATr["Achievement tracker column min width"],
							type = "range",
							set = minWidthSetter,
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 13,
						},
						statusMaxWidth = {
							name = LATr["Achievement Tracker status column max width"],
							type = "range",
							set = statusMaxWidthSetter,
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 14,
						},
						statusMinWidth = {
							name = LATr["Achievement Tracker status column min width"],
							type = "range",
							set = statusMinWidthSetter,
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 15,
						},
						numBooleanCriteriaColumns = {
							name = LATr["Number of columns in the achievement tracker for boolean criteria"],
							type = "range",
							set = numBooleanCriteriaColumnsSetter,
							min = 1,
							max = 8,
							step = 1,
							width = "full",
							order = 16,
						},
					},
				},
				color = {
					type = "group",
					name = LATr["Colors"],
					args = {
						backDropColor = {
							name = LATr["Achievement tracker background color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 1,
						},
						borderColor = {
							name = LATr["Achievement tracker border color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 2,
						},
						colorizeACName = {
							name = LATr["Color achievements category name"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						defaultACNameColor = {
							name = LATr["Default achievements name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 4,
						},
						colorizeACCounter = {
							name = LATr["Color achievements category counter"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						defaultACCounterColor = {
							name = LATr["Default achievements category counter color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 6,
						},
						colorizeASCName = {
							name = LATr["Color achievements subcategory name"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						defaultASCNameColor = {
							name = LATr["Default achivements subcategory name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 8,
						},
						colorizeASCCounter = {
							name = LATr["Color achievements subcategory counter"],
							type = "toggle",
							width = "full",
							order = 9,
						},
						defaultASCCounterColor = {
							name = LATr["Default achievements subcategory counter color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 10,
						},
						colorizeAname = {
							name = LATr["Color achievements name"],
							type = "toggle",
							width = "full",
							order = 11,
						},
						defaultANameColor = {
							name = LATr["Default achievements name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 12,
						},
						colorizeATimer = {
							name = LATr["Color achievement timer"],
							type = "toggle",
							width = "full",
							order = 13,
						},
						defaultATimerColor = {
							name = LATr["Default achievement timer color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 14,
						},
						colorizeACrName = {
							name = LATr["Color achievements criteria name"],
							type = "toggle",
							width = "full",
							order = 15,
						},
						defaultACrNameColor = {
							name = LATr["Default achievements criteria name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 16,
						},
						colorizeACrStatus = {
							name = LATr["Color achievements cirteria status"],
							type = "toggle",
							width = "full",
							order = 17,
						},
						defaultACrStatusColor = {
							name = LATr["Default achievements criteria status color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 18,
						},
						colorizeStatusBar = {
							name = LATr["Color status bar"],
							type = "toggle",
							width = "full",
							order = 19,
						},
						defaultStatusBarColor = {
							name = LATr["Default status bar color"],
							type = "color",
							hasAlpha = false,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 20,
						},
						colorizeStatusBarText ={
							name = LATr["Color status bar text"],
							type = "toggle",
							width = "full",
							order = 21,
						},
						defaultStatusBarTextColor = {
							name = LATr["Default status bar text color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 22,
						},
						defaultDescriptionColor = {
							name = LATr["Default achievement description color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 23,
						},
					},
				},
				statusBar = {
					type = "group",
					name = LATr["Status Bar"],
					args = {
						displayAsStatusBar = {
							name = LATr["Display status bar for criteria progression when possible"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						statusBarTexture = {
							name = LATr["Status bar texture"],
							type = "select",
							dialogControl = 'LSM30_Statusbar',
							values = AceGUIWidgetLSMlists.statusbar,
							order = 2,
						},
						statusBarHeight = {
							name = LATr["Status bar height"],
							type = "range",
							min = 6,
							max = 20,
							step = 1,
							width = "full",
							order = 3,
						},
						leftStatusBarPadding = {
							name = LATr["Left padding of the status bar"],
							type = "range",
							min = 0,
							max = 100,
							step = 1,
							width = "full",
							order = 4,
						},
						rightStatusBarPadding = {
							name = LATr["Right padding of the status bar"],
							type = "range",
							min = 0,
							max = 100,
							step = 1,
							width = "full",
							order = 5,
						},
					},
				},
				interaction = {
					type = "group",
					name = LATr["Tracker interaction"],
					args = {
						disableInteraction = {
							type = "toggle",
							name = LATr["Completely disable tracker interaction, should not be used when display by category is used or if you wish to make it scrollable."],
							set = interactionSetter,
							width = "full",
							order = 1,
						},
						overTooltip = {
							name = LATr["Display achievement tooltip on mouse over"],
							type = "toggle",
							width = "full",
							hidden = interactionDisabled,
							order = 2,
						},
						mouseRightShift = {
							name = LATr["Action on right shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 3,
						},
						mouseRightControl = {
							name = LATr["Action on right control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 4,
						},
						mouseRightAlt = {
							name = LATr["Action on right alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 5,
						},
						mouseRight = {
							name = LATr["Action on right click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 6,
						},
						mouseLeftShift = {
							name = LATr["Action on left shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 7,
						},
						mouseLeftControl = {
							name = LATr["Action on left control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 8,
						},
						mouseLeftAlt = {
							name = LATr["Action on left alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 9,
						},
						mouseLeft = {
							name = LATr["Action on left click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 10,
						},
						mouseMiddleShift = {
							name = LATr["Action on middle shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 11,
						},
						mouseMiddleControl = {
							name = LATr["Action on middle control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 12,
						},
						mouseMiddleAlt = {
							name = LATr["Action on middle alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 13,
						},
						mouseMiddle = {
							name = LATr["Action on middle click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 14,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("achievementTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	chdb.achievementTracked = chdb.achievementTracked or {}
	chdb.categoryDisplay = chdb.categoryDisplay or {}
	self:SetEnabledState(core:getModuleEnabled("achievementTracker"))
	core.moduleState["achievementTracker"] = core:getModuleEnabled("achievementTracker")
	core:registerModuleOptions("achievementTracker", getOptions, module.modName)
end

function module:OnEnable()
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
	self.cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	self.scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	self.aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
	self.acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	self.sbATrFontTQ2:SetFont(LSM:Fetch("font", prdb.statusBarFont), prdb.statusBarFontSize, prdb.statusBarFontOutline)
	self.fixATrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	self.anchor = CreateFrame("Button", nil, UIParent)
	self.anchor:SetHeight(24)
	self.anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	self.anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
	self.anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
	self.anchor.text = self.anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
	self.anchor.text:SetPoint("CENTER")
	self.anchor.text:SetText(LATr["Achievement Tracker"])
	self.anchor:SetWidth(self.anchor.text:GetStringWidth() + 8)
	self.anchor:SetMovable(true)
	self.anchor:RegisterForDrag("LeftButton")
	self.anchor:SetClampedToScreen(true)
	self.anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	self.anchor:SetScript("OnDragStart", function(self) self:StartMoving() end)
	self.anchor:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
		module.trackerAnchor = module.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateAchievementTracker()
	end)
	if not prdb.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
		self.disableInteraction = prdb.disableInteraction
		self.maxWidth = prdb.maxWidth
		self.statusMaxWidth = prdb.statusMaxWidth
		self.columnMaxWidth = self.maxWidth / prdb.numBooleanCriteriaColumns
		self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H * prdb.numBooleanCriteriaColumns
		self.minWidth = normalizeWidth(prdb.minWidth, self.maxWidth)
		self.statusMinWidth = normalizeWidth(prdb.statusMinWidth, self.statusMaxWidth)
		self.columnMinWidth = normalizeWidth(self.minWidth / prdb.numBooleanCriteriaColumns, self.columnMaxWidth)
		self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H * prdb.numBooleanCriteriaColumns, self.totalMaxWidth)
	else
		core.cbh:Fire("updateAchievementsTrackerParameters")
	end
	if prdb.displayAchievementImage then
		self.achievementImageSize = prdb.achievementImageSize
		self.achievementCellPadding = prdb.achievementCellPadding + self.achievementImageSize
	else
		self.achievementImageSize = 0
		self.achievementCellPadding = prdb.achievementCellPadding
	end
	self.categoryCellPadding = prdb.categoryCollapseButtonSize
	self.subCategoryCellPadding = prdb.subCategoryCellPadding + prdb.subCategoryCollapseButtonSize
	self.achievementIconPadding = prdb.achievementCellPadding
	self.criteriaCellPadding = prdb.criteriaCellPadding
	if prdb.showByCategory then
		self.subAchievementCellPadding = self.achievementCellPadding + self.subCategoryCellPadding
		self.subAchievementIconPadding = prdb.achievementCellPadding + self.subCategoryCellPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding + self.subCategoryCellPadding
	else
		self.subAchievementCellPadding = self.achievementCellPadding
		self.subAchievementIconPadding = self.achievementIconPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding
	end
	self:SecureHook("AddTrackedAchievement")
	self:SecureHook("RemoveTrackedAchievement")
	self:RawHook("GetNumTrackedAchievements", true)
	self:RawHook("IsTrackedAchievement", true)
	self:RegisterBucketEvent("CRITERIA_UPDATE", 1)
	self:RegisterEvent("ACHIEVEMENT_EARNED")
	core.RegisterCallback(self, "updateAchievementTracker")
	if prdb.timedAchievements then
		self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	end
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
	for aid in pairs(chdb.achievementTracked) do
		local _, _, _, completed = GetAchievementInfo(aid)
		if not completed then
			AddTrackedAchievement(aid, true)
		else
			chdb.achievementTracked[aid] = nil
		end
	end
	if AchievementFrame then
		AchievementFrameAchievements_ForceUpdate()
	end
	self:updateAchievementTracker()
end

function module:OnDisable()
	self.anchor:Hide()
	QTIP:Release(self.tracker)
	self.tracker = nil
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not core.moduleState["achievementTracker"] then return end
	self.cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	self.scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	self.aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
	self.acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	self.sbATrFontTQ2:SetFont(LSM:Fetch("font", prdb.statusBarFont), prdb.statusBarFontSize, prdb.statusBarFontOutline)
	self.fixATrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	self.anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if not prdb.showInQuestTracker or not core.moduleState["questsTracker"] then
		self.disableInteraction = prdb.disableInteraction
		self.maxWidth = prdb.maxWidth
		self.statusMaxWidth = prdb.statusMaxWidth
		self.columnMaxWidth = self.maxWidth / prdb.numBooleanCriteriaColumns
		self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H * prdb.numBooleanCriteriaColumns
		self.minWidth = normalizeWidth(prdb.minWidth, self.maxWidth)
		self.statusMinWidth = normalizeWidth(prdb.statusMinWidth, self.statusMaxWidth)
		self.columnMinWidth = normalizeWidth(self.minWidth / prdb.numBooleanCriteriaColumns, self.columnMaxWidth)
		self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H * prdb.numBooleanCriteriaColumns, self.totalMaxWidth)
	else
		core.cbh:Fire("updateAchievementsTrackerParameters")
	end
	if prdb.displayAchievementImage then
		self.achievementImageSize = prdb.achievementImageSize
		self.achievementCellPadding = prdb.achievementCellPadding + self.achievementImageSize
	else
		self.achievementImageSize = 0
		self.achievementCellPadding = prdb.achievementCellPadding
	end
	self.categoryCellPadding = prdb.categoryCollapseButtonSize
	self.subCategoryCellPadding = prdb.subCategoryCellPadding + prdb.subCategoryCollapseButtonSize
	self.achievementIconPadding = prdb.achievementCellPadding
	self.criteriaCellPadding = prdb.criteriaCellPadding
	if prdb.showByCategory then
		self.subAchievementCellPadding = self.achievementCellPadding + self.subCategoryCellPadding
		self.subAchievementIconPadding = prdb.achievementCellPadding + self.subCategoryCellPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding + self.subCategoryCellPadding
	else
		self.subAchievementCellPadding = self.achievementCellPadding
		self.subAchievementIconPadding = self.achievementIconPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding
	end
	if prdb.timedAchievements then
		self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	end
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
end

-----------------
--- Callbacks ---
-----------------
function module:inCombat(event)
	self.hideInCombat = true
	if self.tracker then self.tracker:Hide() end
end

function module:outOfCombat(event)
	self.hideInCombat = nil
	self:updateAchievementTracker()
end

function module:updateUsedMedia(event, mediaType, key)
	if not core.moduleState["achievementTracker"] then return end
	if mediaType == "font" then
		local updateTracker = nil
		if key == prdb.categoryFont then
			updateTracker = true
			self.cATrFontTQ2:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
		end
		if key == prdb.subCategoryFont then
			updateTracker = true
			self.scATrFontTQ2:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
		end
		if key == prdb.achievementFont then
			updateTracker = true
			self.aATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achievementFontOutline)
		end
		if key == prdb.achievementCriteriaFont then
			updateTracker = true
			self.acATrFontTQ2:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
		end
		if updateTracker then
			self:updateAchievementTracker()
		end
	end
end

--------------
--- Events ---
--------------
function module:ACHIEVEMENT_EARNED(event, aid)
	if self.achievements[aid] then
		RemoveTrackedAchievement(aid)
	end
	local subCategory = GetAchievementCategory(aid)
	local subCategoryName, category = GetCategoryInfo(subCategory)
	if category and category > 0 then
		if self.category[category] then
			self.category[category].numTA, self.category[category].numTCA  = GetCategoryNumAchievements(category)
			self.category[category].color = core:getColorFromCompletionHEX(self.category[category].numTCA / self.category[category].numTA)
		end
		if self.category[subCategory] then
			self.category[subCategory].numTA, self.category[subCategory].numTCA  = GetCategoryNumAchievements(subCategory)
			self.category[subCategory].color = core:getColorFromCompletionHEX(self.category[subCategory].numTCA / self.category[subCategory].numTA)
		end
	elseif self.category[subCategory] then
		self.category[subCategory].numTA, self.category[subCategory].numTCA  = GetCategoryNumAchievements(subCategory)
		self.category[subCategory].color = core:getColorFromCompletionHEX(self.category[subCategory].numTCA / self.category[subCategory].numTA)
	end
end

function module:CRITERIA_UPDATE()
	local updateTracker
	for aid in pairs(self.achievements) do
		for cid = 1, self.achievements[aid].numCriteria do
			local _, _, completed, quantity, reqQuantity = GetAchievementCriteriaInfo(aid, cid)
			completed = completed or 0
			if quantity ~= self.achievements[aid].criteria[cid].quantity or completed ~= self.achievements[aid].criteria[cid].completed then
				self.achievements[aid].criteria[cid].completed = completed
				self.achievements[aid].criteria[cid].quantity = quantity
				self.achievements[aid].criteria[cid].color = core:getColorFromCompletionHEX(quantity / self.achievements[aid].criteria[cid].reqQuantity)
				if self.visibleCriteria[self.achievements[aid].criteria[cid].id] then
					self:updateAchievementTracker()
				end
			end
		end
	end
	if updateTracker then
		self:updateAchievementTracker()
	end
end

function module:TRACKED_ACHIEVEMENT_UPDATE(event, aid, cid, elapsed, duration)
	if not aid or not duration or not elapsed then return end
	local startTime = GetTime()
	if duration > elapsed then
		if not self.achievements[aid] then
			AddTrackedAchievement(aid)
		end
		self.achievements[aid].startTime = startTime - elapsed
		self.achievements[aid].duration = duration
	elseif self.achievements[aid] and self.achievements[aid].startTime then
		self.achievements[aid].startTime = nil
		self.achievements[aid].duration = nil
	end
	if self.visibleAchievements[aid] then
		self:updateAchievementTracker()
	end
end

-------------
--- Hooks ---
-------------
function module:AddTrackedAchievement(aid, batch)
	if not batch and chdb.achievementTracked[aid] then
		RemoveTrackedAchievement(aid)
	else
		chdb.achievementTracked[aid] = true
		self.numTrackedAchievement = self.numTrackedAchievement + 1
		self.achievements[aid] = core.new()
		_, self.achievements[aid].name, _, self.achievements[aid].completed, _, _, _, self.achievements[aid].description, _, self.achievements[aid].image = GetAchievementInfo(aid)
		self.achievements[aid].numCriteria = GetAchievementNumCriteria(aid)
		self.achievements[aid].criteria = core.new()
		for cid = 1, self.achievements[aid].numCriteria do
			local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
			if reqQuantity == 0 then reqQuantity = 1 end
			self.achievements[aid].criteria[cid] = core.new()
			self.achievements[aid].criteria[cid].id = cid2
			self.achievements[aid].criteria[cid].name = cString
			self.achievements[aid].criteria[cid].cType = cType
			self.achievements[aid].criteria[cid].completed = completed or 0
			self.achievements[aid].criteria[cid].quantity = quantity
			self.achievements[aid].criteria[cid].reqQuantity = reqQuantity
			self.achievements[aid].criteria[cid].color = core:getColorFromCompletionHEX(quantity / reqQuantity)
			self.achievements[aid].criteria[cid].statusBar = band(flags, ACHIEVEMENT_CRITERIA_PROGRESS_BAR) == ACHIEVEMENT_CRITERIA_PROGRESS_BAR
		end
		local categoryName, subCategoryName
		self.achievements[aid].subCategory = GetAchievementCategory(aid)
		subCategoryName, self.achievements[aid].category = GetCategoryInfo(self.achievements[aid].subCategory)
		categoryName = GetCategoryInfo(self.achievements[aid].category)
		if self.achievements[aid].category and self.achievements[aid].category > 0 then
			if not self.category[self.achievements[aid].category] then
				self.category[self.achievements[aid].category] = core.new()
				self.category[self.achievements[aid].category].subCategory = core.new()
				self.category[self.achievements[aid].category].numSubCategory = 0
				self.category[self.achievements[aid].category].achievements = core.new()
				self.category[self.achievements[aid].category].numAchievements = 0
				self.category[self.achievements[aid].category].name = categoryName
				self.category[self.achievements[aid].category].numTA, self.category[self.achievements[aid].category].numTCA  = GetCategoryNumAchievements(self.achievements[aid].category)
				self.category[self.achievements[aid].category].color = core:getColorFromCompletionHEX(self.category[self.achievements[aid].category].numTCA / self.category[self.achievements[aid].category].numTA)
			end
			if not self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory] then
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory] = core.new()
				self.category[self.achievements[aid].category].numSubCategory = self.category[self.achievements[aid].category].numSubCategory + 1
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements = 0
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].achievements = core.new()
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].name = subCategoryName
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTA, self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(self.achievements[aid].subCategory)
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].color = core:getColorFromCompletionHEX(self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTCA / self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTA)
			end
			self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements + 1
			self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].achievements[aid] = true
			self.achievements[aid].sc = true
		else
			if not self.category[self.achievements[aid].subCategory] then
				self.category[self.achievements[aid].subCategory] = core.new()
				self.category[self.achievements[aid].subCategory].subCategory = core.new()
				self.category[self.achievements[aid].subCategory].numSubCategory = 0
				self.category[self.achievements[aid].subCategory].achievements = core.new()
				self.category[self.achievements[aid].subCategory].numAchievements = 0
				self.category[self.achievements[aid].subCategory].name = subCategoryName
				self.category[self.achievements[aid].subCategory].numTA, self.category[self.achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(self.achievements[aid].subCategory)
				self.category[self.achievements[aid].subCategory].color = core:getColorFromCompletionHEX(self.category[self.achievements[aid].subCategory].numTCA / self.category[self.achievements[aid].subCategory].numTA)
			end
			self.category[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].subCategory].numAchievements + 1
			self.category[self.achievements[aid].subCategory].achievements[aid] = true
		end
		if not batch then
			if AchievementFrame then
				AchievementFrameAchievements_ForceUpdate()
			end
			self:updateAchievementTracker()
		end
	end
end

function module:RemoveTrackedAchievement(aid)
	if not chdb.achievementTracked[aid] or not self.achievements[aid] then return end
	self.numTrackedAchievement = self.numTrackedAchievement - 1
	if self.achievements[aid].startTime then
		self.achievements[aid].startTime = nil
		self.achievements[aid].duration = nil
	end
	chdb.achievementTracked[aid] = nil
	if self.achievements[aid].category and self.achievements[aid].category > 0 and self.achievements[aid].subCategory and self.achievements[aid].subCategory > 0 then
		self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].achievements[aid] = nil
		self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements - 1
		if self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements == 0 then
			self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory] = core.del(self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory])
			self.category[self.achievements[aid].category].numSubCategory = self.category[self.achievements[aid].category].numSubCategory - 1
			if self.category[self.achievements[aid].category].numSubCategory == 0 and self.category[self.achievements[aid].category].numAchievements == 0 then
				self.category[self.achievements[aid].category] = core.del(self.category[self.achievements[aid].category])
			end
		end
	elseif self.achievements[aid].subCategory and self.achievements[aid].subCategory > 0 then
		self.category[self.achievements[aid].subCategory].achievements[aid] = nil
		self.category[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].subCategory].numAchievements - 1
		if self.category[self.achievements[aid].subCategory].numSubCategory == 0 and self.category[self.achievements[aid].subCategory].numAchievements == 0 then
			self.category[self.achievements[aid].subCategory] = core.del(self.category[self.achievements[aid].subCategory])
		end
	end
	self.achievements[aid] = core.del(self.achievements[aid])
	self:updateAchievementTracker()
	if AchievementFrame then
		AchievementFrameAchievements_ForceUpdate()
	end
end

function module:GetNumTrackedAchievements()
	return self.numTrackedAchievement
end

function module:IsTrackedAchievement(aid)
	return self.achievements[aid] and true or false
end

---------------
--- Tracker ---
---------------
local interactivity = {
	chatLink = function(cValue1)
		ChatFrameEditBox:Show()
		ChatEdit_InsertLink(GetAchievementLink(cValue1))
	end,
	achievementTooltip = function(cValue1)
		ItemRefTooltip:ClearLines()
		ShowUIPanel(ItemRefTooltip)
		if not ItemRefTooltip:IsVisible() then
			ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
		end
		local link = GetAchievementLink(cValue1)
		if link then
			ItemRefTooltip:SetHyperlink(link)
		end
	end,
	removeTrackedAchievement = function(cValue1)
		RemoveTrackedAchievement(cValue1)
	end,
	displayAchievement = function(cValue1)
		ToggleAchievementFrame()
		if AchievementFrame:IsShown() then
			AchievementFrame_SelectAchievement(cValue1)
		end
	end,
	nothing = function(cValue1)
		return
	end,
}

local function achievementOnMouseDown(cell, button)
	local cType1, cValue1 = cell.cType1, cell.cValue1
	if cType1 == "category" then
		if chdb.categoryDisplay[cValue1] == 1 then
			chdb.categoryDisplay[cValue1] = 0
		else
			chdb.categoryDisplay[cValue1] = 1
		end
		module:updateAchievementTracker()
	elseif cType1 == "achievement" then
		if button == "RightButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseRightShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseRightControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseRightAlt](cValue1)
			else
				interactivity[prdb.mouseRight](cValue1)
			end
		elseif button == "LeftButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseLeftShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseLeftControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseLeftAlt](cValue1)
			else
				interactivity[prdb.mouseLeft](cValue1)
			end
		elseif button == "MiddleButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseMiddleShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseMiddleControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseMiddleAlt](cValue1)
			else
				interactivity[prdb.mouseMiddle](cValue1)
			end
		end
	end
end

local function achievementOnEnter(cell)
	if prdb.overTooltip and cell.cType1 == "achievement" then
		GameTooltip:SetOwner(module.tracker, module.trackerAnchor)
		GameTooltip:SetHyperlink(GetAchievementLink(cell.cValue1))
	end
end

local function achievementOnLeave(cell)
	GameTooltip:Hide()
end

local function buttonOnClick(button)
	local category = button.category
	if not category then return end
	if chdb.categoryDisplay[category] == 1 then
		chdb.categoryDisplay[category] = 0
	else
		chdb.categoryDisplay[category] = 1
	end
	module:updateAchievementTracker()
end

local function timedAchievementUpdate(self, elapsed)
	self.elapsed = self.elapsed + elapsed
	if self.elapsed >= 1 then
		local aid = self.cValue1
		module.achievements[aid].remainingTime = module.achievements[aid].remainingTime - self.elapsed
		if prdb.colorizeATimer then
			self.fontString:SetText("|cff"..prdb.defaultANameColor..self.achievements[aid].name.."|r |cff"..core:getColorFromCompletionHEX(self.achievements[aid].remaining / self.achievements[aid].duration)..SecondsToTime(self.achievements[aid].remaining).."|r")
		else
			self.fontString:SetText("|cff"..prdb.defaultANameColor..self.achievements[aid].name.."|r |cff"..prdb.defaultATimerColor..SecondsToTime(self.achievements[aid].remaining).."|r")
		end
		self.elapsed = 0
	end
end

local cProvider, cPrototype, basecCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cType1, cType2, cValue1, cValue2)
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	elseif self.statusBar then
		core:recycleStatusBar(self.statusBar)
		self.statusBar = nil
	end
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	if not module.disableInteraction then
		self:EnableMouse(true)
		self.cType1, self.cValue1 = cType1, cValue1
		self:SetScript("OnEnter", achievementOnEnter)
		self:SetScript("OnLeave", achievementOnLeave)
		self:SetScript("OnMouseDown", achievementOnMouseDown)
		if cType2 == "icon" and cType1 == "category" then
			self.button = core:getCollapseButton()
			self.button:RegisterForClicks("AnyDown")
			self.button.iconTexture:SetTexture(cValue2)
			self.button:SetScript("OnClick", buttonOnClick)
			self.button.category = cValue1
			if prdb.categoryCollapseButtonSize > height then
				self.button:SetHeight(height)
				self.button:SetWidth(height)
			else
				self.button:SetHeight(prdb.categoryCollapseButtonSize)
				self.button:SetWidth(prdb.categoryCollapseButtonSize)
			end
			self.button:SetParent(self)
			if module.category[cValue1] then
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT")
			else
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", prdb.subCategoryCellPadding, 0)
			end
			self.button:SetToplevel(true)
			self.button:Show()
		end
	else
		self:EnableMouse(false)
		self.cType1, self.cValue1 = nil, nil
		self:SetScript("OnEnter", nil)
		self:SetScript("OnLeave", nil)
		self:SetScript("OnMouseDown", nil)
	end
	if cType2 == "icon" and cType1 == "achievement" then
		if prdb.displayAchievementImage then
			self.button = core:getCollapseButton()
			self.button.iconTexture:SetTexture(cValue2)
			if prdb.achievementImageSize > height then
				self.button:SetHeight(height)
				self.button:SetWidth(height)
			else
				self.button:SetHeight(prdb.achievementImageSize)
				self.button:SetWidth(prdb.achievementImageSize)
			end
			self.button:SetParent(self)
			if module.achievements[cValue1].sc then
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.subAchievementIconPadding, 0)
			else
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.achievementIconPadding, 0)
			end
			self.button:SetToplevel(true)
			self.button:Show()
		end
		if module.achievements[cValue1].remainingTime and module.achievements[cValue1].remainingTime > 0 then
			self:SetScript("OnUpdate", timedAchievementUpdate)
			self.elapsed = 0
		else
			self:SetScript("OnUpdate", nil)
		end
	elseif cType2 == "statusBar" then
		local sbHeight = prdb.statusBarHeight + 2
		if sbHeight > height then
			height = sbHeight
			self:SetHeight(height)
		end
		self.statusBar = core:getStatusBar()
		self.statusBar:SetParent(self)
		self.statusBar:SetPoint("TOPLEFT", prdb.leftStatusBarPadding, -1)
		self.statusBar:SetPoint("BOTTOMRIGHT", -prdb.rightStatusBarPadding, 1)
		self.statusBar:SetMinMaxValues(0, module.achievements[cValue1].criteria[cValue2].reqQuantity)
		self.statusBar:SetValue(module.achievements[cValue1].criteria[cValue2].quantity)
		self.statusBar:SetStatusBarTexture(LSM:Fetch("statusbar", prdb.statusBarTexture))
		self.statusBar.bg:SetTexture(LSM:Fetch("statusbar", prdb.statusBarTexture))
		local r, g, b
		if prdb.colorizeStatusBar then
			r, g, b = core:getColorFromCompletionRGB(module.achievements[cValue1].criteria[cValue2].quantity / module.achievements[cValue1].criteria[cValue2].reqQuantity) 
		else
			r, g, b = unpack(prdb.defaultStatusBarColor)
		end
		self.statusBar.bg:SetVertexColor(r*0.33, g*0.33, b*0.33)
		self.statusBar:SetStatusBarColor(r, g, b)
		self.statusBar.status:SetFontObject(module.sbATrFontTQ2)
		if module.achievements[cValue1].criteria[cValue2].cType == 62 or module.achievements[cValue1].criteria[cValue2].cType == 67 then
			self.statusBar.status:SetText(core:formatMoney(module.achievements[cValue1].criteria[cValue2].quantity))
		elseif prdb.colorizeStatusBarText then
			self.statusBar.status:SetText("|cff"..core:getColorFromCompletionHEX(module.achievements[cValue1].criteria[cValue2].quantity / module.achievements[cValue1].criteria[cValue2].reqQuantity)..module.achievements[cValue1].criteria[cValue2].quantity.." / "..module.achievements[cValue1].criteria[cValue2].reqQuantity.."|r")
		else
			self.statusBar.status:SetText("|cff"..prdb.defaultStatusBarTextColor..module.achievements[cValue1].criteria[cValue2].quantity.." / "..module.achievements[cValue1].criteria[cValue2].reqQuantity.."|r")
		end
		self.statusBar:Show()
	end
	return width, height
end

function cPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	elseif self.statusBar then
		core:recycleStatusBar(self.statusBar)
		self.statusBar = nil
	end
	self:EnableMouse(false)
	self.cType1, self.cValue1, self.elapsed = nil, nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
	self:SetScript("OnUpdate", nil)
end

function module:addAchievementToTracker(aid, achievementPadding, criteriaPadding)
	local tempLine = self.tracker:AddLine()
	self.visibleAchievements[aid] = tempLine
	if self.achievements[aid].startTime then
		self.achievements[aid].remainingTime = self.achievements[aid].duration + self.achievements[aid].startTime - GetTime()
		if prdb.colorizeATimer then
			self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultANameColor..self.achievements[aid].name.."|r |cff"..core:getColorFromCompletionHEX(self.achievements[aid].remaining / self.achievements[aid].duration)..SecondsToTime(self.achievements[aid].remaining).."|r", self.aATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
		else
			self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultANameColor..self.achievements[aid].name.."|r |cff"..prdb.defaultATimerColor..SecondsToTime(self.achievements[aid].remaining).."|r", self.aATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
		end
	else
		self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultANameColor..self.achievements[aid].name.."|r", self.aATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
	end
	if self.achievements[aid].numCriteria == 0 then
		tempLine = self.tracker:AddLine()
		self.tracker:SetCell(tempLine, 1, self.achievements[aid].description, self.acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns + 1, cProvider, criteriaPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", nil, aid)
	elseif self.achievements[aid].numCriteria == 1 then
		if self.achievements[aid].criteria[1].statusBar and prdb.displayAsStatusBar then
			if prdb.showSBDescription then
				tempLine = self.tracker:AddLine()
				self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultDescriptionColor..self.achievements[aid].description.."|r", self.acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns + 1, cProvider, criteriaPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", nil, aid)
			end
			tempLine = self.tracker:AddLine()
			self.tracker:SetCell(tempLine, 1, " ", self.acATrFontTQ2, "CENTER", prdb.numBooleanCriteriaColumns + 1, cProvider, 0, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "achievement", "statusBar", aid, 1)
		else
			tempLine = self.tracker:AddLine()
			if prdb.colorizeACrName then
				self.tracker:SetCell(tempLine, 1, "|cff"..self.achievements[aid].criteria[1].color..self.achievements[aid].description.."|r", self.acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
			else
				self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultACrNameColor..self.achievements[aid].description.."|r", self.acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
			end
			if prdb.colorizeACrStatus then
				self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..self.achievements[aid].criteria[1].color..self.achievements[aid].criteria[1].quantity.." / "..self.achievements[aid].criteria[1].reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
			else
				self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..prdb.defaultACrStatusColor..self.achievements[aid].criteria[1].quantity.." / "..self.achievements[aid].criteria[1].reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
			end
		end
		self.visibleCriteria[self.achievements[aid].criteria[1].id] = tempLine
	elseif self.achievements[aid].numCriteria > 1 then
		local cIndex = 1
		for cid, acData in pairs (self.achievements[aid].criteria) do
			if (prdb.hideCompletedCriteria and acData.completed == 0) or (not prdb.hideCompletedCriteria) then
				if acData.reqQuantity > 1 then
					if (acData.cType == 7 and acData.quantity > 0) or acData.cType ~= 7 then
						tempLine = self.tracker:AddLine()
						if prdb.colorizeACrName then
							self.tracker:SetCell(tempLine, 1, "|cff"..acData.color..acData.name.."|r", self.acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultACrNameColor..acData.name.."|r", self.acATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
						end
						if prdb.colorizeACrStatus then
							self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..acData.color..acData.quantity.." / "..acData.reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..prdb.defaultACrStatusColor..acData.quantity.." / "..acData.reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
						end
					end
					self.visibleCriteria[self.achievements[aid].criteria[cid].id] = tempLine.."~"..criteriaPadding
				else
					if cIndex == 1 then
						tempLine = self.tracker:AddLine()
						self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, " ", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
						if prdb.colorizeACrName then
							self.tracker:SetCell(tempLine, cIndex, "|cff"..acData.color..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, criteriaPadding, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, cIndex, "|cff"..prdb.defaultACrNameColor..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, criteriaPadding, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						end
					else
						if prdb.colorizeACrName then
							self.tracker:SetCell(tempLine, cIndex, "|cff"..acData.color..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, 0, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, cIndex, "|cff"..prdb.defaultACrNameColor..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, 0, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						end
					end
					self.visibleCriteria[self.achievements[aid].criteria[cid].id] = tempLine.."~"..criteriaPadding.."~"..cIndex
					if cIndex == prdb.numBooleanCriteriaColumns then cIndex = 1 else cIndex = cIndex + 1 end
				end
			end
		end
	end
end

function module:updateTimedAchievement()
	for _, timedAchievement in ipairs(self.timerPosition) do
		local tempLine, aid, achievementPadding = timedAchievement:match("^(%d-)~(%d-)~(%d-)$")
		if not aid or not tempLine or not achievementPadding then return end
		tempLine, aid, achievementPadding = tonumber(tempLine), tonumber(aid), tonumber(achievementPadding)
		local remaining = self.achievements[aid].duration + self.achievements[aid].startTime - GetTime()
		if prdb.colorizeATimer then
			self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultANameColor..self.achievements[aid].name.."|r |cff"..core:getColorFromCompletionHEX(remaining / self.achievements[aid].duration)..SecondsToTime(remaining).."|r", self.aATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
		else
			self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultANameColor..self.achievements[aid].name.."|r |cff"..prdb.defaultATimerColor..SecondsToTime(remaining).."|r", self.aATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
		end
	end
end

function module:updateAchievements()
	wipe(self.visibleAchievements)
	wipe(self.visibleCriteria)
	local tempLine
	if prdb.showByCategory then
		for cid, cData in pairs(self.category) do
			chdb.categoryDisplay[cid] = chdb.categoryDisplay[cid] or 1
			tempLine = self.tracker:AddLine()
			if prdb.colorizeACName then
				self.tracker:SetCell(tempLine, 1, "|cff"..cData.color..cData.name.."|r", self.cATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, self.categoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", cid, chdb.categoryDisplay[cid] == 0 and self.plus or self.minus)
			else
				self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultACrNameColor..cData.name.."|r", self.cATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, self.categoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", cid, chdb.categoryDisplay[cid] == 0 and self.plus or self.minus)
			end
			if prdb.colorizeACCounter then
				self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..cData.color..cData.numTCA.." / "..cData.numTA.."|r", self.cATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, cid)
			else
				self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..prdb.defaultACCounterColor..cData.numTCA.." / "..cData.numTA.."|r", self.cATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, cid)
			end
			if chdb.categoryDisplay[cid] == 1 or prdb.disableInteraction then
				if cData.numAchievements > 0 then
					for aid in pairs(cData.achievements) do
						self:addAchievementToTracker(aid, self.achievementCellPadding, self.criteriaCellPadding)
					end
				end
				if cData.numSubCategory > 0 then
					for scid, scData in pairs(cData.subCategory) do
						chdb.categoryDisplay[scid] = chdb.categoryDisplay[scid] or 1
						tempLine = self.tracker:AddLine()
						if prdb.colorizeASCName then
							self.tracker:SetCell(tempLine, 1, "|cff"..scData.color..scData.name.."|r", self.scATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, self.subCategoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", scid, chdb.categoryDisplay[scid] == 0 and self.plus or self.minus)
						else
							self.tracker:SetCell(tempLine, 1, "|cff"..prdb.defaultASCNameColor..scData.name.."|r", self.scATrFontTQ2, "LEFT", prdb.numBooleanCriteriaColumns, cProvider, self.subCategoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", scid, chdb.categoryDisplay[scid] == 0 and self.plus or self.minus)
						end
						if prdb.colorizeASCCounter then
							self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..scData.color..scData.numTCA.." / "..scData.numTA.."|r", self.scATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, scid)
						else
							self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, "|cff"..prdb.defaultASCCounterColor..scData.numTCA.." / "..scData.numTA.."|r", self.scATrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, scid)
						end
						if chdb.categoryDisplay[scid] == 1 or prdb.disableInteraction then
							if scData.numAchievements > 0 then
								for aid in pairs(scData.achievements) do
									self:addAchievementToTracker(aid, self.subAchievementCellPadding, self.subCriteriaCellPadding)
								end
							end
						end
					end
				end
			end
		end
	else
		for aid, aData in pairs(self.achievements) do
			self:addAchievementToTracker(aid, self.achievementCellPadding, self.criteriaCellPadding)
		end
	end
end

function module:updateAchievementTracker()
	if prdb.showInQuestTracker and core.moduleState["questsTracker"] then
		core.cbh:Fire("updateQuestsTracker")
	elseif prdb.toggleTracker and self.numTrackedAchievement > 0 and not self.hideInCombat then
		self.tracker = QTIP:Acquire("tomQuest2Achievements", prdb.numBooleanCriteriaColumns + 1)
		self.tracker:Clear()
		self.tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
		self.tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		self.tracker:SetScale(prdb.scale)
		self.tracker:SetFrameStrata("BACKGROUND")
		self:updateAchievements()
		local tempLine = self.tracker:AddLine()
		for i = 1, prdb.numBooleanCriteriaColumns do
			self.tracker:SetCell(tempLine, i, " ", self.fixATrFontTQ2, "LEFT", 1, nil, 0, 0, self.columnMaxWidth, self.columnMinWidth)
		end
		self.tracker:SetCell(tempLine, prdb.numBooleanCriteriaColumns + 1, " ", self.fixATrFontTQ2, "RIGHT", 1, nil, 0, 0, self.tatusMaxWidth, self.statusMinWidth)
		self.tracker:ClearAllPoints()
		if prdb.showAnchor then
			self.tracker:SetPoint(self.anchorSetPoint[prdb.anchorPosition].point, self.anchor, self.anchorSetPoint[prdb.anchorPosition].relPoint)
		else
			self.tracker:SetPoint(self.anchorSetPoint[prdb.anchorPosition].point, self.anchor, self.anchorSetPoint[prdb.anchorPosition].relPoint, self.anchorSetPoint[prdb.anchorPosition].x, self.anchorSetPoint[prdb.anchorPosition].y)
		end
		if not prdb.disableInteraction then
			self.tracker:UpdateScrolling(prdb.hasMaxHeight and prdb.maxHeight or nil)
			self.tracker:EnableMouse(true)
		else
			self.tracker:EnableMouse(false)
		end
		self.tracker:Show()
	else
		QTIP:Release(self.tracker)
		self.tracker = nil
		return
	end
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2ACHIEVEMENTTRACKERTOGGLE = LATr["Achievement Tracker"]

function tQ2toggleAchievementTracker(display)
	if not display then
		prdb.toggleTracker = not prdb.toggleTracker
		if not prdb.toggleTracker then
			QTIP:Release(module.tracker)
			module.tracker = nil
		else
			module:updateAchievementTracker()
		end
	elseif display == 1 and not prdb.toggleTracker then
		prdb.toggleTracker = not prdb.toggleTracker
		module:updateAchievementTracker()
	elseif display == 0 and prdb.toggleTracker then
		prdb.toggleTracker = not prdb.toggleTracker
		QTIP:Release(module.tracker)
		module.tracker = nil
	end
end