-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("announce", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LA = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Announce")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local format, SendChatMessage, pairs = string.format, SendChatMessage, pairs

-----------------
--- variables ---
-----------------
local prdb, options
module.modName = LA["Quests Announcement"]

module.callBacks = {
	["playerQuestComplete"] = true,
	["playerQuestFailed"] = true,
}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		playerObjectiveProgress = true,
		playerObjectiveComplete = true,
		playerQuestComplete = true,
		playerQuestFailed = true,
		objectiveNotificationWhenComplete = false,
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["announce"] then return end
		if module.callBacks[info[#info]] then
			if value then
				core.RegisterCallback(module, info[#info])
			else
				core.UnregisterCallback(module, info[#info])
			end
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				playerObjectiveProgress = {
					name = LA["Quests objectives progress annoucement"],
					type = "toggle",
					width = "full",
					order = 12,
				},
				playerObjectiveComplete = {
					name = LA["Quests objectives complete announcement"],
					type = "toggle",
					width = "full",
					order = 13,
				},
				playerQuestComplete = {
					name = LA["Quests complete announcement"],
					type = "toggle",
					width = "full",
					order = 14,
				},
				playerQuestFailed = {
					name = LA["Quests failed announcement"],
					type = "toggle",
					width = "full",
					order = 15,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("announce", defaults, "Default")
	prdb = self.db.profile
	self:SetEnabledState(core:getModuleEnabled("announce"))
	core.moduleState["announce"] = core:getModuleEnabled("announce")
	core:registerModuleOptions("announce", getOptions, module.modName)
end

function module:OnEnable()
	core.RegisterCallback(self, "playerJoinedParty")
	core.RegisterCallback(self, "playerLeftParty")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	if not core.moduleState["announce"] then return end
	core.UnregisterAllCallbacks(self)
	core.RegisterCallback(self, "playerJoinedParty")
	core.RegisterCallback(self, "playerLeftParty")
	if core.numPartyMembers == 0 then return end
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	for callback in pairs(self.callBacks) do
		if prdb[callback] then
			core.RegisterCallback(self, callback)
		end
	end
end

function module:playerJoinedParty()
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	for callback in pairs(self.callBacks) do
		if prdb[callback] then
			core.RegisterCallback(self, callback)
		end
	end
end

function module:playerLeftParty()
	core.UnregisterCallback(self, "playerQuestObjectiveProgress")
	for callback in pairs(self.callBacks) do
		if prdb[callback] then
			core.UnregisterCallback(self, callback)
		end
	end
end

---------------------------
--- Main Module Program ---
---------------------------
function module:playerQuestObjectiveProgress(event, commHandler, uid, objective, got, nGot)
	if nGot == core.questsObjectives[uid][objective].nNeed and prdb.playerObjectiveComplete then
		if not prdb.objectiveNotificationWhenComplete and core.quests[uid].complete[core.playerName] ~= 0 then return end
		SendChatMessage(format(LA["(%s) %s : finished"], core.quests[uid].title, core.questsObjectives[uid][objective].desc), "PARTY")
	elseif prdb.playerObjectiveProgress and nGot > 0 then
		SendChatMessage("("..core.quests[uid].title..") "..core.questsObjectives[uid][objective].desc.." : "..got.."/"..core.questsObjectives[uid][objective].need, "PARTY")
	end
end

function module:playerQuestComplete(event, commHandler, uid)
	SendChatMessage(format(LA["(%s) : finished"], core.quests[uid].title), "PARTY")
end

function module:playerQuestFailed(event, commHandler, uid)
	SendChatMessage(format(LA["(%s) : failed"], core.quests[uid].title), "PARTY")
end