-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("notifications", "AceEvent-3.0", "AceHook-3.0", "LibSink-2.0")

-----------------------
--- Library loading ---
-----------------------
local LN = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Notifications")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local format, gsub, pairs = string.format, string.gsub, pairs
local origiOnEvent = UIErrorsFrame:GetScript("OnEvent")

-----------------
--- variables ---
-----------------
local prdb, options, output

module.modName = LN["Quests Notifications"]
module.callBacks = {
	["playerQuestComplete"] = true,
	["playerQuestFailed"] = true,
	["partyQuestComplete"] = true,
	["partyQuestFailed"] = true,
	["partyQuestAccepted"] = true,
	["itemStartingQuestLooted"] = true
}
local function paternize(pattern)
	pattern = pattern:gsub('[()]', '%%%1')
	return '^'..pattern:gsub('(%%.)', '(.+)')..'$'
end
local found = paternize(ERR_QUEST_ADD_FOUND_SII)
local item = paternize(ERR_QUEST_ADD_ITEM_SII)
local kill = paternize(ERR_QUEST_ADD_KILL_SII)
local playerKilled = paternize(ERR_QUEST_ADD_PLAYER_KILL_SII)
local oComplete = paternize(ERR_QUEST_OBJECTIVE_COMPLETE_S)
local qComplete = paternize(ERR_QUEST_COMPLETE_S)
local qFailed = paternize(ERR_QUEST_FAILED_S)

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		playerObjectiveProgress = true,
		playerObjectiveComplete = true,
		playerQuestComplete = true,
		playerQuestFailed = true,
		partyObjectiveProgress = true,
		partyObjectiveComplete = true,
		partyQuestComplete = true,
		partyQuestFailed = true,
		partyQuestAccepted = true,
		partyQuestTurnedIn = true,
		partyQuestAbandonned = true,
		hideBlizzardNotification = true,
		itemStartingQuestLooted = true,
		objectiveNotificationWhenComplete = false,
		sinkOptions = {
		},
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["notifications"] then return end
		if module.callBacks[info[#info]] then
			if value then
				core.RegisterCallback(module, info[#info])
			else
				core.UnregisterCallback(module, info[#info])
			end
		elseif info[#info] == "hideBlizzardNotification" then
			if value then
				UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
					if event == "UI_INFO_MESSAGE" and (message:match(found) or message:match(item) or message:match(kill) or message:match(playerKilled) or message:match(oComplete) or message:match(qComplete) or message:match(qFailed)) then return end
					return origiOnEvent(self, event, message, ...)
				end)
			else
				UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
					return origiOnEvent(self, event, message, ...)
				end)
			end
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				common = {
					type = "group",
					name = LN["Common Settings"],
					args = {
						hideBlizzardNotification = {
							name = LN["Hide default blizzard quests notification"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						objectiveNotificationWhenComplete = {
							name = LN["Objective notification on quest completion"],
							type = "toggle",
							width = "full",
							order = 2,
						},
					},
				},
				player = {
					type = "group",
					name = LN["Player Notifications"],
					args = {
						playerObjectiveProgress = {
							name = LN["Quest objective progress notification"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						playerObjectiveComplete = {
							name = LN["Quest objective complete notification"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						playerQuestComplete = {
							name = LN["Quest complete notification"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						playerQuestFailed = {
							name = LN["Quest failed notification"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						itemStartingQuestLooted = {
							name = LN["Item starting a quest notification"],
							type = "toggle",
							width = "full",
							order = 5,
						},
					},
				},
				party = {
					type = "group",
					name = LN["Party Notifications"],
					args = {
						partyObjectiveProgress = {
							name = LN["Quest objective progress notification"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						partyObjectiveComplete = {
							name = LN["Quest objective complete notification"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						partyQuestComplete = {
							name = LN["Quest complete notification"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						partyQuestFailed = {
							name = LN["Quest failed notification"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						partyQuestAccepted = {
							name = LN["Quest accepted notification"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						partyQuestTurnedIn = {
							name = LN["Quest turned in notification"],
							type = "toggle",
							width = "full",
							order = 6,
						},
						partyQuestAbandonned = {
							name = LN["Quest abandonned notification"],
							type = "toggle",
							width = "full",
							order = 7,
						},
					},
				},
				output = module:GetSinkAce3OptionsDataTable()
			},
		}
	end
	
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("notifications", defaults, "Default")
	prdb = self.db.profile
	self:SetSinkStorage(prdb.sinkOptions)
	self:SetEnabledState(core:getModuleEnabled("notifications"))
	core.moduleState["notifications"] = core:getModuleEnabled("notifications")
	core:registerModuleOptions("notifications", getOptions, module.modName)
end

function module:OnEnable()
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestLost")
	for callback in pairs(self.callBacks) do
		if prdb[callback] then
			core.RegisterCallback(self, callback)
		end
	end
	if prdb.hideBlizzardNotification then
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			if event == "UI_INFO_MESSAGE" and (message:match(found) or message:match(item) or message:match(kill) or message:match(playerKilled) or message:match(oComplete) or message:match(qComplete) or message:match(qFailed)) then return end
			return origiOnEvent(self, event, message, ...)
		end)
	end
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
		return origiOnEvent(self, event, message, ...)
	end)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	if not core.moduleState["notifications"] then return end
	core.UnregisterAllCallbacks(self)
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestLost")
	for callback in pairs(self.callBacks) do
		if prdb[callback] then
			core.RegisterCallback(self, callback)
		end
	end
	if prdb.hideBlizzardNotification then
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			if event == "UI_INFO_MESSAGE" and (message:match(found) or message:match(item) or message:match(kill) or message:match(playerKilled) or message:match(oComplete) or message:match(qComplete) or message:match(qFailed)) then return end
			return origiOnEvent(self, event, message, ...)
		end)
	else
		UIErrorsFrame:SetScript("OnEvent", function(self, event, message, ...)
			return origiOnEvent(self, event, message, ...)
		end)
	end
end

---------------------------
--- Main Module Program ---
---------------------------
function module:playerQuestObjectiveProgress(event, commHandler, uid, objective, got, nGot)
	if nGot == core.questsObjectives[uid][objective].nNeed and prdb.playerObjectiveComplete then
		if not prdb.objectiveNotificationWhenComplete and core.quests[uid].complete[core.playerName] ~= 0 then return end
		self:Pour(format(LN["(%s) %s : finished"], core.quests[uid].title, core.questsObjectives[uid][objective].desc), 0, 1, 0)
	elseif prdb.playerObjectiveProgress and nGot > 0 then
		self:Pour("("..core.quests[uid].title..") "..core.questsObjectives[uid][objective].desc.." : "..got.."/"..core.questsObjectives[uid][objective].need, core:getColorFromCompletionRGB(nGot/core.questsObjectives[uid][objective].nNeed))
	end
end

function module:playerQuestComplete(event, commHandler, uid)
	self:Pour(format(LN["(%s) : finished"], core.quests[uid].title), 0, 1, 0)
end

function module:playerQuestFailed(event, commHandler, uid)
	self:Pour(format(LN["(%s) : failed"], core.quests[uid].title), 1, 0, 0)
end

function module:partyQuestObjectiveProgress(event, commHandler, uid, objective, got, nGot, sender)
	if nGot == core.questsObjectives[uid][objective].nNeed and prdb.partyObjectiveComplete then
		if not prdb.objectiveNotificationWhenComplete and core.quests[uid].complete[sender] ~= 0 then return end
		self:Pour(format(LN["(%s : %s) %s : finished"], sender, core.quests[uid].title, core.questsObjectives[uid][objective].desc), 0, 1, 0)
	elseif prdb.partyObjectiveProgress and nGot > 0 then
		self:Pour("("..sender.." : "..core.quests[uid].title..") "..core.questsObjectives[uid][objective].desc.." : "..got.."/"..core.questsObjectives[uid][objective].need, core:getColorFromCompletionRGB(nGot/core.questsObjectives[uid][objective].nNeed))
	end
end

function module:partyQuestComplete(event, commHandler, uid, sender)
	self:Pour(format(LN["(%s : %s) : finished"], sender, core.quests[uid].title), 0, 1, 0)
end

function module:partyQuestFailed(event, commHandler, uid, sender)
	self:Pour(format(LN["(%s : %s) : failed"], sender, core.quests[uid].title), 1, 0, 0)
end

function module:partyQuestLost(event, commHandler, uid, abandoning, title, sender)
	if abandoning and prdb.partyQuestAbandonned then
		self:Pour(format(LN["(%s : %s) : abandonned"], sender, title), 1, 0, 0)
	elseif not abandoning and prdb.partyQuestTurnedIn then
		self:Pour(format(LN["(%s : %s) : turned in"], sender, title), 0, 1, 0)
	end
end

function module:partyQuestAccepted(event, sender, uid, started, batch)
	if not batch then
		self:Pour(format(LN["(%s : %s) : accepted"], sender, core.quests[uid].title), 0, 1, 0)
	end
end

function module:itemStartingQuestLooted(event, itemLink)
	self:Pour(format(LN["You've just looted %s which start a quest"], itemLink), 0, 1, 0)
end