-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("informations", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LI = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsInformations")
local LSM = LibStub("LibSharedMedia-3.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local format, gsub, type, abs, tonumber = format, gsub, type, math.abs, tonumber
local unpack, tremove, pairs, tconcat = unpack, tremove, pairs, table.concat
local CreateFrame, CreateFont = CreateFrame, CreateFont
local GetQuestLogSelection, SelectQuestLogEntry, QuestLogPushQuest = GetQuestLogSelection, SelectQuestLogEntry, QuestLogPushQuest
local SetAbandonQuest, GetAbandonQuestName, GetAbandonQuestItems = SetAbandonQuest, GetAbandonQuestName, GetAbandonQuestItems
local UIParent, IsShiprdftKeyDown, IsControlKeyDown, IsAltKeyDown = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local GetItemInfo, DressUpItemLink, SetItemRef = GetItemInfo, DressUpItemLink, SetItemRef

-----------------
--- variables ---
-----------------
local prdb, options, _

module.modName = LI["Quests Informations"]
module.stack, module.temp = {}, {}
module.attachSelect = {[1] = LI["Right"], [2] = LI["Left"], [3] = LI["Dettached"]}
module.repl = {["<"] = "&lt;", [">"] = "&gt;"}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showQuestLog = true,
		showLightHeaded = true,
		backDropColor = {0.09, 0.09, 0.19, 0.5},
		borderColor = {0.5, 0.5, 0.5, 1},
		attach = 1,
		showAllCommentsOnOnePage = false,
		qlFont = LSM.DefaultMedia.font,
		qlFontSizeH1 = 14,
		qlFontSizeH2 = 12,
		qlFontSize = 10,
		lhFont = LSM.DefaultMedia.font,
		lhFontSize = 10,
		tomQuest2ParentFrame = {
			width = 640,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		tomQuest2QlFrame = {
			width = 320,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		tomQuest2LhFrame = {
			width = 320,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		color = "ffffd100",
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter, colorGetter, colorSetter, frameSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["informations"] then return end
		local uid
		if tomQuest2QlFrame then
			uid = tomQuest2QlFrame.currentUid
			tomQuest2QlFrame.currentUid = nil
		end
		if tomQuest2LhFrame then
			uid = tomQuest2LhFrame.currentUid or uid
			tomQuest2LhFrame.currentUid = nil
		end
		if uid then
			module:displayQuestInformations(nil, uid)
		end
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
		if not core.moduleState["informations"] then return end
		local uid
		if tomQuest2ParentFrame then
			tomQuest2ParentFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tomQuest2ParentFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		end
		if tomQuest2QlFrame then
			tomQuest2QlFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tomQuest2QlFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			module:lockUnlockQlFrame()
			uid = tomQuest2QlFrame.currentUid
			tomQuest2QlFrame.currentUid = nil
		end
		if tomQuest2LhFrame then
			tomQuest2LhFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tomQuest2LhFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			module:lockUnlockLhFrame()
			uid = tomQuest2LhFrame.currentUid or uid
			tomQuest2LhFrame.currentUid = nil
		end
		if uid then
			module:displayQuestInformations(nil, uid)
		end
	end
	function frameSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["informations"] then return end
		local uid
		if tomQuest2QlFrame then
			module:lockUnlockQlFrame()
			uid = tomQuest2QlFrame.currentUid
			tomQuest2QlFrame.currentUid = nil
		end
		if tomQuest2LhFrame then
			module:lockUnlockLhFrame()
			uid = tomQuest2LhFrame.currentUid or uid
			tomQuest2LhFrame.currentUid = nil
		end
		if uid then
			module:displayQuestInformations(nil, uid)
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				common = {
					type = "group",
					name = LI["Common settings"],
					args = {
						backDropColor = {
							name = LI["Background color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 1,
						},
						borderColor = {
							name = LI["Border color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 2,
						},
						attach = {
							name = LI["Attach LightHeaded to QuestLog"],
							type = "select",
							values = module.attachSelect,
							set = frameSetter,
							width = "full",
							order = 3,
						},
					},
				},
				questLog = {
					type = "group",
					name = LI["Quests log settings"],
					args = {
						showQuestLog = {
							name = LI["Show tomQuest2 quest log"],
							type = "toggle",
							set = frameSetter,
							width = "full",
							order = 1,
						},
						qlFont = {
							name = LI["Font"],
							type = "select",
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							width = "full",
							order = 2,
						},
						qlFontSizeH1 = {
							name = LI["Quest title font size"],
							type="range",
							min = 6,
							max = 20,
							step = 1,
							width = "full",
							order = 3,
						},
						qlFontSizeH2 = {
							name = LI["Headers font size"],
							type="range",
							min = 6,
							max = 20,
							step = 1,
							width = "full",
							order = 4,
						},
						qlFontSize = {
							name = LI["text font size"],
							type="range",
							min = 6,
							max = 20,
							step = 1,
							width = "full",
							order = 5,
						},
					},
				},
				lightHeaded = {
					type = "group",
					name = LI["LightHeaded settings"],
					args = {
						showLightHeaded = {
							name = LI["Show LightHeaded into tomQuest2"],
							type = "toggle",
							set = frameSetter,
							width = "full",
							order = 1,
						},
						showAllCommentsOnOnePage = {
							name = LI["Show all LightHeaded comments on the same page"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						lhFont = {
							name = LI["Font"],
							type = "select",
							dialogControl = 'LSM30_Font',
							values = AceGUIWidgetLSMlists.font,
							width = "full",
							order = 3,
						},
						lhFontSize = {
							name = LI["Font size"],
							type="range",
							min = 6,
							max = 20,
							step = 1,
							width = "full",
							order = 4,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("informations", defaults, "Default")
	prdb = self.db.profile
	self:SetEnabledState(core:getModuleEnabled("informations"))
	core.moduleState["informations"] = core:getModuleEnabled("informations")
	core:registerModuleOptions("informations", getOptions, module.modName)
end

function module:OnEnable()
	core.RegisterCallback(self, "displayQuestInformations")
	core.RegisterCallback(self, "playerQuestLost", "questLost")
	core.RegisterCallback(self, "partyQuestLost", "questLost")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
end

---------------------------
--- Main Module Program ---
---------------------------
function module:questLost(event, commHandler, uid, abandoning, title, sender)
	if core.quests[uid] then return end
	local tq2QlFrame = self.tq2QlFrame
	local tq2LhFrame = self.tq2LhFrame
	local tq2ParentFrame = self.tq2ParentFrame
	if prdb.showQuestLog and prdb.showLightHeaded and LightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		if tq2ParentFrame and tq2ParentFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2ParentFrame:Hide()
			tq2QlFrame.currentUid = nil
			tq2LhFrame.currentUid = nil
			return
		end
	else
		if tq2QlFrame and tq2QlFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2QlFrame:Hide()
			tq2QlFrame.currentUid = nil
			return
		elseif tq2LhFrame and tq2LhFrame:IsVisible() and tq2LhFrame.currentUid and tq2LhFrame.currentUid == uid then
			tq2LhFrame:Hide()
			tq2LhFrame.currentUid = nil
			return
		end
	end
end

-----------------------
--- Frame Functions ---
-----------------------
function module:createLhGUI()
	if self.tq2LhFrame then return end
	local tq2LhFrame = CreateFrame("Frame", "tomQuest2LhFrame", UIParent)
	self.tq2LhFrame = tq2LhFrame
	UISpecialFrames[#UISpecialFrames+1] = "tomQuest2LhFrame"
	tq2LhFrame:SetWidth(640)
	tq2LhFrame:SetHeight(512)
	tq2LhFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	tq2LhFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])

	local close = CreateFrame("Button", nil, tq2LhFrame, "UIPanelCloseButton")
	close:SetPoint("TOPRIGHT")
	close:SetScript("OnClick", function()
		tq2LhFrame.currentUid = nil
		tq2LhFrame:Hide()
	end)

	local resize = CreateFrame("Button", nil, tq2LhFrame)
	resize:SetNormalTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHighlightTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHeight(12)
	resize:SetWidth(12)
	resize:SetPoint("BOTTOMRIGHT", -3, 3)
	self:setFrameTooltip(resize, LI["Click to resize tomQuest2"])

	local titlereg = CreateFrame("Button", nil, tq2LhFrame)
	titlereg:SetPoint("TOPLEFT", 5, -5)
	titlereg:SetPoint("TOPRIGHT", 0, 0)
	titlereg:SetHeight(20)
	titlereg:SetScript("OnMouseDown", function(frame)
		tq2LhFrame:StartMoving()
	end)
	titlereg:SetScript("OnMouseUp", function(frame)
		tq2LhFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2LhFrame")
	end)

	tq2LhFrame:EnableMouse()
	tq2LhFrame:SetMovable(1)
	tq2LhFrame:SetResizable(1)
	tq2LhFrame:SetMinResize(300, 300)
	tq2LhFrame:SetWidth(400)
	tq2LhFrame:SetFrameLevel(0)
	tq2LhFrame:SetWidth(325)
	tq2LhFrame:SetHeight(450)

	tq2LhFrame.close = close
	tq2LhFrame.resize = resize
	tq2LhFrame.titlereg = titlereg

	tq2LhFrame.icon = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.icon:SetHeight(64)
	tq2LhFrame.icon:SetWidth(64)
	tq2LhFrame.icon:SetNormalTexture("Interface\\AddOns\\LightHeaded\\images\\wh_icon")
	tq2LhFrame.icon:SetHighlightTexture("Interface\\AddOns\\LightHeaded\\images\\wh_icon_hover")
	tq2LhFrame.icon:GetHighlightTexture():SetBlendMode("BLEND")
	tq2LhFrame.icon:SetPoint("BOTTOMRIGHT", -10, -10)

	StaticPopupDialogs["LIGHTHEADED_ABOUT_DIALOG"] = {
		text = "LightHeaded (C) 2007-2008 by Jim Whitehead\n\nThis add-on only provides you with the comments and some basic information on quests. For the full information, be sure to visit http://www.wowhead.com",
		button1 = TEXT(OKAY),
		OnAccept = function()end,
		timeout = 0,
		hideOnEscape = 1
	}

	tq2LhFrame.icon:SetScript("OnClick", function()
		StaticPopup_Show("LIGHTHEADED_ABOUT_DIALOG")
	end)

	tq2LhFrame.page = tq2LhFrame:CreateFontString(nil, "ARTWORK")
	tq2LhFrame.page:SetFontObject(GameFontNormalSmall)
	tq2LhFrame.page:SetPoint("BOTTOM", -32, 15)

	tq2LhFrame.prev = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.prev:SetWidth(21)
	tq2LhFrame.prev:SetHeight(21)
	tq2LhFrame.prev:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up")
	tq2LhFrame.prev:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down")
	tq2LhFrame.prev:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled")
	tq2LhFrame.prev:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	tq2LhFrame.prev:SetPoint("RIGHT", tq2LhFrame.page, "LEFT", -25, 0)

	self:setFrameTooltip(tq2LhFrame.prev, LI["Previous page"])
	tq2LhFrame.prev:SetScript("OnClick", function(frame)
		if tq2LhFrame.currentPage then
			self:updateQuestsLightHeadedInformations(tq2LhFrame.currentUid, tq2LhFrame.currentPage - 1)
		end
	end)

	tq2LhFrame.next = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.next:SetWidth(21)
	tq2LhFrame.next:SetHeight(21)
	tq2LhFrame.next:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up")
	tq2LhFrame.next:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down")
	tq2LhFrame.next:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled")
	tq2LhFrame.next:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	tq2LhFrame.next:SetPoint("LEFT", tq2LhFrame.page, "RIGHT", 25, 0)

	self:setFrameTooltip(tq2LhFrame.next, LI["Next page"])
	tq2LhFrame.next:SetScript("OnClick", function(frame)
		if tq2LhFrame.currentPage then
			self:updateQuestsLightHeadedInformations(tq2LhFrame.currentUid, tq2LhFrame.currentPage + 1)
		end
	end)

	tq2LhFrame.title = tq2LhFrame:CreateFontString(nil, "ARTWORK")
	tq2LhFrame.title:SetPoint("TOP", 0, -4)

	tq2LhFrame.scroll = CreateFrame("ScrollFrame", "tomQuest2LhScrollFrame", tq2LhFrame, "UIPanelScrollFrameTemplate")
	tq2LhFrame.scroll:SetPoint("TOPLEFT", 30, -30)
	tq2LhFrame.scroll:SetPoint("BOTTOMRIGHT", -35, 55)

	tq2LhFrame.text = CreateFrame("SimpleHTML", "tomQuest2LhSimpleHTML", tq2LhFrame.scroll)
	tq2LhFrame.text:SetPoint("TOPLEFT", 0, 0)
	tq2LhFrame.text:SetHeight(300)
	tq2LhFrame.text:SetFontObject(GameFontHighlight)

	local function tq2LhOnUpdateFix(self, elapsed)
		self.text:SetText(self.updateText)
		self.updateText = nil
		self:SetScript("OnUpdate", nil)
	end

	function tq2LhFrame.text:mySetText(msg)
		self.text = msg
		tq2LhFrame.updateText = msg
		tq2LhFrame:SetScript("OnUpdate", tq2LhOnUpdateFix)
	end

	function tq2LhFrame.text:updateSize()
		local lhw = tomQuest2LhFrame:GetWidth()
		local width = lhw - 65
		self:SetWidth(width)
		self:mySetText(self.text)
	end
	tq2LhFrame.text:SetScript("OnHyperlinkClick", function(...) self:onHyperlinkClick(...) end)
	tq2LhFrame.text:SetScript("OnHyperlinkEnter", function(...) self:onHyperlinkEnter(...) end)
	tq2LhFrame.text:SetScript("OnHyperlinkLeave", function() GameTooltip:Hide() end)
	tq2LhFrame.scroll:SetScrollChild(tq2LhFrame.text)
	resize:SetScript("OnMouseDown", function(frame)
		tomQuest2LhFrame:StartSizing()
	end)
	resize:SetScript("OnMouseUp", function(frame)
		tomQuest2LhFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2LhFrame")
		tq2LhFrame.text:updateSize()
	end)
	tq2LhFrame:SetAlpha(1)
	tq2LhFrame:Show()
	self:lockUnlockLhFrame()
end

function module:createQlGUI()
	if self.tq2QlFrame then return end
	local tq2QlFrame = CreateFrame("Frame", "tomQuest2QlFrame", UIParent)
	self.tq2QlFrame = tq2QlFrame
	UISpecialFrames[#UISpecialFrames+1] = "tomQuest2QlFrame"
	tq2QlFrame:SetWidth(640)
	tq2QlFrame:SetHeight(512)
	tq2QlFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	tq2QlFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])

	local close = CreateFrame("Button", nil, tq2QlFrame, "UIPanelCloseButton")
	close:SetPoint("TOPRIGHT")
	close:SetScript("OnClick", function()
		tq2QlFrame.currentUid = nil
		tq2QlFrame:Hide()
	end)

	local resize = CreateFrame("Button", nil, tq2QlFrame)
	resize:SetNormalTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHighlightTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHeight(12)
	resize:SetWidth(12)
	resize:SetPoint("BOTTOMRIGHT", -3, 3)
	self:setFrameTooltip(resize, LI["Click to resize tomQuest2"])

	local titlereg = CreateFrame("Button", nil, tq2QlFrame)
	titlereg:SetPoint("TOPLEFT", 5, -5)
	titlereg:SetPoint("TOPRIGHT", 0, 0)
	titlereg:SetHeight(20)
	titlereg:SetScript("OnMouseDown", function(frame)
		tq2QlFrame:StartMoving()
	end)
	titlereg:SetScript("OnMouseUp", function(frame)
		tq2QlFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2QlFrame")
	end)

	tq2QlFrame:EnableMouse()
	tq2QlFrame:SetMovable(1)
	tq2QlFrame:SetResizable(1)
	tq2QlFrame:SetMinResize(300, 300)
	tq2QlFrame:SetWidth(400)
	tq2QlFrame:SetFrameLevel(0)
	tq2QlFrame:SetWidth(325)
	tq2QlFrame:SetHeight(450)

	tq2QlFrame.close = close
	tq2QlFrame.resize = resize
	tq2QlFrame.titlereg = titlereg

	tq2QlFrame.title = tq2QlFrame:CreateFontString(nil, "ARTWORK")
	tq2QlFrame.title:SetPoint("TOP", 0, -4)

	tq2QlFrame.scroll = CreateFrame("ScrollFrame", "tomQuest2QlScrollFrame", tq2QlFrame, "UIPanelScrollFrameTemplate")
	tq2QlFrame.scroll:SetPoint("TOPLEFT", 30, -30)
	tq2QlFrame.scroll:SetPoint("BOTTOMRIGHT", -35, 55)

	tq2QlFrame.text = CreateFrame("SimpleHTML", "tomQuest2QlSimpleHTML", tq2QlFrame.scroll)
	tq2QlFrame.text:SetPoint("TOPLEFT", 0, 0)
	tq2QlFrame.text:SetHeight(300)
	tq2QlFrame.text:SetFontObject(GameFontHighlight)

	tq2QlFrame.abandon = CreateFrame("Button", "tq2QlAbandonFrame", tq2QlFrame, "UIPanelButtonTemplate")
	tq2QlFrame.abandon:SetWidth(125)
	tq2QlFrame.abandon:SetHeight(21)
	tq2QlFrame.abandon:SetPoint("BOTTOMLEFT", tq2QlFrame, "BOTTOMLEFT", 15, 10)
	tq2QlFrame.abandon:SetText(ABANDON_QUEST)

	tq2QlFrame.share = CreateFrame("Button", "tq2QlShareFrame", tq2QlFrame, "UIPanelButtonTemplate")
	tq2QlFrame.share:SetWidth(125)
	tq2QlFrame.share:SetHeight(21)
	tq2QlFrame.share:SetPoint("BOTTOMRIGHT", tq2QlFrame, "BOTTOMRIGHT", -15, 10)
	tq2QlFrame.share:SetText(SHARE_QUEST)
	
	local function tq2QlOnUpdateFix(self, elapsed)
		self.text:SetText(self.updateText)
		self.updateText = nil
		self:SetScript("OnUpdate", nil)
	end

	function tq2QlFrame.text:mySetText(msg)
		self.text = msg
		tq2QlFrame.updateText = msg
		tq2QlFrame:SetScript("OnUpdate", tq2QlOnUpdateFix)
	end

	function tq2QlFrame.text:updateSize()
		local qlw = tomQuest2QlFrame:GetWidth()
		local width = qlw - 65
		self:SetWidth(width)
		self:mySetText(self.text)
	end
	tq2QlFrame.text:SetScript("OnHyperlinkClick", function(...) self:onHyperlinkClick(...) end)
	tq2QlFrame.text:SetScript("OnHyperlinkEnter", function(...) self:onHyperlinkEnter(...) end)
	tq2QlFrame.text:SetScript("OnHyperlinkLeave", function() GameTooltip:Hide() end)
	tq2QlFrame.scroll:SetScrollChild(tq2QlFrame.text)
	resize:SetScript("OnMouseDown", function(frame)
		tomQuest2QlFrame:StartSizing()
	end)
	resize:SetScript("OnMouseUp", function(frame)
		tomQuest2QlFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2QlFrame")
		tq2QlFrame.text:updateSize()
	end)
	tq2QlFrame:SetAlpha(1)
	tq2QlFrame:Show()
	self:lockUnlockQlFrame()
end

function module:lockUnlockQlFrame()
	local tq2QlFrame = self.tq2QlFrame
	local tq2ParentFrame = self.tq2ParentFrame
	tq2QlFrame:ClearAllPoints()
	if prdb.showQuestLog and prdb.showLightHeaded and LightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		if not tq2ParentFrame then
			tq2ParentFrame = CreateFrame("frame", "tomQuest2ParentFrame", UIParent)
			tq2ParentFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
			tq2ParentFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tq2ParentFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			UISpecialFrames[#UISpecialFrames+1] = "tomQuest2ParentFrame"
			tq2ParentFrame.close = CreateFrame("Button", nil, tq2ParentFrame, "UIPanelCloseButton")
			tq2ParentFrame.close:SetPoint("TOPRIGHT")
			tq2ParentFrame.resize = CreateFrame("Button", nil, tq2ParentFrame)
			tq2ParentFrame.resize:SetNormalTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
			tq2ParentFrame.resize:SetHighlightTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
			tq2ParentFrame.resize:SetHeight(12)
			tq2ParentFrame.resize:SetWidth(12)
			tq2ParentFrame.resize:SetPoint("BOTTOMRIGHT", -3, 3)
			self:setFrameTooltip(tq2ParentFrame.resize, LI["Click to resize tomQuest2"])
			tq2ParentFrame.resize:SetScript("OnMouseDown", function(frame)
				tq2ParentFrame:StartSizing()
			end)
			tq2ParentFrame.titlereg = CreateFrame("Button", nil, tq2ParentFrame)
			tq2ParentFrame.titlereg:SetPoint("TOPLEFT", 5, -5)
			tq2ParentFrame.titlereg:SetPoint("TOPRIGHT", 0, 0)
			tq2ParentFrame.titlereg:SetHeight(20)
			tq2ParentFrame.titlereg:SetScript("OnMouseDown", function(frame)
				tq2ParentFrame:StartMoving()
			end)
			tq2ParentFrame.titlereg:SetScript("OnMouseUp", function(frame)
				tq2ParentFrame:StopMovingOrSizing()
				self:savePosition("tomQuest2ParentFrame")
			end)
			tq2ParentFrame:EnableMouse()
			tq2ParentFrame:SetMovable(1)
			tq2ParentFrame:SetResizable(1)
			tq2ParentFrame:SetMinResize(600, 300)
			tq2ParentFrame:SetFrameLevel(0)
			tq2ParentFrame:SetWidth(prdb["tomQuest2ParentFrame"].width)
			tq2ParentFrame:SetHeight(prdb["tomQuest2ParentFrame"].height)
			self.tq2ParentFrame = tq2ParentFrame
		end
		tq2ParentFrame:SetAlpha(1)
		tq2ParentFrame:SetFrameStrata("HIGH")
		tq2QlFrame:SetParent(tomQuest2ParentFrame)
		tq2QlFrame.title:SetFontObject(GameFontHighlight)
		tq2QlFrame.title:SetText("tomQuest2 - Quest Log")
		tq2QlFrame.close:Hide()
		tq2QlFrame.resize:Hide()
		tq2QlFrame.titlereg:Hide()
		tq2QlFrame:SetBackdropColor(0, 0, 0, 0)
		tq2QlFrame:SetBackdropBorderColor(0, 0, 0, 0)
		self:restorePosition("tomQuest2ParentFrame")
		tq2QlFrame:SetWidth(prdb["tomQuest2ParentFrame"].width / 2)
		tq2QlFrame:SetHeight(prdb["tomQuest2ParentFrame"].height)
		if prdb.attach == 1 then
			tq2QlFrame:SetPoint("TOPLEFT")
			tq2QlFrame:SetPoint("BOTTOMLEFT")
		elseif prdb.attach == 2 then
			tq2QlFrame:SetPoint("TOPRIGHT")
			tq2QlFrame:SetPoint("BOTTOMRIGHT")
		end
		tq2QlFrame.text:updateSize()
	elseif prdb.showQuestLog then
		tq2QlFrame:SetParent(UIParent)
		tq2QlFrame.title:SetFontObject(GameFontHighlight)
		tq2QlFrame.title:SetText("tomQuest2 - Quest Log")
		if tq2ParentFrame then
			tq2ParentFrame:Hide()
		end
		tq2QlFrame:SetAlpha(1)
		tq2QlFrame.close:Show()
		tq2QlFrame.resize:Show()
		tq2QlFrame.titlereg:Show()
		tq2QlFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
		tq2QlFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		tq2QlFrame:SetFrameStrata("HIGH")
		self:restorePosition("tomQuest2QlFrame")
		tq2QlFrame.text:updateSize()
	elseif not prdb.showQuestLog then
		tq2QlFrame:Hide()
	end

end

function module:lockUnlockLhFrame()
	local tq2LhFrame = self.tq2LhFrame
	local tq2QlFrame = self.tq2QlFrame
	local tq2ParentFrame = self.tq2ParentFrame
	tq2LhFrame:ClearAllPoints()
	if prdb.showQuestLog and LightHeaded and prdb.showLightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		tq2ParentFrame.close:SetScript("OnClick", function()
			tq2LhFrame.currentUid = nil
			tq2QlFrame.currentUid = nil
			tq2ParentFrame:Hide()
		end)
		tq2ParentFrame.resize:SetScript("OnMouseUp", function(frame)
			tq2ParentFrame:StopMovingOrSizing()
			local width = tq2ParentFrame:GetWidth()/2
			local height = tq2ParentFrame:GetHeight()
			self:savePosition("tomQuest2ParentFrame")
			tq2QlFrame:SetWidth(width)
			tq2LhFrame:SetWidth(width)
			tq2QlFrame:SetHeight(height)
			tq2LhFrame:SetHeight(height)
			tq2QlFrame.text:updateSize()
			tq2LhFrame.text:updateSize()
		end)
		tq2LhFrame:SetParent(tq2ParentFrame)
		tq2LhFrame.title:SetFontObject(GameFontHighlight)
		tq2LhFrame.title:SetText("tomQuest2 - LightHeaded")
		tq2LhFrame:SetBackdropColor(0, 0, 0, 0)
		tq2LhFrame:SetBackdropBorderColor(0, 0, 0, 0)
		tq2LhFrame.close:Hide()
		tq2LhFrame.resize:Hide()
		tq2LhFrame.titlereg:Hide()
		tq2LhFrame:SetWidth(prdb["tomQuest2ParentFrame"].width / 2)
		tq2LhFrame:SetHeight(prdb["tomQuest2ParentFrame"].height)
		if prdb.attach == 1 then
			tq2LhFrame:SetPoint("TOPRIGHT")
			tq2LhFrame:SetPoint("BOTTOMRIGHT")
		elseif prdb.attach == 2 then
			tq2LhFrame:SetPoint("TOPLEFT")
			tq2LhFrame:SetPoint("BOTTOMLEFT")
		end
		tq2LhFrame.text:updateSize()
	elseif (prdb.showLightHeaded and LightHeaded) then
		tq2LhFrame:SetParent(UIParent)
		tq2LhFrame.title:SetFontObject(GameFontHighlight)
		tq2LhFrame.title:SetText("tomQuest2 - LightHeaded")
		if tq2ParentFrame then
			tq2ParentFrame:Hide()
		end
		tq2LhFrame:SetAlpha(1)
		tq2LhFrame.close:Show()
		tq2LhFrame.resize:Show()
		tq2LhFrame.titlereg:Show()
		tq2LhFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
		tq2LhFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		tq2LhFrame:SetFrameStrata("HIGH")
		self:restorePosition("tomQuest2LhFrame")
		tq2LhFrame.text:updateSize()
	elseif not (prdb.showLightHeaded and LightHeaded) then
		tq2LhFrame:Hide()
	end
end


function module:savePosition(name)
	local f = _G[name]
	prdb[name].point, prdb[name].x, prdb[name].y = "BOTTOMLEFT", f:GetLeft(), f:GetBottom()
	prdb[name].width = f:GetWidth()
	prdb[name].height = f:GetHeight()
end

function module:restorePosition(name)
	local f = _G[name]
	f:ClearAllPoints()
	f:SetPoint(prdb[name].point, prdb[name].x, prdb[name].y)
	f:SetWidth(prdb[name].width)
	f:SetHeight(prdb[name].height)
end

function module:setFrameTooltip(frame, text)
	frame:SetScript("OnEnter", function(frame, ...)
		GameTooltip:SetOwner(frame, "ANCHOR_TOPLEFT")
		GameTooltip:SetText(text)
		GameTooltip:Show()
	end)
	frame:SetScript("OnLeave", function(frame, ...)
		GameTooltip:Hide()
	end)
end

function module:updateQuestsLightHeadedInformations(uid, page)
	local tq2LhFrame = self.tq2LhFrame
	if not tq2LhFrame:IsVisible() then return end
	local lhText
	local num = LightHeaded:GetNumQuestComments(uid) or 0
	if #(self.stack) > 0 then
		lhText = "|cFF0066FF|Hlhref:back|h[Back]|h|r\n"..self:getQuestLightHeadedInformations(uid, page, num)
	else
		lhText = self:getQuestLightHeadedInformations(uid, page, num)
	end
	tq2LhFrame.text:mySetText(lhText)
	if showAllCommentsOnOnePage then
		tq2LhFrame.page:Hide()
		tq2LhFrame.prev:Hide()
		tq2LhFrame.next:Hide()
	else
		tq2LhFrame.page:SetText(format(LI["Page %s of %s"], page, num + 1))
		tq2LhFrame.page:Show()
		tq2LhFrame.prev:Show()
		tq2LhFrame.next:Show()
		if page == 1 then
			tq2LhFrame.prev:Disable()
		else
			tq2LhFrame.prev:Enable()
		end
		if page == num + 1 then
			tq2LhFrame.next:Disable()
		else
			tq2LhFrame.next:Enable()
		end
	end
	tq2LhFrame.currentUid = uid
	tq2LhFrame.currentPage = page
end


------------------------------
--- LightHeaded & QuestLog ---
------------------------------
function module:clean(text)
	text = gsub(text, "%[ul%]", "")
	text = gsub(text, "%[/ul%]", "")
	text = gsub(text, "%[li%]", "")
	text = gsub(text, "%[/li%]", "")
	text = gsub(text, "%[url=", "")
	text = gsub(text, "%[/url%]", "")
	return text
end

function module:displayQuestInformations(event, uid)
	if not core.quests[uid] then return end
	if prdb.showQuestLog and not self.tq2QlFrame then
		self:createQlGUI()
	end
	if LightHeaded and prdb.showLightHeaded and not tomQuest2LhFrame then
		self:createLhGUI()
	end
	local tq2QlFrame = self.tq2QlFrame
	local tq2LhFrame = self.tq2LhFrame
	local tq2ParentFrame = self.tq2ParentFrame
	if prdb.showQuestLog and prdb.showLightHeaded and LightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		if tq2ParentFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2ParentFrame:Hide()
			tq2QlFrame.currentUid = nil
			tq2LhFrame.currentUid = nil
			return
		end
		tq2ParentFrame:Show()
	else
		if tq2QlFrame and tq2QlFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2QlFrame:Hide()
			tq2QlFrame.currentUid = nil
			return
		elseif tq2LhFrame and tq2LhFrame:IsVisible() and tq2LhFrame.currentUid and tq2LhFrame.currentUid == uid then
			tq2LhFrame:Hide()
			tq2LhFrame.currentUid = nil
			return
		end
	end
	if prdb.showQuestLog then
		tq2QlFrame:Show()
		tq2QlFrame.currentUid = uid
		local fontN = CreateFont("tq2NFont")
		fontN:SetFont(LSM:Fetch("font", prdb.qlFont), prdb.qlFontSize)
		local fontH1 = CreateFont("tq2H1Font")
		fontH1:SetFont(LSM:Fetch("font", prdb.qlFont), prdb.qlFontSizeH1, "OUTLINE")
		local fontH2 = CreateFont("tq2H2Font")
		fontH2:SetFont(LSM:Fetch("font", prdb.qlFont), prdb.qlFontSizeH2, "OUTLINE")
		tq2QlFrame.title:SetFontObject(fontN)
		tq2QlFrame.text:SetFontObject(fontN)
		tq2QlFrame.text:SetFontObject("h1", fontH1)
		tq2QlFrame.text:SetFontObject("h2", fontH2)

		if core.quests[uid].complete[core.playerName] and core.quests[uid].id then
			tq2QlFrame.abandon:Enable()
			tq2QlFrame.abandon:SetScript("OnClick", function()
				if core.quests[uid].complete[core.playerName] and core.quests[uid].id then
					local wasSelected = GetQuestLogSelection()
					SelectQuestLogEntry(core.quests[uid].id)
					SetAbandonQuest()
					local items = GetAbandonQuestItems()
					if items then
						StaticPopup_Hide("ABANDON_QUEST")
						StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
					else
						StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
						StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
					end
					SelectQuestLogEntry(wasSelected)
				end
			end)
		else
			tq2QlFrame.abandon:Disable()
		end

 		if core.numPartyMembers > 0 and core.quests[uid].complete[core.playerName] and core.quests[uid].id and core.quests[uid].sharable then
			tq2QlFrame.share:Enable()
			tq2QlFrame.share:SetScript("OnClick", function()
				local wasSelected = GetQuestLogSelection()
				SelectQuestLogEntry(core.quests[uid].id)
				QuestLogPushQuest()
				SelectQuestLogEntry(wasSelected)
			end)
		else
			tq2QlFrame.share:Disable()
 		end

		local tIndex = 2
		if core.quests[uid].playersOnQuest > 1 then
			self.temp[1] = "<HTML><BODY><H1>|cff"..core.quests[uid].color..core.quests[uid].tag..core.quests[uid].title.."|r &lt;"..core.quests[uid].playersOnQuest.."&gt;</H1>"
		else
			self.temp[1] = "<HTML><BODY><H1>|cff"..core.quests[uid].color..core.quests[uid].tag..core.quests[uid].title.."|r </H1>"
		end
		if not core.quests[uid].description then
			self.temp[tIndex] = LI["Quests Informations not yet available, please try once synchronization with party is complete"]
			tq2QlFrame.text:mySetText(tconcat(self.temp, "<P><BR/>", 1, tIndex).."</P></BODY></HTML>")
		else
			if core.questsObjectives[uid] then
				self.temp[tIndex] = "<P>"..gsub(core.quests[uid].objective, "[<>]", self.repl).."</P><P>"
				for objective, qoData in pairs(core.questsObjectives[uid]) do
					tIndex = tIndex + 1
					self.temp[tIndex] = "|cff"..qoData.qoColor..qoData.desc.."|r "..qoData.qoStatus
				end
				tIndex = tIndex + 1
				self.temp[tIndex] = "</P><BR/><H2>"..LI["Description:"].."</H2>"
			else
				self.temp[tIndex] = "<P>"..gsub(core.quests[uid].objective, "[<>]", self.repl).."</P>"
				tIndex = tIndex + 1
				self.temp[tIndex] = "<H2>"..LI["Description:"].."</H2>"
			end
			tIndex = tIndex + 1
			self.temp[tIndex] = "<P>"..gsub(core.quests[uid].description, "[<>]", self.repl).."</P>"
			if (core.quests[uid].titleReward ~= "" or core.quests[uid].rewardSpellLink or (core.quests[uid].numRewards + core.quests[uid].numChoices + core.quests[uid].money + core.quests[uid].honor + core.quests[uid].talent) > 0 ) then
				tIndex = tIndex + 1
				self.temp[tIndex] = "<H2>"..LI["Rewards:"].."</H2><P>"
				if core.quests[uid].money > 0 then
					tIndex = tIndex + 1
					self.temp[tIndex] = LI["You will rececive: "]..core:formatMoney(core.quests[uid].money).."<BR/>"
				end
				if core.quests[uid].honor > 0 then
					tIndex = tIndex + 1
					self.temp[tIndex] = LI["Honor points: "]..core.quests[uid].honor.."<BR/>"
				end
				if core.quests[uid].talent > 0 then
					tIndex = tIndex + 1
					self.temp[tIndex] = LI["Talent points: "]..core.quests[uid].talent.."<BR/>"
				end
				if core.quests[uid].titleReward ~= "" then
					tIndex = tIndex + 1
					self.temp[tIndex] = LI["Title: "]..core.quests[uid].titleReward.."<BR/>"
				end
				if core.quests[uid].rewardSpellName ~= "" then
					tIndex = tIndex + 1
					self.temp[tIndex] = "|T"..core.quests[uid].rewardSpellTexture..":32|t |cff71d5ff|Hlhref:spell:"..uid.."|h["..core.quests[uid].rewardSpellName.."]|h|r".."<BR/>"
				end
				if core.quests[uid].numRewards > 0 then
					for i = 1, core.quests[uid].numRewards do
						local numItems = core.quests[uid].itemRewards[i].numItems
						if numItems > 1 then
							if core.quests[uid].itemRewards[i].texture and core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = numItems.." |T"..core.quests[uid].itemRewards[i].texture..":32|t"..core.quests[uid].itemRewards[i].link
							elseif core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = numItems.." "..core.quests[uid].itemRewards[i].link
							end
						else
							if core.quests[uid].itemRewards[i].texture and core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = "|T"..core.quests[uid].itemRewards[i].texture..":32|t"..core.quests[uid].itemRewards[i].link
							elseif core.quests[uid].itemRewards[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = core.quests[uid].itemRewards[i].link
							end
						end
					end
				end
				if core.quests[uid].numChoices > 0 then
					tIndex = tIndex + 1
					self.temp[tIndex] = LI["Choose one of the following rewards:"].."<BR/>"
					for i = 1, core.quests[uid].numChoices do
						local numItems = core.quests[uid].itemChoices[i].numItems
						if numItems > 1 then
							if core.quests[uid].itemChoices[i].texture and  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = numItems.." |T"..core.quests[uid].itemChoices[i].texture..":32|t"..core.quests[uid].itemChoices[i].link
							elseif  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = numItems.." "..core.quests[uid].itemChoices[i].link
							end
						else
							if core.quests[uid].itemChoices[i].texture and  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = "|T"..core.quests[uid].itemChoices[i].texture..":32|t"..core.quests[uid].itemChoices[i].link
							elseif  core.quests[uid].itemChoices[i].link then
								tIndex = tIndex + 1
								self.temp[tIndex] = core.quests[uid].itemChoices[i].link
							end
						end
					end
				end
				tq2QlFrame.text:mySetText(tconcat(self.temp, "<BR/>", 1, tIndex).."</P></BODY></HTML>")
			else
				tq2QlFrame.text:mySetText(tconcat(self.temp, "<BR/>", 1, tIndex).."</BODY></HTML>")
			end
		end
	end
	if LightHeaded and prdb.showLightHeaded then
		tq2LhFrame:Show()
		local font = CreateFont("tq2LhFont")
		font:SetFont(LSM:Fetch("font", prdb.lhFont), prdb.lhFontSize)
		tq2LhFrame.title:SetFontObject(font)
		tq2LhFrame.page:SetFontObject(font)
		tq2LhFrame.text:SetFontObject(font)
		tq2LhFrame.currentUid = uid
		local num = LightHeaded:GetNumQuestComments(uid) or 0
		if prdb.showAllCommentsOnOnePage then
			tq2LhFrame.currentPage = 0
			tq2LhFrame.next:Hide()
			tq2LhFrame.prev:Hide()
			tq2LhFrame.page:Hide()
			local lhText = self:getQuestLightHeadedInformations(uid, 0, num)
			if lhText then
				tq2LhFrame.text:mySetText(lhText)
			end
		else
			tq2LhFrame.currentPage = 1
			tq2LhFrame.page:Show()
			local lhText = self:getQuestLightHeadedInformations(uid, 1, num)
			if lhText then
				tq2LhFrame.page:SetText(format(LI["Page %s of %s"], 1, num + 1))
				tq2LhFrame.text:mySetText(lhText)
				if num == 0 then
					tq2LhFrame.prev:Hide()
					tq2LhFrame.next:Hide()
				else
					tq2LhFrame.prev:Show()
					tq2LhFrame.next:Show()
					tq2LhFrame.prev:Disable()
					if tq2LhFrame.currentPage == num + 1 then
						tq2LhFrame.next:Disable()
					else
						tq2LhFrame.next:Enable()
					end
				end
			end
		end
	end
end

function module:getQuestLightHeadedInformations(uid, page, numComments)
	local title = core.quests[uid] and core.quests[uid].title or LightHeaded:GetQuestName(uid)
	if page == 0 or page == 1 then
		local tIndex = 3
		for uid, sharable, level, reqlev, stype, sname, sid, etype, ename, eid, exp, rep, series in LightHeaded:IterateQuestInfo(uid) do
			self.temp[1] = "|c"..prdb.color..LI["Name:|r "]..title
			self.temp[2] = LI["Quest ID:|r"].." |Hlhref:external:http://www.wowhead.com/?quest="..uid.."|h"..uid.."|h"
			if sharable then
				self.temp[3] = LI["Sharable:|r "]..LI["Yes"]
			else
				self.temp[3] = LI["Sharable:|r "]..LI["No"]
			end
			if level ~= "" then
				tIndex = tIndex + 1
				self.temp[tIndex] = LI["Level:|r "]..level
			end
			if reqlev ~= "" then
				tIndex = tIndex + 1
				self.temp[tIndex] = LI["Required Level:|r"]..reqlev
			end
			if stype ~= "" and sname ~= "" then
				tIndex = tIndex + 1
				if stype == "npc" then
					self.temp[tIndex] = LI["Starts:|r "].."|cFF0066FF|Hlhref:npc:"..sid..":"..sname.."|h["..sname.."]|h|r"
				else
					self.temp[tIndex] = LI["Starts:|r "]..sname
				end
			end
			if etype ~= "" and ename ~= "" then
				tIndex = tIndex + 1
				if etype == "npc" then
					self.temp[tIndex] = LI["Ends:|r "].."|cFF0066FF|Hlhref:npc:"..eid..":"..ename.."|h["..ename.."]|h|r"
				else
					self.temp[tIndex] = LI["Ends:|r "]..ename
				end
				
			end
			if exp ~= "" then
				tIndex = tIndex + 1
				self.temp[tIndex] = LI["Experience:|r "]..exp
			end
			if rep ~= "" then
				tIndex = tIndex + 1
				self.temp[tIndex] = LI["Reputation Gains:|r"]
				for name, value in rep:gmatch("([^\029]+)\029([^\029]+)") do
					tIndex = tIndex + 1
					self.temp[tIndex] = " - "..name.." |r: "..value
				end
			end
			if series ~= "" then
				tIndex = tIndex + 1
				self.temp[tIndex] = LI["Quest Series:|r"]
				for step, id in series:gmatch("([^\029]+)\029([^\029]+)") do
					tIndex = tIndex + 1
					if uid == id then
						self.temp[tIndex] = step.." - "..title.." |r"
					else
						self.temp[tIndex] = step.." - |r|cFF0066FF|Hlhref:quest:"..id.."|h["..LightHeaded:GetQuestName(id).."]|h|r"
					end
				end
			end
		end
		if page == 0 and numComments > 0 then
			tIndex = tIndex + 1
			self.temp[tIndex] = LI["Comments for:|r "]..title
			for page = 1, numComments do
				tIndex = tIndex + 1
				self.temp[tIndex] = self:clean(LightHeaded:GetCommentText(uid, page))
			end
			return tconcat(self.temp, "\n|c"..prdb.color, 1, tIndex - numComments).."\n"..tconcat(self.temp, "\n", tIndex - numComments + 1, tIndex)
		else
			return tconcat(self.temp, "\n|c"..prdb.color, 1, tIndex)
		end
	elseif page > 1 then
		return "|c"..prdb.color..LI["Comments for:|r "]..title.."\n"..self:clean(LightHeaded:GetCommentText(uid, page - 1))
	end
end

function module:onHyperlinkClick(frame, link)
	local tq2LhFrame = self.tq2LhFrame
	if not link then return end
	if link:match("^lhref:npc") then
		self.stack[#(self.stack)+1] = tq2LhFrame.currentPage
		self.stack[#(self.stack)+1] = tq2LhFrame.currentUid
		local id, name = link:match("^lhref:npc:(.+):(.+)$")
		id = tonumber(id)
		if LightHeaded:LoadNPCData(id) then
			local tIndex = 1
			self.temp[tIndex] = "|cFF0066FF|Hlhref:back|h[Back]|h|r\n\n\n"..LI["Known locations for "].."|cFFFFFFFF"..name.."|r:"
			for c, z, x, y in LightHeaded:IterateNPCLocs(id) do
				tIndex = tIndex + 1
				self.temp[tIndex] = "|cFF0066FF|Hlhref:zcoord:"..c..":"..z..":"..x..":"..y..":"..name.."|h["..LightHeaded:GetCoordText(x, y).."]|h|r"
			end
			tq2LhFrame.text:mySetText(tconcat(self.temp, "\n", 1, tIndex))
		else
			tq2LhFrame.text:mySetText("|cFF0066FF|Hlhref:back|h[Back]|h|r\n\n\n"..LI["Location unknown"].."|cFFFFFFFF"..name.."|r")
		end
		tq2LhFrame.next:Disable()
		tq2LhFrame.prev:Disable()
	elseif link:match("^lhref:coord") then
		local xt, yt, uid = link:match("^lhref:coord:([^:]+):([^:]+):(.+)")
		local x = tonumber(xt)
		local y = tonumber(yt)
		if not x or not y then
			core:Print(format(LI["The coordinate you clicked on was not valid: %s"], link))
			return
		end
		local note
		if not tonumber(uid) then
			note = uid
		else
			note = "[LH] "..LightHeaded:GetQuestName(uid)
		end
		if TomTom then
			TomTom:AddWaypoint(x, y, note)
		elseif Cartographer_Waypoints and type(Cartographer_Waypoints.AddLHWaypoint) == "function" then
			Cartographer_Waypoints:AddLHWaypoint(nil, nil, x, y, note)
		elseif MN_ThottInterface and type(MN_ThottInterface_Local) == "function" then
			MN_ThottInterface_Local(x, y, note)
		elseif Cartographer_Waypoints then
			core:Print(LI["LightHeaded waypoints are not supported in this version of Cartographer_Waypoints."])
		end
	elseif link:match("^lhref:zcoord") then
		local c, z, x, y, uid = link:match("^lhref:zcoord:([^:]+):([^:]+):([^:]+):([^:]+):(.+)")
		c, z, x, y = tonumber(c), tonumber(z), tonumber(x), tonumber(y)
		local note
		if not tonumber(uid) then
			note = uid
		else
			note = "[LH]"..LightHeaded:GetQuestName(uid)
		end
		if TomTom and type(TomTom.AddZWaypoint) == "function" then
			TomTom:AddZWaypoint(c, z, x, y, note)
		elseif  Cartographer_Waypoints and type(Cartographer_Waypoints.AddLHWaypoint) == "function" then
			Cartographer_Waypoints:AddLHWaypoint(c, z, x, y, note)
		elseif MN_ThottInterface and type(MN_ThottInterface_Legacy) == "function" then
			MN_ThottInterface_Legacy(c, z, x, y, note)
		elseif TomTom then
			core:Print(LI["NPC locations are not supported in this version of TomTom."])
		elseif Cartographer_Waypoints then
			core:Print(LI["LightHeaded waypoints are not supported in this version of Cartographer_Waypoints."])
		end
	elseif link:match("^lhref:external") then
		local url = link:match("^lhref:external:(.+)$")
		ChatFrameEditBox:Show()
		ChatFrameEditBox:SetText(url)
		ChatFrameEditBox:SetFocus()
	elseif link:match("^lhref:quest:") then
		self.stack[#(self.stack)+1] = tq2LhFrame.currentPage
		self.stack[#(self.stack)+1] = tq2LhFrame.currentUid
		local uid = link:match("^lhref:quest:(.+)$")
		uid = tonumber(uid)
		if prdb.showAllCommentsOnOnePage then
			module:updateQuestsLightHeadedInformations(uid, 0)
			tq2LhFrame.next:Hide()
			tq2LhFrame.prev:Hide()
			tq2LhFrame.page:Hide()
		else
			module:updateQuestsLightHeadedInformations(uid, 1)
		end
	elseif link:match("^lhref:back$") then
		local uid = tremove(self.stack)
		local page = tremove(self.stack)
		module:updateQuestsLightHeadedInformations(uid, page)
	elseif link:match("^lhref:item:") or link:match("^item:") then
		if link:match("^lhref:item:") then
			link = link:sub(7)
		end
		local name, chatLink = GetItemInfo(link)
		if name then
			if IsControlKeyDown() then
				DressUpItemLink(link)
			elseif IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
				ChatFrameEditBox:Insert(chatLink)
			else
				SetItemRef(link)
			end
		end
	elseif not link:match("^lhref:spell:(%d+)") then
		core:Print(format(LI["Unmanaged lhref:%s "], link))
	end
end

function module:onHyperlinkEnter(frame, link)
	if not link then return end
	if link:match("^lhref:item") then
		GameTooltip:SetOwner(frame,"ANCHOR_CURSOR")
		GameTooltip:SetHyperlink(link:sub(7))
	elseif link:match("^item") then
		GameTooltip:SetOwner(frame,"ANCHOR_CURSOR")
		GameTooltip:SetHyperlink(link)
	elseif link:match("^lhref:spell:(%d+)") then
		local uid = tonumber(link:match("^lhref:spell:(%d+)"))
		if not uid then return end
		GameTooltip:SetOwner(frame,"ANCHOR_CURSOR")
		GameTooltip:SetQuestLogRewardSpell(core.quests[uid].id)
	end
end