-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTooltip", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LQTo = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTooltip")
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")
local LDB = LibStub("LibDataBroker-1.1")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, ipairs, unpack, tonumber, format, sub, tconcat = pairs, ipairs, unpack, tonumber, string.format, string.sub, table.concat
local QuestLogPushQuest, GetQuestLink, SelectQuestLogEntry, GetQuestLogSelection, WorldMap_OpenToQuest = QuestLogPushQuest, GetQuestLink, SelectQuestLogEntry, GetQuestLogSelection, WorldMap_OpenToQuest
local ItemRefTooltip, ChatEdit_InsertLink, GetItemCount, GetItemCooldown = ItemRefTooltip, ChatEdit_InsertLink, GetItemCount, GetItemCooldown
local CreateFrame, CreateFont, SetFont, ToggleFrame, ShowUIPanel, MouseIsOver = CreateFrame, CreateFont, SetFont, ToggleFrame, ShowUIPanel, MouseIsOver
local UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest = GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest

-----------------
--- Variables ---
-----------------
local prdb, chdb, options, _

module.modName = LQTo["Broker Quests Tooltip"]
module.maxQuests = MAX_QUESTS
module.minus = "Interface\\Buttons\\UI-MinusButton-Up"
module.plus = "Interface\\Buttons\\UI-PlusButton-Up"
module.CELL_MARGIN_H = 6
module.hToFontTQ2 = CreateFont("hToFontTQ2")
module.qToFontTQ2 = CreateFont("qToFontTQ2")
module.qoToFontTQ2 = CreateFont("qoToFontTQ2")

module.tempTable = {}

module.outlines = {[""] = LQTo["None"], ["OUTLINE"] = LQTo["Outline"], ["THICKOUTLINE"] = LQTo["Thick Outline"]}
module.hideObjectives = {[1] = LQTo["When you've completed the objective"], [2] = LQTo["When everyone in the party has completed the objective"], [3] = LQTo["Never"]}
module.defaultColors = {"defaultHeaderNameColor", "defaultHeaderLevelColor", "defaultQuestNameColor", "defaultQuestLevelColor", "defaultQuestObjectiveColor"}
module.interactivityList = {
	["chatLink"] = LQTo["Insert quest link in chat"],
	["objectiveProgress"] = LQTo["Insert quest objectives progress in chat"],
	["macroUpdate"] = LQTo["Update special quest item macro"],
	["questTooltip"] = LQTo["Display of quest tooltip"],
	["shareQuest"] = LQTo["Share quest"],
	["trackQuest"] = LQTo["Track quest"],
	["abandonQuest"] = LQTo["Abandon quest"],
	["displayQuest"] = LQTo["Diplay quest log"],
	["displayQuestInfo"] = LQTo["Display quest on map"],
	["nothing"] = LQTo["Do nothing"],
}
module.ldbInteractivityList = {
	["openOptions"] = LQTo["Open tomQuest2 option frame"],
	["toggleQuestTracker"] = LQTo["Toggle quests tracking frame"],
	["toggleAchievementTracker"] = LQTo["Toggle achievements tracking frame"],
	["toggleAnnouncementModule"] = LQTo["toggle chat announcement module"],
	["toggleQuestsLog"] = LQTo["Toggle Blizzard quests log"],
	["toggleSyncPartyModule"] = LQTo["Toggle party synchronization module"],
	["toggleQuestsTrackerModule"] = LQTo["Toggle quests tracking module"],
	["toggleAchievementTrackerModule"] = LQTo["Toggle achievements tracking module"],
	["nothing"] = LQTo["Do nothing"],
}
module.interactivityAction = {
	mouseRightShift = LQTo["Shift + Right click"],
	mouseRightControl = LQTo["Ctrl + Right click"],
	mouseRightAlt = LQTo["Alt + Right click"],
	mouseRight = LQTo["Right click"],
	mouseLeftShift = LQTo["Shift + Left click"],
	mouseLeftControl = LQTo["Ctrl + Left click"],
	mouseLeftAlt = LQTo["Alt + Left click"],
	mouseLeft = LQTo["Left click"],
	mouseMiddleShift = LQTo["Shift + Middle click"],
	mouseMiddleControl = LQTo["Ctrl + Middle click"],
	mouseMiddleAlt = LQTo["Alt + Middle click"],
	mouseMiddle = LQTo["Middle click"],
}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showNumCompletedQuests = true,
		showNumQuests = true,
		showMaxQuest = true,
		colorizeQuestsCounter = true,
		hideCompletedObjective = 2,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "", 
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8, 
		questObjectiveFontOutline = "",
		maxWidth = 250,
		minWidth = 250,
		statusMaxWidth = 70,
		statusMinWidth = 70,
		hasMaxHeight = true,
		maxHeight = 600,
		scale = 1,
		showQuestItemIcon = true,
		questItemIconSize = 20,
		questsCellPadding = 5,
		objectivesCellPadding = 10,
		rightCellPadding = 5,
		collapseButtonSize = 14,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		showHints = true,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "objectiveProgress",
		mouseRightAlt = "macroUpdate",
		mouseRight = "questTooltip",
		mouseLeftShift = "trackQuest",
		mouseLeftControl = "shareQuest",
		mouseLeftAlt = "abandonQuest",
		mouseLeft = "displayQuest",
		mouseMiddleShift = "displayQuestInfo",
		mouseMiddleControl = "displayQuestInfo",
		mouseMiddleAlt = "displayQuestInfo",
		mouseMiddle = "displayQuestInfo",
		ldbMouseRightShift = "openOptions",
		ldbMouseRightControl = "openOptions",
		ldbMouseRightAlt = "openOptions",
		ldbMouseRight = "openOptions",
		ldbMouseLeftShift = "toggleQuestTracker",
		ldbMouseLeftControl = "toggleAnnouncementModule",
		ldbMouseLeftAlt = "toggleAchievementTracker",
		ldbMouseLeft = "toggleQuestsLog",
		ldbMouseMiddleShift = "toggleQuestTracker",
		ldbMouseMiddleControl = "toggleAnnouncementModule",
		ldbMouseMiddleAlt = "toggleAchievementTracker",
		ldbMouseMiddle = "toggleQuestsLog",
	}
}

---------------
--- Options ---
---------------
local function normalizeWidth(min, max)
	if max < min then return max else return min end
end

local optGetter, optSetter, maxWidthSetter, minWidthSetter, statusMaxWidthSetter, statusMinWidthSetter, showQuestItemIconSetter, questItemIconSizeSetter, questsCellPaddingSetter, objectivesCellPaddingSetter, colorGetter, colorSetter, hexColorGetter, hexColorSetter, ldbTextSetter, headFontSetter, questFontSetter, qoFontSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
	end
	function maxWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module.maxWidth = value
		module.minWidth = normalizeWidth(prdb.minWidth, value)
		module.totalMaxWidth = value + module.statusMaxWidth + module.CELL_MARGIN_H
		module.totalMinWidth = normalizeWidth(module.minWidth + module.statusMinWidth + module.CELL_MARGIN_H, module.totalMaxWidth)
	end
	function minWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module.minWidth = normalizeWidth(value, module.maxWidth)
		module.totalMinWidth = normalizeWidth(value + module.statusMinWidth + module.CELL_MARGIN_H, module.totalMaxWidth)
	end
	function statusMaxWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module.statusMaxWidth = value
		module.statusMinWidth = normalizeWidth(prdb.statusMinWidth, value)
		module.totalMaxWidth = module.maxWidth + value + module.CELL_MARGIN_H
		module.totalMinWidth = normalizeWidth(module.minWidth + module.statusMinWidth + module.CELL_MARGIN_H, module.totalMaxWidth)
	end
	function statusMinWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module.statusMinWidth = normalizeWidth(value, module.statusMaxWidth)
		module.totalMinWidth = normalizeWidth(module.minWidth + value + module.CELL_MARGIN_H, module.totalMaxWidth)
	end
	function showQuestItemIconSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		if value then
			module.questItemIconSize = prdb.questItemIconSize
			module.questsCellPadding = prdb.questsCellPadding + module.questItemIconSize
			module.objectivesCellPadding = prdb.objectivesCellPadding + module.questItemIconSize
		else
			module.questItemIconSize = 0
			module.questsCellPadding = prdb.questsCellPadding
			module.objectivesCellPadding = prdb.objectivesCellPadding
		end
	end
	function questItemIconSizeSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		if prdb.showQuestItemIcon then
			module.questItemIconSize = prdb.questItemIconSize
			module.questsCellPadding = prdb.questsCellPadding + module.questItemIconSize
			module.objectivesCellPadding = prdb.objectivesCellPadding + module.questItemIconSize
		end
	end
	function questsCellPaddingSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		if prdb.showQuestItemIcon then
			module.questsCellPadding = prdb.questsCellPadding + module.questItemIconSize
		else
			module.questsCellPadding = prdb.questsCellPadding
		end
	end
	function objectivesCellPaddingSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		if prdb.showQuestItemIcon then
			module.objectivesCellPadding = prdb.objectivesCellPadding + module.questItemIconSize
		else
			module.objectivesCellPadding = prdb.objectivesCellPadding
		end
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
	end
	function hexColorGetter(info)
		return tonumber(sub(prdb[info[#info]], 1, 2), 16) / 255, tonumber(sub(prdb[info[#info]], 3, 4), 16) / 255, tonumber(sub(prdb[info[#info]], 5, 6), 16) / 255
	end
	function hexColorSetter(info, arg1, arg2, arg3)
		prdb[info[#info]] = format("%02x%02x%02x", arg1*255, arg2*255, arg3*255)
	end
	function ldbTextSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module:questLogUpdated(nil, true, false)
	end
	function headFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module.hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	end
	function questFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module.qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	end
	function qoFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTooltip"] then return end
		module.qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				questsCounter = {
					type = "group",
					name = LQTo["Quests counter"],
					set = ldbTextSetter,
					args = {
						showNumCompletedQuests = {
							name = LQTo["Show number of completed quest"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						showNumQuests = {
							name = LQTo["Show number of quests"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						showMaxQuest = {
							name = LQTo["Show maximum quests number"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						colorizeQuestsCounter = {
							name = LQTo["Color quests counter"],
							type = "toggle",
							width = "full",
							order = 4,
						},
					},
				},
				display = {
					type = "group",
					name = LQTo["Display"],
					args = {
						hideCompletedObjective = {
							name = LQTo["Hide completed quests Objectives"],
							type = "select",
							values = module.hideObjectives,
							width = "full",
							order = 1,
						},
						hideZoneLevelRange = {
							name = LQTo["Hide zone level range"],
							type = "toggle",
							width = "full",
							order = 2,
						},
					},
				},
				fonts = {
					type = "group",
					name = LQTo["Tooltip fonts"],
					args = {
						zone = {
							type = "group",
							name = LQTo["Zone"],
							set = headFontSetter,
							inline = true,
							order = 1,
							args = {
								headerFont = {
									name = LQTo["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								headerFontSize = {
									name = LQTo["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								headerFontOutline = {
									name = LQTo["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						}, 
						quests = {
							type = "group",
							name = LQTo["Quests"],
							set = questFontSetter,
							inline = true,
							order = 2,
							args = {
								questFont = {
									name = LQTo["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questFontSize = {
									name = LQTo["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questFontOutline = {
									name = LQTo["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
						questsObjectives = {
							type = "group",
							name = LQTo["Quests objectives"],
							set = qoFontSetter,
							inline = true,
							order = 3,
							args = {
								questObjectiveFont = {
									name = LQTo["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questObjectiveFontSize = { 
									name = LQTo["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questObjectiveFontOutline = {
									name = LQTo["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
					},
				},
				appearance = {
					type = "group",
					name = LQTo["Tooltip appearance"],
					args = {
						scale = {
							name = LQTo["Quests Tooltip scale"],
							type="range",
							min = 0,
							max = 1,
							step = 0.1,
							width = "full",
							order = 1,
						},
						hasMaxHeight = {
							name = LQTo["Set quests tooltip maximum height after which it will be scrollable"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						maxHeight = {
							name = LQTo["Quests tooltip max height"],
							type = "range",
							min = 100,
							max = 2000,
							step = 5,
							width = "full",
							order = 3,
						},
						collapseButtonSize = {
							name = LQTo["Collapse Button Size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 4,
						},
						questsCellPadding = {
							name = LQTo["Quests cell padding"],
							type="range",
							set = questsCellPaddingSetter,
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 5,
						},
						objectivesCellPadding = {
							name = LQTo["Quests objectives cell padding"],
							type="range",
							set = objectivesCellPaddingSetter,
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 6,
						},
						rightCellPadding = {
							name = LQTo["Right-most cell right padding to avoid cell clipping"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							width = "full",
							order = 7,
						},
						maxWidth = {
							name = LQTo["Quests Tooltip column max width"],
							type="range",
							set = maxWidthSetter,
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 8,
						},
						minWidth = {
							name = LQTo["Quests Tooltip column min width"],
							type="range",
							set = minWidthSetter,
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 9,
						},
						statusMaxWidth = {
							name = LQTo["Quests Tooltip status column max width"],
							type="range",
							set = statusMaxWidthSetter,
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 10,
						},
						statusMinWidth = {
							name = LQTo["Quests Tooltip status column min width"],
							type="range",
							set = statusMinWidthSetter,
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 11,
						},
					},
				},
				colors = {
					type = "group",
					name = LQTo["Colors"],
					args = {
						colorizeHeaderName = {
							name = LQTo["Color quests header name"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						defaultHeaderNameColor = {
							name = LQTo["Default header name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 2,
						},
						colorizeHeaderLevel = {
							name = LQTo["Color quests header level"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						defaultHeaderLevelColor = {
							name = LQTo["Default header level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 4,
						},
						colorizeQuestName = {
							name = LQTo["Color quests name"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						defaultQuestNameColor = {
							name = LQTo["Default quest name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 6,
						},
						colorizeQuestLevel = {
							name = LQTo["Color quests level"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						defaultQuestLevelColor = {
							name = LQTo["Default quest level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 8,
						},
						colorizeQuestObjective = {
							name = LQTo["Color quests objectives"],
							type = "toggle",
							width = "full",
							order = 9,
						},
						defaultQuestObjectiveColor = {
							name = LQTo["Default quest objective color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 10,
						},
						colorizeQuestObjectiveCounter = {
							name = LQTo["Color quests objectives counter"],
							type = "toggle",
							width = "full",
							order = 11,
						},
					},
				},
				questsItems = {
					type = "group",
					name = LQTo["Quests items"],
					args = {
						showQuestItemIcon = {
							name = LQTo["Show special quest item icon behind quest name"],
							type = "toggle",
							set = showQuestItemIconSetter,
							width = "full",
							order = 1,
						},
						questItemIconSize = {
							name = LQTo["Quest item icon size"],
							type = "range",
							set = questItemIconSizeSetter,
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 2,
						},
					},
				},
				tooltipInteraction = {
					type = "group",
					name = LQTo["Tooltip interaction"],
					args = {
						showHints = {
							name = LQTo["Show interractivity hints at the bottom of the tooltip"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						overTooltip = {
							name = LQTo["Display quest tooltip on mouse over"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						mouseRightShift = {
							name = LQTo["Action on right shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 3,
						},
						mouseRightControl = {
							name = LQTo["Action on right control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 4,
						},
						mouseRightAlt = {
							name = LQTo["Action on right alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 5,
						},
						mouseRight = {
							name = LQTo["Action on right click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 6,
						},
						mouseLeftShift = {
							name = LQTo["Action on left shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 7,
						},
						mouseLeftControl = {
							name = LQTo["Action on left control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 8,
						},
						mouseLeftAlt = {
							name = LQTo["Action on left alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 9,
						},
						mouseLeft = {
							name = LQTo["Action on left click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 10,
						},
						mouseMiddleShift = {
							name = LQTo["Action on middle shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 11,
						},
						mouseMiddleControl = {
							name = LQTo["Action on middle control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 12,
						},
						mouseMiddleAlt = {
							name = LQTo["Action on middle alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 13,
						},
						mouseMiddle = {
							name = LQTo["Action on middle click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							order = 14,
						},
					},
				},
				ldbInteraction = {
					type = "group",
					name = LQTo["Quests counter interaction"],
					args = {
						ldbMouseRightShift = {
							name = LQTo["Action on right shift click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 1,
						},
						ldbMouseRightControl = {
							name = LQTo["Action on right control click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 2,
						},
						ldbMouseRightAlt = {
							name = LQTo["Action on right alt click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 3,
						},
						ldbMouseRight = {
							name = LQTo["Action on right click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 4,
						},
						ldbMouseLeftShift = {
							name = LQTo["Action on left shift click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 5,
						},
						ldbMouseLeftControl = {
							name = LQTo["Action on left control click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 6,
						},
						ldbMouseLeftAlt = {
							name = LQTo["Action on left alt click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 7,
						},
						ldbMouseLeft = {
							name = LQTo["Action on left click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 8,
						},
						ldbMouseMiddleShift = {
							name = LQTo["Action on middle shift click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 9,
						},
						ldbMouseMiddleControl = {
							name = LQTo["Action on middle control click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 10,
						},
						ldbMouseMiddleAlt = {
							name = LQTo["Action on middle alt click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 11,
						},
						ldbMouseMiddle = {
							name = LQTo["Action on middle click"],
							type = "select",
							values = module.ldbInteractivityList,
							width = "full",
							order = 12,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTooltip", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:getModuleEnabled("questsTooltip"))
	core.moduleState["questsTooltip"] = core:getModuleEnabled("questsTooltip")
	core:registerModuleOptions("questsTooltip", getOptions, module.modName)
	chdb.collapseAll = chdb.collapseAll or 0
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not core.moduleState["questsTooltip"] then return end
	self.hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	self.qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	self.qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	self.maxWidth = prdb.maxWidth
	self.minWidth = normalizeWidth(prdb.minWidth, self.maxWidth)
	self.statusMaxWidth = prdb.statusMaxWidth
	self.statusMinWidth = normalizeWidth(prdb.statusMinWidth, self.statusMaxWidth)
	self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H
	self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H, self.totalMaxWidth)
	if prdb.showQuestItemIcon then
		self.questItemIconSize = prdb.questItemIconSize
		self.questsCellPadding = prdb.questsCellPadding + self.questItemIconSize
		self.objectivesCellPadding = prdb.objectivesCellPadding + self.questItemIconSize
	else
		self.questItemIconSize = 0
		self.questsCellPadding = prdb.questsCellPadding
		self.objectivesCellPadding = prdb.objectivesCellPadding
	end

end

function module:updateUsedMedia(event, mediatype, key)
	if not core.moduleState["questsTooltip"] then return end
	if mediatype == "font" then
		local updateTooltip = nil
		if key == prdb.headerFont then
			updateTooltip = true
			self.hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		end
		if key == prdb.questFont then
			updateTooltip = true
			self.qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		end
		if key == prdb.questObjectiveFont then
			updateTooltip = true
			self.qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		end
		if self.tooltipIsShown and updateTooltip then
			self:updateTooltip()
		end
	end
end

local function hideTooltip()
	if not module.tip or MouseIsOver(module.tip) then return end
	module.tooltipIsShown = nil
	module.tip:SetScript("OnUpdate", nil)
	QTIP:Release(module.tip)
	module.tip = nil
end

---------------------------
--- Main Module Program ---
---------------------------
function module:questLogUpdated(event, ldbText, tooltip)
	if not core.moduleState["questsTooltip"] then return end
	if ldbText then
		local tIndex = 0
		self.ldbColor = nil
		if prdb.showNumCompletedQuests then
			tIndex = tIndex + 1
			self.tempTable[tIndex] = core.numCompletedQuests
			if core.numQuests > 0 then
				self.ldbColor = core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests)
			else
				self.ldbColor = "ffffff"
			end
		end
		if prdb.showNumQuests then
			tIndex = tIndex + 1
			self.tempTable[tIndex] = core.numQuests
			self.ldbColor = self.ldbColor or core:getColorFromCompletionHEX(core.numQuests/self.maxQuests)
		end
		if prdb.showMaxQuest then
			tIndex = tIndex + 1
			self.tempTable[tIndex] = self.maxQuests
			self.ldbColor = self.ldbColor or "ffffff"
		end
		if tIndex > 0 then
			if prdb.colorizeQuestsCounter then
				self.dataobj.text = "|cff"..self.ldbColor..tconcat(self.tempTable, "/", 1, tIndex).."|r"
			else
				self.dataobj.text = tconcat(self.tempTable, "/", 1, tIndex)
			end
		else
			self.dataobj.text = ""
		end
	end
	if tooltip and self.tooltipIsShown then
		self:updateTooltip()
	end
end

---------------------------------------------------
--- QTIP Mouse Handler and Custom Cell Provider ---
---------------------------------------------------
local interactivity = {
	chatLink = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ChatFrameEditBox:Show()
			ChatEdit_InsertLink(GetQuestLink(core.quests[cValue1].id))
		end
	end,
	objectiveProgress = function(cValue1)
		if core.quests[cValue1].numObjectives > 0 and core.quests[cValue1].complete[core.playerName] then
			ChatFrameEditBox:Show()
			for objective, qoData in pairs(core.questsObjectives[cValue1]) do
				ChatFrameEditBox:Insert("{"..qoData.desc.." "..qoData.got[core.playerName].."/"..qoData.need.."}")
			end
		end
	end,
	macroUpdate = function(cValue1)
		core:generateSpecialItemMacro(cValue1)
	end,
	questTooltip = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ItemRefTooltip:ClearLines()
			ShowUIPanel(ItemRefTooltip)
			if not ItemRefTooltip:IsVisible() then
				ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
			end
			local link = GetQuestLink(core.quests[cValue1].id)
			if link then
				ItemRefTooltip:SetHyperlink(link)
			end
		end
	end,
	shareQuest = function(cValue1)
		if core.numPartyMembers > 0 and core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id and core.quests[cValue1].sharable then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			QuestLogPushQuest()
			SelectQuestLogEntry(wasSelected)
		end
	end,
	trackQuest = function(cValue1)
		core.cbh:Fire("toggleTrackedQuest", cValue1)
		module:updateTooltip()
	end,
	abandonQuest = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			SetAbandonQuest()
			local items = GetAbandonQuestItems()
			if items then
				StaticPopup_Hide("ABANDON_QUEST")
				StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
			else
				StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
				StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
			end
			SelectQuestLogEntry(wasSelected)
		end
	end,
	displayQuest = function(cValue1)
		core:displayQuestInformations(cValue1)
	end,
	displayQuestInfo = function(cValue1)
		WorldMap_OpenToQuest(cValue1)
	end,
	nothing = function(cValue1)
		return
	end
}

local questsOnMouseDown = function(cell, button)
	local cType1, cValue1 = cell.cType1, cell.cValue1
	if cType1 == "header" then
		if chdb.collapseAll == 1 then
			chdb.collapseAll = 0
			for zone in pairs(chdb.zoneDisplay) do
				chdb.zoneDisplay[zone] = 0
			end
		else
			chdb.collapseAll = 1
			for zone in pairs(chdb.zoneDisplay) do
				chdb.zoneDisplay[zone] = 1
			end
		end
		module:updateTooltip()
	elseif cType1 == "zone" then
		if chdb.zoneDisplay[cValue1] == 1 then
			chdb.zoneDisplay[cValue1] = 0
			chdb.collapseAll = 0
		else
			chdb.zoneDisplay[cValue1] = 1
			chdb.collapseAll = 1
		end
		module:updateTooltip()
	elseif cType1 == "quest" then
		if button == "RightButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseRightShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseRightControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseRightAlt](cValue1)
			else
				interactivity[prdb.mouseRight](cValue1)
			end
		elseif button == "LeftButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseLeftShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseLeftControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseLeftAlt](cValue1)
			else
				interactivity[prdb.mouseLeft](cValue1)
			end
		elseif button == "MiddleButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseMiddleShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseMiddleControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseMiddleAlt](cValue1)
			else
				interactivity[prdb.mouseMiddle](cValue1)
			end
		end
	end
end

local function questsOnEnter(cell)
	local cType1, cValue1 = cell.cType1, cell.cValue1
	if prdb.overTooltip and cType1 == "quest" and core.quests[cValue1] and core.quests[cValue1].id then
		local link = GetQuestLink(core.quests[cValue1].id)
		if link then
			GameTooltip:SetOwner(module.tip, module.tipAnchor)
			GameTooltip:SetHyperlink(link)
		end
	end
end

local function questsOnLeave(cell)
	GameTooltip:Hide()
end

local function buttonOnEnter(self)
	local link = self.link
	if not link then return end
	GameTooltip:SetOwner(self, "ANCHOR_LEFT")
	GameTooltip:SetHyperlink(link)
end

local function buttonOnLeave()
	GameTooltip:Hide() 
end

local function buttonOnClick(button)
	local category = button.category
	if chdb.zoneDisplay[category] == 1 then
		chdb.zoneDisplay[category] = 0
	else
		chdb.zoneDisplay[category] = 1
	end
	module:updateTooltip()
end

local function headerOnClick(button)
	if chdb.collapseAll == 1 then
		chdb.collapseAll = 0
		for zone in pairs(chdb.zoneDisplay) do
			chdb.zoneDisplay[zone] = 0
		end
	else
		chdb.collapseAll = 1
		for zone in pairs(chdb.zoneDisplay) do
			chdb.zoneDisplay[zone] = 1
		end
	end
	module:updateTooltip()
end

local function wrongButtonOnClick(button)
	core:Print(core.wrongButton)
end

local function itemOnEvent(self, event, ...)
	if event == "BAG_UPDATE_COOLDOWN" then
		local start, duration, enable = GetItemCooldown(self.link)
		if start > 0 and duration > 0 and enable > 0 then
			self.coolDown:SetCooldown(start, duration)
			self.coolDown:Show()
		else
			self.coolDown:Hide()
		end
	end
end

local cProvider, cPrototype, baseCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cType1, cType2, cValue1, cValue2)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad + 10, maxWidth, minWidth)
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	self:EnableMouse(true)
	self.cType1, self.cValue1 = cType1, cValue1
	self:SetScript("OnEnter", questsOnEnter)
	self:SetScript("OnLeave", questsOnLeave)
	self:SetScript("OnMouseDown", questsOnMouseDown)
	if cType2 == "icon" then
		if cType1 == "quest" then
			if cValue2 ~= 0 and prdb.showQuestItemIcon and (prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and core.quests[cValue1].partyComplete == 0) or (prdb.hideCompletedObjective == 1 and core.quests[cValue1].playerComplete == 0)) then
				if core.quests[cValue1].item then
					self.button = core:getItemButton()
					self.button:SetID(core.quests[cValue1].id)
					self.button:SetScript("OnEvent", itemOnEvent)
					local itemCount = GetItemCount(cValue2)
					if itemCount > 1 then
						self.button.count:SetText(itemCount)
						self.button.count:Show()
					end
					self.button:RegisterEvent("BAG_UPDATE_COOLDOWN")
					local start, duration, enable = GetItemCooldown(cValue2)
					if start > 0 and duration > 0 and enable > 0 then
						self.button.coolDown:SetCooldown(start, duration)
						self.button.coolDown:Show()
					else
						self.button.coolDown:Hide()
					end
				else
					self.button = core:getNormalButton()
					self.button:SetScript("OnEnter", buttonOnEnter)
					self.button:SetScript("OnLeave", buttonOnLeave)
					self.button:SetScript("OnClick", wrongButtonOnClick)
				end
				self.button.iconTexture:SetTexture(core.quests[cValue1].specialItemIcon)
				self.button.link = cValue2
				self.button:SetScale(prdb.questItemIconSize/self.button.height)
				self.button:SetParent(self)
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", prdb.questsCellPadding, 0)
				self.button:SetToplevel(true)
				self.button:Show()
			end
			if core.quests[cValue1].tracked then
				self.check = core:getCollapseButton()
				self.check.iconTexture:SetTexture("Interface\\Buttons\\UI-CheckBox-Check")
				self.check:SetHeight(height)
				self.check:SetWidth(height)
				self.check:SetParent(self)
				self.check:SetPoint("TOPLEFT", self, "TOPLEFT", self.fontString:GetStringWidth() + module.questsCellPadding, 0)
				self.check:SetToplevel(true)
				self.check:Show()
			end
		elseif cType1 == "zone" then
			self.button = core:getCollapseButton()
			self.button.iconTexture:SetTexture(cValue2)
			self.button:SetScript("OnClick", buttonOnClick)
			self.button.category = cValue1
			if prdb.collapseButtonSize > height then
				self.button:SetHeight(height)
				self.button:SetWidth(height)
			else
				self.button:SetHeight(prdb.collapseButtonSize)
				self.button:SetWidth(prdb.collapseButtonSize)
			end
			self.button:SetParent(self)
			self.button:SetPoint("TOPLEFT", self, "TOPLEFT")
			self.button:SetToplevel(true)
			self.button:Show()
		elseif cType1 == "header" then
			self.button = core:getCollapseButton()
			self.button.iconTexture:SetTexture(cValue2)
			self.button:SetScript("OnClick", headerOnClick)
			if prdb.collapseButtonSize > height then
				self.button:SetHeight(height)
				self.button:SetWidth(height)
			else
				self.button:SetHeight(prdb.collapseButtonSize)
				self.button:SetWidth(prdb.collapseButtonSize)
			end
			self.button:SetParent(self)
			self.button:SetPoint("TOPRIGHT", self, "TOPLEFT", (width - self.fontString:GetStringWidth())/2 - 10, 0)
			self.button:SetToplevel(true)
			self.button:Show()
		end
	end
	return width, height
end

function cPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	end
	if self.check then
		core:recycleButton(self.check)
		self.check = nil
	end
	self:EnableMouse(false)
	self.cType1, self.cValue1 = nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
end

function module:updateTooltip()
	self.tip = QTIP:Acquire("tomQuest2ToolTip", 2, "LEFT", "RIGHT")
	self.tip:Clear()
	self.tooltipIsShown = true
	self.tip:SetScale(prdb.scale)
	local tempLine = self.tip:AddLine()
	self.tip:SetCell(tempLine, 1, "tomQuest2", self.hToFontTQ2, "CENTER", 2, cProvider, 0, 0, self.totalMaxWidth, self.totalMinWidth, "header", "icon", nil, chdb.collapseAll == 0 and self.plus or self.minus)
	self.tip:AddLine(" ")
	for _, zone in ipairs(core.sortedZones) do
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		tempLine = self.tip:AddLine()
		if not prdb.hideZoneLevelRange then
			self.tip:SetCell(tempLine, 1, "|cff"..(prdb.colorizeHeaderName and core.zones[zone].color or prdb.defaultHeaderNameColor)..zone.."|r |cff"..(prdb.colorizeHeaderLevel and core.zones[zone].color or prdb.defaultHeaderLevelColor)..core.zones[zone].level.."|r", self.hToFontTQ2, "LEFT", 2, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
		else
			self.tip:SetCell(tempLine, 1, "|cff"..(prdb.colorizeHeaderName and core.zones[zone].color or prdb.defaultHeaderNameColor)..zone.."|r", self.hToFontTQ2, "LEFT", 2, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
		end
		if chdb.zoneDisplay[zone] == 1 then
			for _, uid in ipairs(core.sortedQuestsByZone[zone]) do
				tempLine = self.tip:AddLine()
				if core.quests[uid].playersOnQuest > 1 then
					self.tip:SetCell(tempLine, 1, "|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest..">", self.qToFontTQ2, "LEFT", 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink or 0)
				else
					self.tip:SetCell(tempLine, 1, "|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r", self.qToFontTQ2, "LEFT", 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink or 0)
				end
				self.tip:SetCell(tempLine, 2, core.quests[uid].pqStatus, self.qToFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid)
				if core.questsObjectives[uid] then
					for objective, qoData in ipairs(core.questsObjectives[uid]) do
						if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
							tempLine = self.tip:AddLine()
							if prdb.colorizeQuestObjective then
								self.tip:SetCell(tempLine, 1, "|cff"..qoData.qoColor..qoData.desc.."|r", self.qoToFontTQ2, "LEFT", 1, cProvider, self.objectivesCellPadding, 0, self.maxWidth, self.minWidth, "quest", nil, uid)
							else
								self.tip:SetCell(tempLine, 1, "|cff"..prdb.defaultQuestObjectiveColor..qoData.desc.."|r", self.qoToFontTQ2, "LEFT", 1, cProvider, self.objectivesCellPadding, 0, self.maxWidth, self.minWidth, "quest", nil, uid)
							end
							if prdb.colorizeQuestObjectiveCounter then
								self.tip:SetCell(tempLine, 2, qoData.qoStatus, self.qoToFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid)
							else
								self.tip:SetCell(tempLine, 2, qoData.qoStatusNC, self.qoToFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid)
							end
						end
					end
				else
					tempLine = self.tip:AddLine()
					self.tip:SetCell(tempLine, 1, core.quests[uid].objective, self.qoToFontTQ2, "LEFT", 2, cProvider, self.objectivesCellPadding, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "quest", nil, uid)
				end
			end
		end
	end
	self.tip:AddLine(" ")
	if prdb.showHints then
		for k, v in pairs(self.interactivityAction) do
			local action = prdb[k]
			if action ~= "nothing" then
				tempLine = self.tip:AddLine()
				self.tip:SetCell(tempLine, 1, "|cffeda55f"..v.."|r |cff19ff19"..self.interactivityList[action]..".|r", self.qoToFontTQ2, "LEFT", 2, cProvider, 0, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth)
			end
		end
	end
	self.tip:SmartAnchorTo(self.anchor)
	self.tip:UpdateScrolling(prdb.hasMaxHeight and prdb.maxHeight or nil)
	self.tip:Show()
end

local function dataobjOnEnter(frame)
	if not module.tooltipIsShown then
		module.anchor = frame
		module.tipAnchor = frame:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateTooltip()
	end
end

local function dataobjOnLeave(frame)
	if module.tip and MouseIsOver(module.tip) then 
		module.tip:SetScript("OnUpdate", hideTooltip)
	else
		hideTooltip()
	end
end

local ldbOnClick = {
	openOptions = function()
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.Profiles)
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.tomQuest2)
	end,
	toggleQuestTracker = function()
		tQ2toggleQuestTracker()
	end,
	toggleAchievementTracker = function()
		tQ2toggleAchievementTracker()
	end,
	toggleAnnouncementModule = function()
		core:toggleAnnounce()
	end,
	toggleQuestsLog = function()
		ToggleFrame(QuestLogFrame)
	end,
	toggleSyncPartyModule = function()
		core:toggleSyncParty()
	end,
	toggleQuestsTrackerModule = function()
		core:toggleQuestsTracker()
	end,
	toggleAchievementTrackerModule = function()
		core:toggleAchievementTracker()
	end,
	nothing = function(cValue1)
		return
	end,
}

local function dataobjOnClick(button)
	if button == "RightButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseRightShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseRightControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseRightAlt]()
		else
			ldbOnClick[prdb.ldbMouseRight]()
		end
	elseif button == "LeftButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseLeftAlt]()
		else
			ldbOnClick[prdb.ldbMouseLeft]()
		end
	elseif button == "MiddleButton" then
		if IsShiftKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleShift]()
		elseif IsControlKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleControl]()
		elseif IsAltKeyDown() then
			ldbOnClick[prdb.ldbMouseMiddleAlt]()
		else
			ldbOnClick[prdb.ldbMouseMiddle]()
		end
	end
end

function module:OnEnable()
	if not self.dataobj then
		self.dataobj = LDB:NewDataObject("tomQuest2", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon", OnEnter = function(frame) dataobjOnEnter(frame) end, OnLeave = function(frame) dataobjOnLeave(frame) end, OnClick = function(frame, button) dataobjOnClick(button) end})
	end
	core.RegisterCallback(self, "questLogUpdated")
	self.hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	self.qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	self.qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	self.maxWidth = prdb.maxWidth
	self.minWidth = normalizeWidth(prdb.minWidth, self.maxWidth)
	self.statusMaxWidth = prdb.statusMaxWidth
	self.statusMinWidth = normalizeWidth(prdb.statusMinWidth, self.statusMaxWidth)
	self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H
	self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H, self.totalMaxWidth)
	if prdb.showQuestItemIcon then
		self.questItemIconSize = prdb.questItemIconSize
		self.questsCellPadding = prdb.questsCellPadding + self.questItemIconSize
		self.objectivesCellPadding = prdb.objectivesCellPadding + self.questItemIconSize
	else
		self.questItemIconSize = 0
		self.questsCellPadding = prdb.questsCellPadding
		self.objectivesCellPadding = prdb.objectivesCellPadding
	end
end