-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTracker", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")
local achievementTracker = core:GetModule("achievementTracker")

-----------------------
--- Library loading ---
-----------------------
local LQTr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, ipairs, unpack, format, type, sub, tonumber, match, tremove, wipe, time, SecondsToTime, GetTime = pairs, ipairs, unpack, string.format, type, string.sub, tonumber, string.match, tremove, wipe, time, SecondsToTime, GetTime
local SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName, WorldMap_OpenToQuest = SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName, WorldMap_OpenToQuest
local SelectQuestLogEntry, QuestLogPushQuest, GetQuestLogSelection, GetQuestLink, GetItemCooldown = SelectQuestLogEntry, QuestLogPushQuest, GetQuestLogSelection, GetQuestLink, GetItemCooldown
local CreateFrame, CreateFont, SetFont, ChatEdit_InsertLink, ItemRefTooltip, ShowUIPanel = CreateFrame, CreateFont, SetFont, ChatEdit_InsertLink, ItemRefTooltip, ShowUIPanel
local GetSubZoneText, GetZoneText, UnitIsDead, GetItemCount, GetQuestLogSpecialItemInfo = GetSubZoneText, GetZoneText, UnitIsDead, GetItemCount, GetQuestLogSpecialItemInfo
local UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown, IsQuestLogSpecialItemInRange = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown, IsQuestLogSpecialItemInRange

-----------------
--- Variables ---
-----------------
local prdb, chdb, options

module.modName = LQTr["Quests Tracker"]
module.numAutoUntrack = 0
module.numDelayedHide = 0
module.subZone = ""
module.currentZone = ""
module.oldZone = ""
module.minus = "Interface\\Buttons\\UI-MinusButton-Up"
module.plus = "Interface\\Buttons\\UI-PlusButton-Up"
module.hTrFontTQ2 = CreateFont("hTrFontTQ2")
module.qTrFontTQ2 = CreateFont("qTrFontTQ2")
module.qoTrFontTQ2 = CreateFont("qoTrFontTQ2")
module.fixQTrFontTQ2 = CreateFont("fixQTrFontTQ2")
module.rangeUpdateDelay = TOOLTIP_UPDATE_TIME
module.CELL_MARGIN_H = 6
module.MAX_PLAYER_LEVEL = MAX_PLAYER_LEVEL

module.autoUntrack, module.delayedHide = {}, {}

module.anchorPosition = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"}
module.autoTrack = {[1] = LQTr["Always"], [2] = LQTr["Below Max Level"], [3] = LQTr["Never"]}
module.hideQuests = {[1] = LQTr["When you've completed the quest"], [2] = LQTr["When everyone in the party has completed the quest"], [3] = LQTr["Never"]}
module.hideObjectives = {[1] = LQTr["When you've completed the objective"], [2] = LQTr["When everyone in the party has completed the objective"], [3] = LQTr["Never"]}
module.outlines = {[""] = LQTr["None"], ["OUTLINE"] = LQTr["Outline"], ["THICKOUTLINE"] = LQTr["Thick Outline"]}
module.defaultColors = {"defaultHeaderNameColor", "defaultHeaderLevelColor", "defaultQuestNameColor", "defaultQuestLevelColor", "defaultQuestObjectiveColor"}
module.interactivityList = {
	["chatLink"] = LQTr["Insert quest link in chat"],
	["objectiveProgress"] = LQTr["Insert quest objectives progress in chat"],
	["macroUpdate"] = LQTr["Update special quest item macro"],
	["questTooltip"] = LQTr["Display of quest tooltip"],
	["shareQuest"] = LQTr["Share quest"],
	["trackQuest"] = LQTr["Track quest"],
	["abandonQuest"] = LQTr["Abandon quest"],
	["displayQuest"] = LQTr["Diplay quest log"],
	["displayQuestInfo"] = LQTr["Display quest on map"],
	["nothing"] = LQTr["Do nothing"],
}
module.anchorSetPoint = {
	[1] = {point = "TOPLEFT", relPoint = "BOTTOMLEFT", x = 0, y = 24},
	[2] = {point = "TOP", relPoint = "BOTTOM", x = 0, y = 24},
	[3] = {point = "TOPRIGHT", relPoint = "BOTTOMRIGHT", x = 0, y = 24},
	[4] = {point = "BOTTOMLEFT", relPoint = "TOPLEFT", x = 0, y = -24},
	[5] = {point = "BOTTOM", relPoint = "TOP", x = 0, y = -24},
	[6] = {point = "BOTTOMRIGHT", relPoint = "TOPRIGHT", x = 0, y = -24}}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		onLogIn = false,
		onQuestAccepted = 3,
		onObjectiveProgress = true,
		unTrackInactive = false,
		unTrackDelay = 5,
		onSubZone = true,
		onZone = true,
		onDaily = true,
		onClass = true,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedQuest = 2,
		delayToHideCompletedQuest = 0,
		hideCompletedObjective = 2,
		onlyCurrentZone = false,
		limitQuestLevel = false,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "",
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8,
		questObjectiveFontOutline = "",
		maxWidth = 250,
		minWidth = 250,
		statusMaxWidth = 70,
		statusMinWidth = 70,
		hasMaxHeight = true,
		maxHeight = 600,
		scale = 1,
		showQuestItemIcon = true,
		oorColor = {0.8, 0.1, 0.1},
		questItemIconSize = 20,
		showItemsStartingQuest = true,
		otherItemIconSize = 12,
		showUsableQuestItemsNotAttached = true,
		questsCellPadding = 5,
		objectivesCellPadding = 10,
		rightCellPadding = 5,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		colorizeQuestTimer = true,
		defaultQuestTimerColor = "ffffff",
		displayHeader = true,
		autoCollapseHeader = true,
		collapseButtonSize = 24,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		disableInteraction = false,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "objectiveProgress",
		mouseRightAlt = "macroUpdate",
		mouseRight = "questTooltip",
		mouseLeftShift = "trackQuest",
		mouseLeftControl = "shareQuest",
		mouseLeftAlt = "abandonQuest",
		mouseLeft = "displayQuest",
		mouseMiddleShift = "displayQuestInfo",
		mouseMiddleControl = "displayQuestInfo",
		mouseMiddleAlt = "displayQuestInfo",
		mouseMiddle = "displayQuestInfo",
	},
}

---------------
--- Options ---
---------------
local function normalizeWidth(min, max)
	if max < min then return max else return min end
end
local optGetter, optSetter, maxWidthSetter, minWidthSetter, statusMaxWidthSetter, statusMinWidthSetter, interactionSetter, showQuestItemIconSetter, questItemIconSizeSetter, questCellPaddingSetter, objectiveCellPaddingSetter, colorGetter, colorSetter, hexColorGetter, hexColorSetter, anchorSetter, combatSetter, maxLDGetter, maxLDSetter, headFontSetter, questFontSetter, qoFontSetter, interactionDisabled
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateQuestsTracker()
	end
	function maxWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module.maxWidth = value
		module.minWidth = normalizeWidth(prdb.minWidth, value)
		if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
			local numBooleanCriteriaColumns = achievementTracker.db.profile.numBooleanCriteriaColumns
			achievementTracker.maxWidth = value
			achievementTracker.columnMaxWidth = value / numBooleanCriteriaColumns
			achievementTracker.totalMaxWidth = value + module.statusMaxWidth + module.CELL_MARGIN_H * numBooleanCriteriaColumns
			achievementTracker.minWidth = module.minWidth
			achievementTracker.columnMinWidth = normalizeWidth(module.minWidth / numBooleanCriteriaColumns, achievementTracker.columnMaxWidth)
			achievementTracker.totalMinWidth = normalizeWidth(module.minWidth + module.statusMinWidth + module.CELL_MARGIN_H * numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
		end
		module:updateQuestsTracker()
	end
	function minWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module.minWidth = normalizeWidth(value, module.maxWidth)
		if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
			local numBooleanCriteriaColumns = achievementTracker.db.profile.numBooleanCriteriaColumns
			achievementTracker.minWidth = module.minWidth
			achievementTracker.columnMinWidth = normalizeWidth(value / numBooleanCriteriaColumns, achievementTracker.columnMaxWidth)
			achievementTracker.totalMinWidth = normalizeWidth(value + module.statusMinWidth + module.CELL_MARGIN_H * numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
		end
		module:updateQuestsTracker()
	end
	function statusMaxWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module.statusMaxWidth = value
		module.statusMinWidth = normalizeWidth(prdb.statusMinWidth, value)
		if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
			local numBooleanCriteriaColumns = achievementTracker.db.profile.numBooleanCriteriaColumns
			achievementTracker.statusMaxWidth = value
			achievementTracker.totalMaxWidth = module.maxWidth + value + module.CELL_MARGIN_H * numBooleanCriteriaColumns
			achievementTracker.statusMinWidth = module.statusMinWidth
			achievementTracker.totalMinWidth = normalizeWidth(module.minWidth + module.statusMinWidth + module.CELL_MARGIN_H * numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
		end
		module:updateQuestsTracker()
	end
	function statusMinWidthSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module.statusMinWidth = normalizeWidth(value, module.statusMaxWidth)
		if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
			achievementTracker.statusMinWidth = module.statusMinWidth
			achievementTracker.totalMinWidth = normalizeWidth(module.minWidth + value + module.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
		end
		module:updateQuestsTracker()
	end
	function interactionSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
			achievementTracker.disableInteraction = value
		end
		module:updateQuestsTracker()
	end
	function showQuestItemIconSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		if value then
			module.questItemIconSize = prdb.questItemIconSize
			module.questsCellPadding = prdb.questsCellPadding + module.questItemIconSize
			module.objectivesCellPadding = prdb.objectivesCellPadding + module.questItemIconSize
		else
			module.questItemIconSize = 0
			module.questsCellPadding = prdb.questsCellPadding
			module.objectivesCellPadding = prdb.objectivesCellPadding
		end
		module:updateQuestsTracker()
	end
	function questItemIconSizeSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		if prdb.showQuestItemIcon then
			module.questItemIconSize = value
		else
			module.questItemIconSize = 0
		end
		module:updateQuestsTracker()
	end
	function questCellPaddingSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		if prdb.showQuestItemIcon then
			module.questsCellPadding = value + module.questItemIconSize
		else
			module.questsCellPadding = value
		end
		module:updateQuestsTracker()
	end
	function objectiveCellPaddingSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		if prdb.showQuestItemIcon then
			module.objectivesCellPadding = value + module.questItemIconSize
		else
			module.objectivesCellPadding = value
		end
		module:updateQuestsTracker()
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateQuestsTracker()
	end
	function hexColorGetter(info)
		return tonumber(sub(prdb[info[#info]], 1, 2), 16) / 255, tonumber(sub(prdb[info[#info]], 3, 4), 16) / 255, tonumber(sub(prdb[info[#info]], 5, 6), 16) / 255
	end
	function hexColorSetter(info, arg1, arg2, arg3)
		prdb[info[#info]] = format("%02x%02x%02x", arg1*255, arg2*255, arg3*255)
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		core:computeQuestsStatus()
	end
	function anchorSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		if value then
			module.anchor:Show()
		else
			module.anchor:Hide()
		end
		module:updateQuestsTracker()
	end
	function combatSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		module.hideInCombat = nil
		if value then
			core.RegisterCallback(module, "outOfCombat")
			core.RegisterCallback(module, "inCombat")
		else
			core.UnregisterCallback(module, "outOfCombat")
			core.UnregisterCallback(module, "inCombat")
		end
	end
	function maxLDGetter(info)
		return core.db.profile.quests[info[#info]]
	end
	function maxLDSetter(info, value)
		core.db.profile.quests[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateQuestsTracker()
	end
	function headFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		module.hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateQuestsTracker()
	end
	function questFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		module.qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateQuestsTracker()
	end
	function qoFontSetter(info, value)
		prdb[info[#info]] = value
		if not core.moduleState["questsTracker"] then return end
		module.qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		QTIP:Release(module.tracker)
		module.tracker = nil
		module:updateQuestsTracker()
	end
	function interactionDisabled()
		if prdb.disableInteraction then
			return true
		else
			return false
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			childGroups = "tab",
			name = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				autotrack = {
					type = "group",
					name = LQTr["Auto track"],
					args = {
						onLogIn = {
							name = LQTr["Track all quests on log in"],
							type = "toggle",
							width = "full",
							order = 1,
						},
						onQuestAccepted = {
							name = LQTr["Auto track quests when you accept them"],
							type = "select",
							values = module.autoTrack,
							width = "full",
							order = 2,
						},
						onObjectiveProgress = {
							name = LQTr["Auto track quests on objective progress"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						unTrackInactive = {
							name = LQTr["Auto untrack inactive quests that were tracked on objective progress"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						unTrackDelay = {
							name = LQTr["Auto untrack delay (in minutes)"],
							type = "range",
							min = 1,
							max = 30,
							step = 0.5,
							width = "full",
							order = 5,
						},
						onSubZone = {
							name = LQTr["Auto track quests on sub zone change"],
							type = "toggle",
							width = "full",
							order = 6,
						},
						onZone = {
							name = LQTr["Auto track quests on zone change"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						onDaily = {
							name = LQTr["Auto track daily quests"],
							type = "toggle",
							width = "full",
							order = 8,
						},
						onClass = {
							name = LQTr["Auto track class quests"],
							type = "toggle",
							width = "full",
							order = 9,
						},
					},
				},
				display = {
					type = "group",
					name = LQTr["Display"],
					args = {
						showAnchor = {
							name = LQTr["Show quest tracker anchor"],
							type = "toggle",
							width = "full",
							set = anchorSetter,
							order = 1,
						},
						anchorPosition = {
							name = LQTr["Anchor position"],
							type = "select",
							values = module.anchorPosition,
							width = "full",
							order = 2,
						},
						displayHeader = {
							name = LQTr["Display quests by zone"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						autoCollapseHeader = {
							name = LQTr["Automatically expand /collapse zone header on zone change"],
							type = "toggle",
							width = "full",
							order = 4,
						},
						toggleTracker = {
							name = LQTr["Display tracker"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						hideTrackerInCombat = {
							name = LQTr["Hide quest tracker in combat"],
							type = "toggle",
							set = module.combatSetter,
							width = "full",
							order = 6,
						},
						hideCompletedQuest = {
							name = LQTr["Hide completed quest"],
							type = "select",
							values = module.hideQuests,
							width = "full",
							order = 7,
						},
						delayToHideCompletedQuest = {
							name = LQTr["Delay to hide completed quest"],
							type = "range",
							min = 0,
							max = 60,
							step = 0.5,
							width = "full",
							order = 8,
						},
						hideCompletedObjective = {
							name = LQTr["Hide completed quests Objectives"],
							type = "select",
							values = module.hideObjectives,
							width = "full",
							order = 9,
						},
						onlyCurrentZone = {
							name = LQTr["Hide quests that are not in the current zone"],
							type = "toggle",
							width = "full",
							order = 10,
						},
						limitQuestLevel = {
							name = LQTr["Hide high level quests (You'll have to choose the maximum level difference)"],
							type = "toggle",
							width = "full",
							order = 11,
						},
						maxLevelDiff = {
							name = LQTr["Maximum level difference between the weighted quest level and your level"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							get = maxLDGetter,
							set = maxLDSetter,
							width = "full",
							order = 12,
						},
						hideZoneLevelRange = {
							name = LQTr["Hide zone level range"],
							type = "toggle",
							width = "full",
							order = 13,
						},
					},
				},
				fonts = {
					type = "group",
					name = LQTr["Tracker fonts"],
					args = {
						zone = {
							type = "group",
							name = LQTr["Zone"],
							set = headFontSetter,
							inline = true,
							order = 1,
							args = {
								headerFont = {
									name = LQTr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								headerFontSize = {
									name = LQTr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								headerFontOutline = {
									name = LQTr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						}, 
						quests = {
							type = "group",
							name = LQTr["Quests"],
							set = questFontSetter,
							inline = true,
							order = 2,
							args = {
								questFont = {
									name = LQTr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questFontSize = {
									name = LQTr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questFontOutline = {
									name = LQTr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
						questsObjectives = {
							type = "group",
							name = LQTr["Quests objectives"],
							set = qoFontSetter,
							inline = true,
							order = 3,
							args = {
								questObjectiveFont = {
									name = LQTr["Font"],
									type = "select",
									dialogControl = 'LSM30_Font',
									values = AceGUIWidgetLSMlists.font,
									width = "full",
									order = 1,
								},
								questObjectiveFontSize = { 
									name = LQTr["Font size"],
									type = "range",
									min = 6,
									max = 20,
									step = 1,
									width = "full",
									order = 2,
								},
								questObjectiveFontOutline = {
									name = LQTr["Font outline"],
									type = "select",
									values = module.outlines,
									width = "full",
									order = 3,
								},
							},
						},
					},
				},
				appearance = {
					type = "group",
					name = LQTr["Tracker appearance"],
					args = {
						scale = {
							name = LQTr["Quests Tracker scale"],
							type = "range",
							min = 0,
							max = 1,
							step = 0.1,
							width = "full",
							order = 1,
						},
						hasMaxHeight = {
							name = LQTr["Set quests tracker maximum height after which it will be scrollable"],
							type = "toggle",
							width = "full",
							order = 2,
						},
						maxHeight = {
							name = LQTr["Quests tracker max height"],
							type = "range",
							min = 100,
							max = 2000,
							step = 5,
							width = "full",
							order = 3,
						},
						collapseButtonSize = {
							name = LQTr["Collapse Button Size"],
							type = "range",
							min = 10,
							max = 40,
							step = 1,
							width = "full",
							order = 4,
						},
						questsCellPadding = {
							name = LQTr["Quests cell padding"],
							type = "range",
							set = questCellPaddingSetter,
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 5,
						},
						objectivesCellPadding = {
							name = LQTr["Quests objectives cell padding"],
							type = "range",
							set = objectiveCellPaddingSetter,
							min = 0,
							max = 50,
							step = 1,
							width = "full",
							order = 6,
						},
						rightCellPadding = {
							name = LQTr["Right-most cell right padding to avoid cell clipping"],
							type = "range",
							min = 0,
							max = 10,
							step = 1,
							width = "full",
							order = 7,
						},
						maxWidth = {
							name = LQTr["Quests Tracker column max width"],
							type = "range",
							set = maxWidthSetter,
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 8,
						},
						minWidth = {
							name = LQTr["Quests Tracker column min width"],
							type = "range",
							set = minWidthSetter,
							min = 50,
							max = 600,
							step = 5,
							width = "full",
							order = 9,
						},
						statusMaxWidth = {
							name = LQTr["Quests Tracker status column max width"],
							type="range",
							set = statusMaxWidthSetter,
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 10,
						},
						statusMinWidth = {
							name = LQTr["Quests Tracker status column min width"],
							type="range",
							set = statusMinWidthSetter,
							min = 30,
							max = 200,
							step = 5,
							width = "full",
							order = 11,
						},
					},
				},
				color = {
					type = "group",
					name = LQTr["Colors"],
					args = {
						backDropColor = {
							name = LQTr["Quests Tracker background color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 1,
						},
						borderColor = {
							name = LQTr["Quests Tracker border color"],
							type = "color",
							hasAlpha = true,
							get = colorGetter,
							set = colorSetter,
							width = "full",
							order = 2,
						},
						colorizeHeaderName = {
							name = LQTr["Color quests header name"],
							type = "toggle",
							width = "full",
							order = 3,
						},
						defaultHeaderNameColor = {
							name = LQTr["Default header name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 4,
						},
						colorizeHeaderLevel = {
							name = LQTr["Color quests header level"],
							type = "toggle",
							width = "full",
							order = 5,
						},
						defaultHeaderLevelColor = {
							name = LQTr["Default header level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 6,
						},
						colorizeQuestName = {
							name = LQTr["Color quests name"],
							type = "toggle",
							width = "full",
							order = 7,
						},
						defaultQuestNameColor = {
							name = LQTr["Default quest name color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 8,
						},
						colorizeQuestLevel = {
							name = LQTr["Color quests level"],
							type = "toggle",
							width = "full",
							order = 9,
						},
						defaultQuestLevelColor = {
							name = LQTr["Default quest level color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 10,
						},
						colorizeQuestTimer = {
							name = LQTr["Color quests timer"],
							type = "toggle",
							width = "full",
							order = 11,
						},
						defaultQuestTimerColor = {
							name = LQTr["Default quest timer color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 12,
						},
						colorizeQuestObjective = {
							name = LQTr["Color quests objectives"],
							type = "toggle",
							width = "full",
							order = 13,
						},
						defaultQuestObjectiveColor = {
							name = LQTr["Default quest objective color"],
							type = "color",
							hasAlpha = false,
							get = hexColorGetter,
							set = hexColorSetter,
							width = "full",
							order = 14,
						},
						colorizeQuestObjectiveCounter = {
							name = LQTr["Color quests objectives counter"],
							type = "toggle",
							width = "full",
							order = 15,
						},
					},
				},
				questsItems = {
					type = "group",
					name = LQTr["Quests items"],
					args = {
						settings = {
							type = "group",
							name = LQTr["Settings"],
							inline = true,
							order = 1,
							args = {
								showQuestItemIcon = {
									name = LQTr["Show special quest item icon behind quest name"],
									type = "toggle",
									set = showQuestItemIconSetter,
									width = "full",
									order = 1,
								},
								questItemIconSize = {
									name = LQTr["Quest item icon size"],
									type = "range",
									set = questItemIconSizeSetter,
									min = 6,
									max = 36,
									step = 1,
									width = "full",
									order = 2,
								},
								oorColor = {
									name = LQTr["Out of range color"],
									type = "color",
									hasAlpha = false,
									set = colorSetter,
									get = colorGetter,
									width = "full",
									order = 3,
								},
								showItemsStartingQuest = {
									name = LQTr["Show items starting a quest in the tracker"],
									type = "toggle",
									width = "full",
									order = 4,
								},
								showUsableQuestItemsNotAttached = {
									name = LQTr["Show usable quest item not linked to a quest"],
									type = "toggle",
									width = "full",
									order = 5,
								},
								otherItemIconSize = {
									name = LQTr["Other quest item icon size"],
									type = "range",
									min = 6,
									max = 36,
									step = 1,
									width = "full",
									order = 6,
								},
							},
						},
					},
				},
				trackerInteraction = {
					type = "group",
					name = LQTr["Tracker interaction"],
					args = {
						disableInteraction = {
							type = "toggle",
							name = LQTr["Completely disable tracker interaction, should not be used when display by category is used or if you wish to make it scrollable."],
							set = interactionSetter,
							width = "full",
							order = 1,
						},
						overTooltip = {
							name = LQTr["Display quest tooltip on mouse over"],
							type = "toggle",
							width = "full",
							hidden = interactionDisabled,
							order = 2,
						},
						mouseRightShift = {
							name = LQTr["Action on right shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 3,
						},
						mouseRightControl = {
							name = LQTr["Action on right control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 4,
						},
						mouseRightAlt = {
							name = LQTr["Action on right alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 5,
						},
						mouseRight = {
							name = LQTr["Action on right click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 6,
						},
						mouseLeftShift = {
							name = LQTr["Action on left shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 7,
						},
						mouseLeftControl = {
							name = LQTr["Action on left control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 8,
						},
						mouseLeftAlt = {
							name = LQTr["Action on left alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 9,
						},
						mouseLeft = {
							name = LQTr["Action on left click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 10,
						},
						mouseMiddleShift = {
							name = LQTr["Action on middle shift click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 11,
						},
						mouseMiddleControl = {
							name = LQTr["Action on middle control click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 12,
						},
						mouseMiddleAlt = {
							name = LQTr["Action on middle alt click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 13,
						},
						mouseMiddle = {
							name = LQTr["Action on middle click"],
							type = "select",
							values = module.interactivityList,
							width = "full",
							hidden = interactionDisabled,
							order = 14,
						},
					},
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:getModuleEnabled("questsTracker"))
	core.moduleState["questsTracker"] = core:getModuleEnabled("questsTracker")
	core:registerModuleOptions("questsTracker", getOptions, module.modName)
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsSubZoneAutoTracked = chdb.questsSubZoneAutoTracked or {}
	chdb.questsZoneAutoTracked = chdb.questsZoneAutoTracked or {}
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:OnEnable()
	self.hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	self.qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	self.qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	self.fixQTrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	self.subZone = GetSubZoneText() or ""
	self.currentZone = GetZoneText() or ""
	self.anchor = CreateFrame("Button", nil, UIParent)
	self.anchor:SetHeight(24)
	self.anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	self.anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
	self.anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
	self.anchor.text = self.anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
	self.anchor.text:SetPoint("CENTER")
	self.anchor.text:SetText(LQTr["Tracker"])
	self.anchor:SetWidth(self.anchor.text:GetStringWidth() + 8)
	self.anchor:SetMovable(true)
	self.anchor:RegisterForDrag("LeftButton")
	self.anchor:SetClampedToScreen(true)
	self.anchor:SetScript("OnDragStart", function(self) self:StartMoving() end)
	self.anchor:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
		module.trackerAnchor = module.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateQuestsTracker()
	end)
	self.anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	self.disableInteraction = prdb.disableInteraction
	self.maxWidth = prdb.maxWidth
	self.statusMaxWidth = prdb.statusMaxWidth
	self.minWidth = normalizeWidth(prdb.minWidth, self.maxWidth)
	self.statusMinWidth = normalizeWidth(prdb.statusMinWidth, self.statusMaxWidth)
	if prdb.showQuestItemIcon then
		self.questItemIconSize = prdb.questItemIconSize
		self.questsCellPadding = prdb.questsCellPadding + self.questItemIconSize
		self.objectivesCellPadding = prdb.objectivesCellPadding + self.questItemIconSize
	else
		self.questItemIconSize = 0
		self.questsCellPadding = prdb.questsCellPadding
		self.objectivesCellPadding = prdb.objectivesCellPadding
	end
	if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
		achievementTracker.disableInteraction = prdb.disableInteraction
		achievementTracker.maxWidth = prdb.maxWidth
		achievementTracker.statusMaxWidth = prdb.statusMaxWidth
		achievementTracker.columnMaxWidth = achievementTracker.maxWidth / achievementTracker.db.profile.numBooleanCriteriaColumns
		achievementTracker.totalMaxWidth = achievementTracker.maxWidth + achievementTracker.statusMaxWidth + self.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns
		achievementTracker.minWidth = normalizeWidth(prdb.minWidth, achievementTracker.maxWidth)
		achievementTracker.statusMinWidth = normalizeWidth(prdb.statusMinWidth, achievementTracker.statusMaxWidth)
		achievementTracker.columnMinWidth = normalizeWidth(achievementTracker.minWidth / achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.columnMaxWidth)
		achievementTracker.totalMinWidth = normalizeWidth(achievementTracker.minWidth + achievementTracker.statusMinWidth + self.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
		core.cbh:Fire("updateAchievementTracker")
	end
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
	self:RegisterEvent("ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "zoneChanged")
	self:RawHook("IsQuestWatched", true)
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	core.RegisterCallback(self, "init")
	core.RegisterCallback(self, "updateAchievementsTrackerParameters")
	core.RegisterCallback(self, "playerQuestAccepted")
	core.RegisterCallback(self, "partyQuestAccepted")
	core.RegisterCallback(self, "playerQuestComplete", "playerQuestComplete")
	core.RegisterCallback(self, "playerQuestUnComplete", "playerQuestUnComplete")
	core.RegisterCallback(self, "playerQuestFailed", "questFailed")
	core.RegisterCallback(self, "partyQuestComplete", "partyQuestComplete")
	core.RegisterCallback(self, "partyQuestUnComplete", "partyQuestUnComplete")
	core.RegisterCallback(self, "partyQuestFailed", "questFailed")
	core.RegisterCallback(self, "playerQuestObjectiveProgress", "questObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress", "questObjectiveProgress")
	core.RegisterCallback(self, "playerQuestLost", "questLost")
	core.RegisterCallback(self, "partyQuestLost", "questLost")
	core.RegisterCallback(self, "toggleTrackedQuest")
	core.RegisterCallback(self, "questObjectiveAdded")
	core.RegisterCallback(self, "questDescriptionAdded")
	core.RegisterCallback(self, "updateQuestsTracker")
end

function module:OnDisable()
	self.anchor:Hide()
	QTIP:Release(self.tracker)
	self.tracker = nil
	core.UnregisterAllCallbacks(self)
	if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
		achievementTracker.disableInteraction = achievementTracker.db.profile.disableInteraction
		achievementTracker.maxWidth = achievementTracker.db.profile.maxWidth
		achievementTracker.statusMaxWidth = achievementTracker.db.profile.statusMaxWidth
		achievementTracker.columnMaxWidth = achievementTracker.maxWidth / achievementTracker.db.profile.numBooleanCriteriaColumns
		achievementTracker.totalMaxWidth = achievementTracker.maxWidth + achievementTracker.statusMaxWidth + achievementTracker.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns
		achievementTracker.minWidth = normalizeWidth(achievementTracker.db.profile.minWidth, achievementTracker.maxWidth)
		achievementTracker.statusMinWidth = normalizeWidth(achievementTracker.db.profile.statusMinWidth, achievementTracker.statusMaxWidth)
		achievementTracker.columnMinWidth = normalizeWidth(achievementTracker.minWidth / achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.columnMaxWidth)
		achievementTracker.totalMinWidth = normalizeWidth(achievementTracker.minWidth + achievementTracker.statusMinWidth + achievementTracker.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
		core.cbh:Fire("updateAchievementTracker")
	end
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not core.moduleState["questsTracker"] then return end
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsSubZoneAutoTracked = chdb.questsSubZoneAutoTracked or {}
	chdb.questsZoneAutoTracked = chdb.questsZoneAutoTracked or {}
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	self.hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	self.qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	self.qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	self.fixQTrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	self.anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	self.disableInteraction = prdb.disableInteraction
	self.maxWidth = prdb.maxWidth
	self.statusMaxWidth = prdb.statusMaxWidth
	self.minWidth = normalizeWidth(prdb.minWidth, self.maxWidth)
	self.statusMinWidth = normalizeWidth(prdb.statusMinWidth, self.statusMaxWidth)
	if prdb.showQuestItemIcon then
		self.questItemIconSize = prdb.questItemIconSize
		self.questsCellPadding = prdb.questsCellPadding + self.questItemIconSize
		self.objectivesCellPadding = prdb.objectivesCellPadding + self.questItemIconSize
	else
		self.questItemIconSize = 0
		self.questsCellPadding = prdb.questsCellPadding
		self.objectivesCellPadding = prdb.objectivesCellPadding
	end
	if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
		achievementTracker.disableInteraction = prdb.disableInteraction
		achievementTracker.maxWidth = prdb.maxWidth
		achievementTracker.statusMaxWidth = prdb.statusMaxWidth
		achievementTracker.columnMaxWidth = achievementTracker.maxWidth / achievementTracker.db.profile.numBooleanCriteriaColumns
		achievementTracker.totalMaxWidth = achievementTracker.maxWidth + achievementTracker.statusMaxWidth + self.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns
		achievementTracker.minWidth = normalizeWidth(prdb.minWidth, achievementTracker.maxWidth)
		achievementTracker.statusMinWidth = normalizeWidth(prdb.statusMinWidth, achievementTracker.statusMaxWidth)
		achievementTracker.columnMinWidth = normalizeWidth(achievementTracker.minWidth / achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.columnMaxWidth)
		achievementTracker.totalMinWidth = normalizeWidth(achievementTracker.minWidth + achievementTracker.statusMinWidth + self.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
		core.cbh:Fire("updateAchievementTracker")
	end
	if prdb.hideTrackerInCombat then
		core.RegisterCallback(self, "outOfCombat")
		core.RegisterCallback(self, "inCombat")
	end
end

function module:updateAchievementsTrackerParameters()
	achievementTracker.disableInteraction = prdb.disableInteraction
	achievementTracker.maxWidth = prdb.maxWidth
	achievementTracker.statusMaxWidth = prdb.statusMaxWidth
	achievementTracker.columnMaxWidth = achievementTracker.maxWidth / achievementTracker.db.profile.numBooleanCriteriaColumns
	achievementTracker.totalMaxWidth = achievementTracker.maxWidth + achievementTracker.statusMaxWidth + self.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns
	achievementTracker.minWidth = normalizeWidth(prdb.minWidth, achievementTracker.maxWidth)
	achievementTracker.statusMinWidth = normalizeWidth(prdb.statusMinWidth, achievementTracker.statusMaxWidth)
	achievementTracker.columnMinWidth = normalizeWidth(achievementTracker.minWidth / achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.columnMaxWidth)
	achievementTracker.totalMinWidth = normalizeWidth(achievementTracker.minWidth + achievementTracker.statusMinWidth + self.CELL_MARGIN_H * achievementTracker.db.profile.numBooleanCriteriaColumns, achievementTracker.totalMaxWidth)
	core.cbh:Fire("updateAchievementTracker")
end

function module:updateUsedMedia(event, mediatype, key)
	if not core.moduleState["questsTracker"] then return end
	if mediatype == "font" then
		local updateQuestTracker
		if key == prdb.headerFont then
			updateQuestTracker = true
			self.hTrFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		end
		if key == prdb.questFont then
			updateQuestTracker = true
			self.qTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		end
		if key == prdb.questObjectiveFont then
			updateQuestTracker = true
			self.qoTrFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		end
		if updateQuestTracker then
			self:updateQuestsTracker()
		end
	end
end

-------------------
--- zoneChanged ---
-------------------
function module:zoneChanged()
	if UnitIsDead("player") then return end
	local trackerNeedUpdate
	self.subZone = GetSubZoneText() or ""
	self.currentZone = GetZoneText() or ""
	if prdb.onZone and self.currentZone ~= self.oldZone then
		for uid in pairs(chdb.questsZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				RemoveQuestWatch(core.quests[uid].id, uid, true, true)
			else
				chdb.questsZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and qData.zone == self.currentZone then
				chdb.questsZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	if prdb.onSubZone and self.subZone ~= "" then
		for uid in pairs(chdb.questsSubZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				RemoveQuestWatch(core.quests[uid].id, uid, true, true)
			else
				chdb.questsSubZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and ((qData.objective and qData.objective:find(self.subZone, 1, true)) or (qData.description and qData.description:find(self.subZone, 1, true))) then
				chdb.questsSubZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	if prdb.displayHeader and prdb.autoCollapseHeader then
		for zone, zData in pairs(core.zones) do
			if self.currentZone == zone then
				chdb.zoneDisplay[zone] = 1
			elseif zData.zone then
				chdb.zoneDisplay[zone] = 0
			end
		end
		trackerNeedUpdate = true
	end
	if trackerNeedUpdate then
		for uid in pairs(self.autoUntrack) do
			if not chdb.questsTracked[uid] then
				self.autoUntrack[uid] = nil
			end
		end
		for uid in pairs(self.delayedHide) do
			if not chdb.questsTracked[uid] then
				self.delayedHide[uid] = nil
			end
		end
		self:updateQuestsTracker()
	end
	self.oldZone = self.currentZone
end

-------------
--- Hooks ---
-------------
function module:IsQuestWatched(id)
	local uid = tonumber(match(GetQuestLink(id) or '', 'quest:(%d+)'))
	return chdb.questsTracked[uid] and 1 or nil
end

function module:AddQuestWatch(id, timeToTrack, uid, batch)
	if not uid then
		uid = tonumber(match(GetQuestLink(id) or '', 'quest:(%d+)'))
		if not uid then return end
	end
	chdb.questsTracked[uid] = true
	core.quests[uid].tracked = true
	if not batch then
		self:updateQuestsTracker()
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

function module:RemoveQuestWatch(id, uid, batch, leaveTimer)
	if not uid then
		uid = tonumber(match(GetQuestLink(id) or '', 'quest:(%d+)'))
		if not uid then return end
	end
	if chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		core.quests[uid].tracked = nil
		if not batch then
			self:updateQuestsTracker()
		end
	end
	if self.autoUntrack[uid] then
		self.autoUntrack[uid] = nil
		self.numAutoUntrack = self.numAutoUntrack - 1
	end
	if self.delayedHide[uid] then
		self.delayedHide[uid] = nil
		self.numDelayedHide = self.numDelayedHide - 1
	end
	if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
		self:CancelTimer(self.timerHandler, true)
		self.timerHandler = nil
	end
	if QuestLogFrame:IsVisible() then
		QuestLog_Update()
	end
end

-----------------
--- Callbacks ---
-----------------
function module:init(event)
	if prdb.onLogIn then
		for uid in pairs (core.quests) do
			AddQuestWatch(core.quests[uid].id, nil, uid, true)
		end
	else
		for uid in pairs(chdb.questsTracked) do
			if not core.quests[uid] then
				chdb.questsTracked[uid] = nil
				chdb.questsSubZoneAutoTracked[uid] = nil
				chdb.questsZoneAutoTracked[uid] = nil
			else
				AddQuestWatch(core.quests[uid].id, nil, uid, true)
			end
		end
	end
	self:updateQuestsTracker()
end

function module:playerQuestAccepted(event, commHandler, uid, started)
	if prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= self.MAX_PLAYER_LEVEL) or (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].daily == 1) or (prdb.onClass and core.quests[uid].class) then
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
		else
			if self.autoUntrack[uid] and (prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= self.MAX_PLAYER_LEVEL) or (prdb.onDaily and core.quests[uid].daily == 1) or (prdb.onClass and core.quests[uid].class)) then
				self.autoUntrack[uid] = nil
				self.numAutoUntrack = self.numAutoUntrack - 1
				if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
					self:CancelTimer(self.timerHandler, true)
					self.timerHandler = nil
				end
			end
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
			self:updateQuestsTracker()
		end
	elseif prdb.onZone and core.quests[uid].zone == self.currentZone then
		if not chdb.questsTracked[uid] then
			chdb.questsZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		else
			chdb.questsSubZoneAutoTracked[uid] = nil
			self:updateQuestsTracker()
		end
	elseif prdb.onSubZone and not chdb.questsTracked[uid] and self.subZone ~= "" and ((core.quests[uid].objective and core.quests[uid].objective:find(self.subZone, 1, true)) or (core.quests[uid].description and core.quests[uid].description:find(self.subZone, 1, true))) then
		chdb.questsSubZoneAutoTracked[uid] = true
		AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
	elseif chdb.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:partyQuestAccepted(event, sender, uid, started, batch)
	if batch then
		if not chdb.questsTracked[uid] and prdb.onLogIn then
			AddQuestWatch(core.quests[uid].id, nil, uid, true)
		end
	elseif prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= self.MAX_PLAYER_LEVEL) or (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].daily == 1) then
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
		else
			if self.autoUntrack[uid] and (prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= self.MAX_PLAYER_LEVEL) or (prdb.onDaily and core.quests[uid].daily == 1)) then
				self.autoUntrack[uid] = nil
				self.numAutoUntrack = self.numAutoUntrack - 1
				if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
					self:CancelTimer(self.timerHandler, true)
					self.timerHandler = nil
				end
			end
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
			self:updateQuestsTracker()
		end
	elseif chdb.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:playerQuestComplete(event, commHandler, uid, sender)
	local completeTime = time()
	core.quests[uid].playerCompleteTime = completeTime
	if core.quests[uid].partyComplete == 1 then
		core.quests[uid].partyCompleteTime = completeTime
	end
	if chdb.questsTracked[uid] then
		if prdb.delayToHideCompletedQuest > 0 and ((prdb.hideCompleteQuests == 1 and core.quests[uid].playerComplete == 1) or (prdb.hideCompleteQuests == 2 and core.quests[uid].partyComplete == 1)) then
			self.delayedHide[uid] = completeTime + prdb.delayToHideCompletedQuest
			if not self.timerHandler then
				self.timerHandler = self:ScheduleRepeatingTimer("checkTimer", 1)
			end
		end
		self:updateQuestsTracker()
	end
	if self.autoUntrack[uid] then
		self.autoUntrack[uid] = nil
		self.numAutoUntrack = self.numAutoUntrack - 1
		if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
			self:CancelTimer(self.timerHandler, true)
			self.timerHandler = nil
		end
	end
end

function module:partyQuestComplete(event, commHandler, uid, sender)
	if core.quests[uid].partyComplete == 1 then
		local completeTime = time()
		core.quests[uid].partyCompleteTime = completeTime
		if chdb.questsTracked[uid] then
			if prdb.delayToHideCompletedQuest > 0 and prdb.hideCompleteQuests == 2 then
				self.delayedHide[uid] = completeTime + prdb.delayToHideCompletedQuest
				if not self.timerHandler then
					self.timerHandler = self:ScheduleRepeatingTimer("checkTimer", 1)
				end
			end
			self:updateQuestsTracker()
		end
		if self.autoUntrack[uid] then
			self.autoUntrack[uid] = nil
			self.numAutoUntrack = self.numAutoUntrack - 1
			if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
				self:CancelTimer(self.timerHandler, true)
				self.timerHandler = nil
			end
		end
	elseif chdb.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:playerQuestUnComplete(event, commHandler, uid, sender)
	core.quests[uid].playerCompleteTime = nil
	core.quests[uid].partyCompleteTime = nil
	if self.delayedHide[uid] then
		self.delayedHide[uid] = nil
		self.numDelayedHide = self.numDelayedHide - 1
		if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
			self:CancelTimer(self.timerHandler, true)
			self.timerHandler = nil
		end
	end
	if chdb.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:partyQuestUnComplete(event, commHandler, uid, sender)
	core.quests[uid].partyCompleteTime = nil
	if self.delayedHide[uid] and prdb.hideCompleteQuests == 2 then
		self.delayedHide[uid] = nil
		self.numDelayedHide = self.numDelayedHide - 1
		if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
			self:CancelTimer(self.timerHandler, true)
			self.timerHandler = nil
		end
	end
	if chdb.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:questFailed(event, commHandler, uid, sender)
	if chdb.questsTracked[uid] then
		if self.autoUntrack[uid] then
			self.autoUntrack[uid] = nil
			self.numAutoUntrack = self.numAutoUntrack - 1
			if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
				self:CancelTimer(self.timerHandler, true)
				self.timerHandler = nil
			end
		end
		self:updateQuestsTracker()
	end
end

function module:questObjectiveProgress(event, commHandler, uid, objective, got, nGot, sender)
	if prdb.onObjectiveProgress then
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid)
			if prdb.unTrackInactive and prdb.unTrackDelay > 0 then
				self.autoUntrack[uid] = time() + prdb.unTrackDelay * 60
				if not self.timerHandler then
					self.timerHandler = self:ScheduleRepeatingTimer("checkTimer", 1)
				end
			end
		else
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
			if prdb.unTrackInactive and prdb.unTrackDelay > 0 and self.autoUntrack[uid] then
				self.autoUntrack[uid] = time() + prdb.unTrackDelay * 60
			end
		end
		self:updateQuestsTracker()
	elseif chdb.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:questLost(event, commHandler, uid, abandoning, title, sender)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		if self.autoUntrack[uid] then
			self.autoUntrack[uid] = nil
			self.numAutoUntrack = self.numAutoUntrack - 1
		end
		if self.delayedHide[uid] then
			self.delayedHide[uid] = nil
			self.numDelayedHide = self.numDelayedHide - 1
		end
		if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
			self:CancelTimer(self.timerHandler, true)
			self.timerHandler = nil
		end
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		self:updateQuestsTracker()
	elseif core.quests[uid] and chdb.questsTracked[uid] then
		self:updateQuestsTracker()
	end
end

function module:toggleTrackedQuest(event, uid)
	if chdb.questsTracked[uid] then
		RemoveQuestWatch(core.quests[uid].id, uid)
	elseif not chdb.questsTracked[uid] then
		AddQuestWatch(core.quests[uid].id, nil, uid)
	end
end

function module:questObjectiveAdded(event, uid)
	if prdb.onSubZone and self.subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].objective and core.quests[uid].objective:find(self.subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		elseif core.quests[uid].numObjectives == 0 then
			self:updateQuestsTracker()
		end
	elseif core.quests[uid].numObjectives == 0 then
		self:updateQuestsTracker()
	end
end

function module:questDescriptionAdded(event, uid)
	if prdb.onSubZone and self.subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].description and core.quests[uid].description:find(self.subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil, true)
		end
	end
end

function module:inCombat(event)
	self.hideInCombat = true
	if self.tracker then self.tracker:Hide() end
end

function module:outOfCombat(event)
	self.hideInCombat = nil
	self:updateQuestsTracker()
end

---------------
--- Tracker ---
---------------
local interactivity = {
	chatLink = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ChatFrameEditBox:Show()
			ChatEdit_InsertLink(GetQuestLink(core.quests[cValue1].id))
		end
	end,
	objectiveProgress = function(cValue1)
		if core.quests[cValue1].numObjectives > 0 and core.quests[cValue1].complete[core.playerName] then
			ChatFrameEditBox:Show()
			for objective, qoData in pairs(core.questsObjectives[cValue1]) do
				ChatFrameEditBox:Insert("{"..qoData.desc.." "..qoData.got[core.playerName].."/"..qoData.need.."}")
			end
		end
	end,
	macroUpdate = function(cValue1)
		core:generateSpecialItemMacro(cValue1)
	end,
	questTooltip = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			ItemRefTooltip:ClearLines()
			ShowUIPanel(ItemRefTooltip)
			if not ItemRefTooltip:IsVisible() then
				ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
			end
			local link = GetQuestLink(core.quests[cValue1].id)
			if link then
				ItemRefTooltip:SetHyperlink(link)
			end
		end
	end,
	shareQuest = function(cValue1)
		if core.numPartyMembers > 0 and core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id and core.quests[cValue1].sharable then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			QuestLogPushQuest()
			SelectQuestLogEntry(wasSelected)
		end
	end,
	trackQuest = function(cValue1)
		module:toggleTrackedQuest(nil, cValue1)
	end,
	abandonQuest = function(cValue1)
		if core.quests[cValue1].complete[core.playerName] and core.quests[cValue1].id then
			local wasSelected = GetQuestLogSelection()
			SelectQuestLogEntry(core.quests[cValue1].id)
			SetAbandonQuest()
			local items = GetAbandonQuestItems()
			if items then
				StaticPopup_Hide("ABANDON_QUEST")
				StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
			else
				StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
				StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
			end
			SelectQuestLogEntry(wasSelected)
		end
	end,
	displayQuest = function(cValue1)
		core:displayQuestInformations(cValue1)
	end,
	displayQuestInfo = function(cValue1)
		WorldMap_OpenToQuest(cValue1)
	end,
	nothing = function(cValue1)
		return
	end
}

local function questsOnMouseDown(cell, button)
	local cType1, cValue1 = cell.cType1, cell.cValue1
	if cType1 == "zone" then
		if chdb.zoneDisplay[cValue1] == 1 then
			chdb.zoneDisplay[cValue1] = 0
		else
			chdb.zoneDisplay[cValue1] = 1
		end
		module:updateQuestsTracker()
	elseif cType1 == "quest" then
		if button == "RightButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseRightShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseRightControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseRightAlt](cValue1)
			else
				interactivity[prdb.mouseRight](cValue1)
			end
		elseif button == "LeftButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseLeftShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseLeftControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseLeftAlt](cValue1)
			else
				interactivity[prdb.mouseLeft](cValue1)
			end
		elseif button == "MiddleButton" then
			if IsShiftKeyDown() then
				interactivity[prdb.mouseMiddleShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[prdb.mouseMiddleControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[prdb.mouseMiddleAlt](cValue1)
			else
				interactivity[prdb.mouseMiddle](cValue1)
			end
		end
	elseif cType1 == "questItem" then
		core:generateSpecialItemMacro(nil, cValue1)
	end
end

local function questsOnEnter(cell)
	local cValue1 = cell.cValue1
	if prdb.overTooltip and cell.cType1 == "quest" and core.quests[cValue1] and core.quests[cValue1].id then
		local link = GetQuestLink(core.quests[cValue1].id)
		if link then
			GameTooltip:SetOwner(module.tracker, module.trackerAnchor)
			GameTooltip:SetHyperlink(link)
		end
	end
end

local function questsOnLeave(cell)
	GameTooltip:Hide()
end

local function buttonOnEnter(self)
	local link = self.link
	if not link then return end
	GameTooltip:SetOwner(self, "ANCHOR_LEFT")
	GameTooltip:SetHyperlink(link)
end

local function buttonOnLeave()
	GameTooltip:Hide() 
end

local function buttonOnClick(button)
	local category = button.category
	if not category then return end
	if chdb.zoneDisplay[category] == 1 then
		chdb.zoneDisplay[category] = 0
	else
		chdb.zoneDisplay[category] = 1
	end
	module:updateQuestsTracker()
end

local function wrongButtonOnClick(button)
	core:Print(core.wrongButton)
end

local function updateItemRange(self, elapsed)
	self.rangeTimer = self.rangeTimer - elapsed
	if self.rangeTimer <= 0  then
		if IsQuestLogSpecialItemInRange(self:GetID()) == 0 then
			self.iconTexture:SetVertexColor(0.8, 0.1, 0.1)
		else
			self.iconTexture:SetVertexColor(1.0, 1.0, 1.0)
		end
		self.rangeTimer = module.rangeUpdateDelay
	end
end

local function itemOnEvent(self, event, ...)
	if event == "BAG_UPDATE_COOLDOWN" then
		local start, duration, enable = GetItemCooldown(self.link)
		if start > 0 and duration > 0 and enable > 0 then
			self.coolDown:SetCooldown(start, duration)
			self.coolDown:Show()
		else
			self.coolDown:Hide()
		end
	elseif event == "PLAYER_TARGET_CHANGED" then
		self.rangeTimer = -1
	end
end

local function timedQuestUpdate(self, elapsed)
	self.elapsed = self.elapsed + elapsed
	if self.elapsed >= 1 then
		local uid = self.cValue1
		core.quests[uid].remainingTime = core.quests[uid].remainingTime - self.elapsed
		if core.quests[uid].playersOnQuest > 1 then
			self.fontString:SetText("|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest.."> "..(prdb.colorizeQuestTimer and core:getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or prdb.defaultQuestTimerColor)..SecondsToTime(core.quests[uid].remainingTime))
		else
			self.fontString:SetText("|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r "..(prdb.colorizeQuestTimer and core:getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or prdb.defaultQuestTimerColor)..SecondsToTime(core.quests[uid].remainingTime))
		end
		self.elapsed = 0
	end
end

local cProvider, cPrototype, baseCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cType1, cType2, cValue1, cValue2)
	local width, height = baseCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	if cType1 == "quest" and cValue2 and core.quests[cValue1].remainingTime and core.quests[cValue1].remainingTime > 0 then
		self:SetScript("OnUpdate", timedQuestUpdate)
		self.elapsed = 0
	else
		self:SetScript("OnUpdate", nil)
	end
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	elseif self.statusBar then
		core:recycleStatusBar(self.statusBar)
		self.statusBar = nil
	end
	if not self.disableInteraction then
		self:EnableMouse(true)
		self.cType1, self.cValue1 = cType1, cValue1
		self:SetScript("OnEnter", questsOnEnter)
		self:SetScript("OnLeave", questsOnLeave)
		self:SetScript("OnMouseDown", questsOnMouseDown)
		if cType2 == "icon" and cValue2 then
			if cType1 == "quest" and prdb.showQuestItemIcon and (prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and core.quests[cValue1].partyComplete == 0) or (prdb.hideCompletedObjective == 1 and core.quests[cValue1].playerComplete == 0)) then
				if core.quests[cValue1].item and core.quests[cValue1].id then
					self.button = core:getItemButton()
					self.button:SetID(core.quests[cValue1].id)
					self.button:SetScript("OnEvent", itemOnEvent)
					local itemCount = GetItemCount(cValue2)
					if itemCount > 1 then
						self.button.count:SetText(itemCount)
						self.button.count:Show()
					end
					if core.quests[cValue1].itemHasRange then
						self.button.rangeTimer = -1
						self.button:SetScript("OnUpdate", updateItemRange)
						self.button:RegisterEvent("PLAYER_TARGET_CHANGED")
					end
					self.button:RegisterEvent("BAG_UPDATE_COOLDOWN")
					local start, duration, enable = GetItemCooldown(cValue2)
					if start > 0 and duration > 0 and enable > 0 then
						self.button.coolDown:SetCooldown(start, duration)
						self.button.coolDown:Show()
					else
						self.button.coolDown:Hide()
					end
				else
					self.button = core:getNormalButton()
					self.button:SetScript("OnEnter", buttonOnEnter)
					self.button:SetScript("OnLeave", buttonOnLeave)
					self.button:SetScript("OnClick", wrongButtonOnClick)
				end
				self.button.iconTexture:SetTexture(core.quests[cValue1].specialItemIcon)
				self.button.link = cValue2
				self.button:SetScale(prdb.questItemIconSize/self.button.height)
				self.button:SetParent(self)
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", prdb.questsCellPadding, 0)
				self.button:SetToplevel(true)
				self.button:Show()
			elseif cType1 == "zone" then
				self.button = core:getCollapseButton()
				self.button:RegisterForClicks("AnyDown")
				self.button.iconTexture:SetTexture(cValue2)
				self.button:SetScript("OnClick", buttonOnClick)
				self.button.category = cValue1
				if prdb.collapseButtonSize > height then
					self.button:SetHeight(height)
					self.button:SetWidth(height)
				else
					self.button:SetHeight(prdb.collapseButtonSize)
					self.button:SetWidth(prdb.collapseButtonSize)
				end
				self.button:SetParent(self)
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT")
				self.button:SetToplevel(true)
				self.button:Show()
			elseif cType1 == "questItem" then
				self.button = core:getNormalButton()
				self.button.iconTexture:SetTexture(cValue2)
				self.button:SetScript("OnEnter", buttonOnEnter)
				self.button:SetScript("OnLeave", buttonOnLeave)
				self.button:SetScript("OnClick", wrongButtonOnClick)
				self.button.link = cValue1
				if prdb.otherItemIconSize >= height then
					self.button:SetScale(height/self.button.height)
				else
					self.button:SetScale(prdb.otherItemIconSize/self.button.height)
				end
				self.button:SetParent(self)
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", prdb.questsCellPadding, 0)
				self.button:SetToplevel(true)
				self.button:Show()
			end
		end
	else
		self:EnableMouse(false)
		self.cType1, self.cValue1 = nil, nil
		self:SetScript("OnEnter", nil)
		self:SetScript("OnLeave", nil)
		self:SetScript("OnMouseDown", nil)
	end
	return width, height
end

function cPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	elseif self.statusBar then
		core:recycleStatusBar(self.statusBar)
		self.statusBar = nil
	end
	self:EnableMouse(false)
	self.cType1, self.cValue1, self.elapsed = nil, nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
	self:SetScript("OnUpdate", nil)
end

function module:addQuestToTracker(uid)
	if chdb.questsTracked[uid] and core.quests[uid] and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or (core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + prdb.delayToHideCompletedQuest > self.displayTime) or core.quests[uid].timer)) or (prdb.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or (core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(prdb.delayToHideCompletedQuest) > self.displayTime) or core.quests[uid].timer))) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not prdb.limitQuestLevel) then
		self.tempLine = self.tracker:AddLine()
		if core.quests[uid].initialTimer then
			core.quests[uid].remainingTime = core.quests[uid].initialTimer + core.quests[uid].startTime - GetTime()
			if core.quests[uid].playersOnQuest > 1 then
				self.tracker:SetCell(self.tempLine, 1, "|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest.."> "..(prdb.colorizeQuestTimer and core:getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or prdb.defaultQuestTimerColor)..SecondsToTime(core.quests[uid].remainingTime), self.qTrFontTQ2, "LEFT", self.numColumns - 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink)
			else
				self.tracker:SetCell(self.tempLine, 1, "|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r "..(prdb.colorizeQuestTimer and core:getColorFromCompletionHEX(core.quests[uid].remainingTime / core.quests[uid].initialTimer) or prdb.defaultQuestTimerColor)..SecondsToTime(core.quests[uid].remainingTime), self.qTrFontTQ2, "LEFT", self.numColumns - 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink)
			end
		else
			if core.quests[uid].playersOnQuest > 1 then
				self.tracker:SetCell(self.tempLine, 1, "|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r <"..core.quests[uid].playersOnQuest..">", self.qTrFontTQ2, "LEFT", self.numColumns - 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink)
			else
				self.tracker:SetCell(self.tempLine, 1, "|cff"..(prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor)..core.quests[uid].tag.."|r |cff"..(prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor)..core.quests[uid].title.."|r", self.qTrFontTQ2, "LEFT", self.numColumns - 1, cProvider, self.questsCellPadding, 0, self.maxWidth, self.minWidth, "quest", "icon", uid, core.quests[uid].specialItemLink)
			end
		end
		self.tracker:SetCell(self.tempLine, self.numColumns, core.quests[uid].pqStatus, self.qTrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid)
		if core.questsObjectives[uid] then
			for objective, qoData in pairs(core.questsObjectives[uid]) do
				if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
					self.tempLine = self.tracker:AddLine()
					if prdb.colorizeQuestObjective then
						self.tracker:SetCell(self.tempLine, 1, "|cff"..qoData.qoColor..qoData.desc.."|r", self.qoTrFontTQ2, "LEFT", self.numColumns - 1, cProvider, self.objectivesCellPadding, 0, self.maxWidth, self.minWidth, "quest", nil, uid, objective)
					else
						self.tracker:SetCell(self.tempLine, 1, "|cff"..prdb.defaultQuestObjectiveColor..qoData.desc.."|r", self.qoTrFontTQ2, "LEFT", self.numColumns - 1, cProvider, self.objectivesCellPadding, 0, self.maxWidth, self.minWidth, "quest", nil, uid, objective)
					end
					if prdb.colorizeQuestObjectiveCounter then
						self.tracker:SetCell(self.tempLine, self.numColumns, qoData.qoStatus, self.qoTrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid, objective)
					else
						self.tracker:SetCell(self.tempLine, self.numColumns, qoData.qoStatusNC, self.qoTrFontTQ2, "RIGHT", 1, cProvider, 0, prdb.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "quest", nil, uid, objective)
					end
				end
			end
		else
			self.tempLine = self.tracker:AddLine()
			self.tracker:SetCell(self.tempLine, 1, core.quests[uid].objective, self.qoTrFontTQ2, "LEFT", self.numColumns, cProvider, self.objectivesCellPadding, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "quest", nil, uid)
		end
		self.tq = true
		self.replaceHeader = nil
	end
end

function module:updateQuests()
	self.tq = nil
	self.replaceHeader = nil
	self.displayTime = time()
	self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H * (self.numColumns - 1)
	self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H * (self.numColumns - 1), self.totalMaxWidth)
	if prdb.displayHeader then
		for _, zone in ipairs(core.sortedZones) do
			if ((prdb.onlyCurrentZone and zone == self.currentZone) or not prdb.onlyCurrentZone) and core.zones[zone] then
				chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
				if not self.replaceHeader then self.tempLine = self.tracker:AddLine() end
				self.replaceHeader = true
				if not prdb.hideZoneLevelRange then
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(prdb.colorizeHeaderName and core.zones[zone].color or prdb.defaultHeaderNameColor)..zone.."|r |cff"..(prdb.colorizeHeaderLevel and core.zones[zone].color or prdb.defaultHeaderLevelColor)..core.zones[zone].level.."|r", self.hTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
				else
					self.tracker:SetCell(self.tempLine, 1, "|cff"..(prdb.colorizeHeaderName and core.zones[zone].color or prdb.defaultHeaderNameColor)..zone.."|r", self.hTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
				end
				if chdb.zoneDisplay[zone] == 1 or self.disableInteraction then
					for _, uid in ipairs(core.sortedQuestsByZone[zone]) do
						self:addQuestToTracker(uid)
					end
				else
					for _, uid in ipairs(core.sortedQuestsByZone[zone]) do
						if chdb.questsTracked[uid] and core.quests[uid] and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + prdb.delayToHideCompletedQuest > self.displayTime)) or (prdb.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(prdb.delayToHideCompletedQuest) > self.displayTime))) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not prdb.limitQuestLevel) then
							self.tq = true
							self.replaceHeader = nil
							break
						end
					end
				end
			end
		end
	else
		for _, uid in ipairs(core.sortedQuests) do
			if chdb.questsTracked[uid] and core.quests[uid] and ((prdb.onlyCurrentZone and core.quests[uid].zone == self.currentZone) or not prdb.onlyCurrentZone) and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and (core.quests[uid].partyComplete == 0 or (core.quests[uid].partyCompleteTime and core.quests[uid].partyCompleteTime + prdb.delayToHideCompletedQuest > self.displayTime) or core.quests[uid].timer)) or (prdb.hideCompletedQuest == 1 and (core.quests[uid].playerComplete == 0 or (core.quests[uid].playerCompleteTime and core.quests[uid].playerCompleteTime + tonumber(prdb.delayToHideCompletedQuest) > self.displayTime) or core.quests[uid].timer))) and ((prdb.limitQuestLevel and (core.quests[uid].weightedLevel - core.playerLevel <= core.db.profile.quests.maxLevelDiff)) or not prdb.limitQuestLevel) then
				self:addQuestToTracker(uid)
			end
		end
	end
	if prdb.showItemsStartingQuest and core.numItemsStartingQuest > 0 then
		self.tq = true
		local zone = LQTr["Items starting a quest"]
		if not self.replaceHeader or not self.tempLine then
			self.tempLine = self.tracker:AddLine()
		end
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		if prdb.colorizeHeaderName then
			self.tracker:SetCell(self.tempLine, 1, "|cffffffff"..zone.."|r", self.hTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
		else
			self.tracker:SetCell(self.tempLine, 1, "|cff"..prdb.defaultHeaderNameColor..zone.."|r", self.hTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
		end
		if chdb.zoneDisplay[zone] == 1 then
			for item, itemData in pairs(core.itemStartingQuest) do
				self.tempLine = self.tracker:AddLine()
				self.tracker:SetCell(self.tempLine, 1, "|cff"..itemData.color..itemData.name.."|r", self.qTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.questsCellPadding + prdb.otherItemIconSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "questItem", "icon", item, itemData.icon)
			end
		end
		self.replaceHeader = nil
	end
	if prdb.showUsableQuestItemsNotAttached and core.numUsableQuestItems > 0 then
		self.tq = true
		local zone = LQTr["Usable quest items not linked to a quest"]
		if not self.replaceHeader or not self.tempLine then
			self.tempLine = self.tracker:AddLine()
		end
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		if prdb.colorizeHeaderName then
			self.tracker:SetCell(self.tempLine, 1, "|cffffffff"..zone.."|r", self.hTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
		else
			self.tracker:SetCell(self.tempLine, 1, "|cff"..prdb.defaultHeaderNameColor..zone.."|r", self.hTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.collapseButtonSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "zone", "icon", zone, chdb.zoneDisplay[zone] == 1 and self.minus or self.plus)
		end
		if chdb.zoneDisplay[zone] == 1 then
			for item, itemData in pairs(core.usableQuestItems) do
				self.tempLine = self.tracker:AddLine()
				self.tracker:SetCell(self.tempLine, 1, "|cff"..itemData.color..itemData.name.."|r", self.qTrFontTQ2, "LEFT", self.numColumns, cProvider, prdb.questsCellPadding + prdb.otherItemIconSize, prdb.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "questItem", "icon", item, itemData.icon)
			end
		end
		self.replaceHeader = nil
	end
	if self.replaceHeader then
		self.tracker:SetCell(self.tempLine, 1, nil, self.hTrFontTQ2, "LEFT", self.numColumns)
	end
end

function module:updateQuestsTracker()
	if prdb.toggleTracker and not self.hideInCombat then
		if core.moduleState["achievementTracker"] and achievementTracker.db.profile.showInQuestTracker then
			if achievementTracker.numTrackedAchievement == 0 then 
				self.achievementDisplayed = nil
				self.numColumns = 2
			else
				self.achievementDisplayed = true
				self.numColumns = achievementTracker.db.profile.numBooleanCriteriaColumns + 1
			end
			self.tracker = QTIP:Acquire("tomQuest2Tracker", self.numColumns)
			achievementTracker.tracker = self.tracker
			self.tracker:Clear()
			self.tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			self.tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			self.tracker:SetScale(prdb.scale)
			self.tracker:SetFrameStrata("BACKGROUND")
			if self.achievementDisplayed and achievementTracker.db.profile.positionInQuestTracker == 1 then
				achievementTracker:updateAchievements()
				self:updateQuests()
			else
				self:updateQuests()
				if self.achievementDisplayed then
					achievementTracker:updateAchievements()
				elseif not self.tq then
					QTIP:Release(self.tracker)
					achievementTracker.tracker = nil
					self.tracker = nil
					return
				end
			end
			local tempLine = self.tracker:AddLine()
			for i = 1, self.numColumns - 1 do
				self.tracker:SetCell(tempLine, i, " ", self.fixQTrFontTQ2, "LEFT", 1, nil, 0, 0, self.maxWidth/(self.numColumns - 1), self.minWidth/(self.numColumns - 1))
			end
			self.tracker:SetCell(tempLine, self.numColumns, " ", self.fixQTrFontTQ2, "RIGHT", 1, nil, 0, 0, self.statusMaxWidth, self.statusMinWidth)
		else
			self.numColumns = 2
			self.tracker = QTIP:Acquire("tomQuest2Tracker", 2)
			self.tracker:Clear()
			self.tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			self.tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			self.tracker:SetScale(prdb.scale)
			self.tracker:SetFrameStrata("BACKGROUND")
			self:updateQuests()
			if not self.tq then
				QTIP:Release(self.tracker)
				self.tracker = nil
				return
			end
			local tempLine = self.tracker:AddLine()
			self.tracker:SetCell(tempLine, 1, " ", self.fixQTrFontTQ2, "LEFT", 1, nil, 0, 0, self.maxWidth, self.minWidth)
			self.tracker:SetCell(tempLine, 2, " ", self.fixQTrFontTQ2, "RIGHT", 1, nil, 0, 0, self.statusMaxWidth, self.statusMinWidth)
		end
	else
		QTIP:Release(self.tracker)
		self.tracker = nil
		return
	end
	self.tracker:ClearAllPoints()
	if prdb.showAnchor then
		self.tracker:SetPoint(self.anchorSetPoint[prdb.anchorPosition].point, self.anchor, self.anchorSetPoint[prdb.anchorPosition].relPoint)
	else
		self.tracker:SetPoint(self.anchorSetPoint[prdb.anchorPosition].point, self.anchor, self.anchorSetPoint[prdb.anchorPosition].relPoint, self.anchorSetPoint[prdb.anchorPosition].x, self.anchorSetPoint[prdb.anchorPosition].y)
	end
	if not self.disableInteraction then
		self.tracker:UpdateScrolling(prdb.hasMaxHeight and prdb.maxHeight or nil)
		self.tracker:EnableMouse(true)
	else
		self.tracker:EnableMouse(false)
	end
	self.tracker:Show()
end

function module:checkTimer()
	local currentTime = time()
	local updateTracker
	for uid, endTime in pairs(self.autoUntrack) do
		if currentTime >= endTime then
			RemoveQuestWatch(core.quests[uid].id, uid, true)
			updateTracker = true
		end
	end
	for uid, endTime in pairs(self.delayedHide) do
		if currentTime >= endTime then
			updateTracker = true
			self.delayedHide[uid] = nil
			self.numDelayedHide = self.numDelayedHide - 1
			if self.numAutoUntrack == 0 and self.numDelayedHide == 0 and self.timerHandler then
				self:CancelTimer(self.timerHandler, true)
				self.timerHandler = nil
			end
		end
	end
	if updateTracker then
		self:updateQuestsTracker()
	end
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2QUESTTRACKERTOGGLE = LQTr["Quests Tracker"]

function tQ2toggleQuestTracker(display)
	if not display then
		prdb.toggleTracker = not prdb.toggleTracker
		if not prdb.toggleTracker then
			QTIP:Release(module.tracker)
			module.tracker = nil
		else
			module:updateQuestsTracker()
		end
	elseif display == 1 and not prdb.toggleTracker then
		prdb.toggleTracker = not prdb.toggleTracker
		module:updateQuestsTracker()
	elseif display == 0 and prdb.toggleTracker then
		prdb.toggleTracker = not prdb.toggleTracker
		QTIP:Release(module.tracker)
		module.tracker = nil
	end
end